/*-
* Copyright (c) 2008 Damien Bergamini <damien.bergamini@free.fr>
*
* Permission to use, copy, modify, and distribute this software for any
* purpose with or without fee is hereby granted, provided that the above
* copyright notice and this permission notice appear in all copies.
*
* THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
* WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
* ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
* WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
* ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
* OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
* This code implements the HMAC algorithm described in RFC 2104 using
* the MD5, SHA1 and SHA-256 hash functions.
*/

#include <stdint.h>
#include <string.h>
#include "amd_md5.h"
#include "amd_sha1.h"
//#include "sha2.h"
#include "amd_hmac.h"

void
  AMD_HMAC_MD5_Init(AMD_HMAC_MD5_CTX *ctx, const uint8_t *key, unsigned int key_len)
{
  uint8_t k_ipad[AMD_MD5_BLOCK_LENGTH];
  int i;

  if (key_len > AMD_MD5_BLOCK_LENGTH) {
    AMD_MD5Init(&ctx->ctx);
    AMD_MD5Update(&ctx->ctx, key, key_len);
    AMD_MD5Final(ctx->key, &ctx->ctx);
    ctx->key_len = AMD_MD5_DIGEST_LENGTH;
  } else {
    memcpy(ctx->key, key, key_len);
    ctx->key_len = key_len;
  }

  memset(k_ipad, 0, AMD_MD5_BLOCK_LENGTH);
  memcpy(k_ipad, ctx->key, ctx->key_len);
  for (i = 0; i < AMD_MD5_BLOCK_LENGTH; i++)
    k_ipad[i] ^= 0x36;

  AMD_MD5Init(&ctx->ctx);
  AMD_MD5Update(&ctx->ctx, k_ipad, AMD_MD5_BLOCK_LENGTH);

  memset(k_ipad, 0, sizeof k_ipad);
}

void
  AMD_HMAC_MD5_Update(AMD_HMAC_MD5_CTX *ctx, const uint8_t *data, unsigned int len)
{
  AMD_MD5Update(&ctx->ctx, data, len);
}

void
  AMD_HMAC_MD5_Final(uint8_t digest[AMD_MD5_DIGEST_LENGTH], AMD_HMAC_MD5_CTX *ctx)
{
  uint8_t k_opad[AMD_MD5_BLOCK_LENGTH];
  int i;

  AMD_MD5Final(digest, &ctx->ctx);

  memset(k_opad, 0, AMD_MD5_BLOCK_LENGTH);
  memcpy(k_opad, ctx->key, ctx->key_len);
  for (i = 0; i < AMD_MD5_BLOCK_LENGTH; i++)
    k_opad[i] ^= 0x5c;

  AMD_MD5Init(&ctx->ctx);
  AMD_MD5Update(&ctx->ctx, k_opad, AMD_MD5_BLOCK_LENGTH);
  AMD_MD5Update(&ctx->ctx, digest, AMD_MD5_DIGEST_LENGTH);
  AMD_MD5Final(digest, &ctx->ctx);

  memset(k_opad, 0, sizeof k_opad);
}

void
  AMD_HMAC_SHA1_Init(AMD_HMAC_SHA1_CTX *ctx, const uint8_t *key, unsigned int key_len)
{
  uint8_t k_ipad[AMD_SHA1_BLOCK_LENGTH];
  int i;

  if (key_len > AMD_SHA1_BLOCK_LENGTH) {
    AMD_SHA1Init(&ctx->ctx);
    AMD_SHA1Update(&ctx->ctx, key, key_len);
    AMD_SHA1Final(ctx->key, &ctx->ctx);
    ctx->key_len = AMD_SHA1_DIGEST_LENGTH;
  } else {
    memcpy(ctx->key, key, key_len);
    ctx->key_len = key_len;
  }

  memset(k_ipad, 0, AMD_SHA1_BLOCK_LENGTH);
  memcpy(k_ipad, ctx->key, ctx->key_len);
  for (i = 0; i < AMD_SHA1_BLOCK_LENGTH; i++)
    k_ipad[i] ^= 0x36;

  AMD_SHA1Init(&ctx->ctx);
  AMD_SHA1Update(&ctx->ctx, k_ipad, AMD_SHA1_BLOCK_LENGTH);

  memset(k_ipad, 0, sizeof k_ipad);
}

void
  AMD_HMAC_SHA1_Update(AMD_HMAC_SHA1_CTX *ctx, const uint8_t *data, unsigned int len)
{
  AMD_SHA1Update(&ctx->ctx, data, len);
}

void
  AMD_HMAC_SHA1_Final(uint8_t digest[AMD_SHA1_DIGEST_LENGTH], AMD_HMAC_SHA1_CTX *ctx)
{
  uint8_t k_opad[AMD_SHA1_BLOCK_LENGTH];
  int i;

  AMD_SHA1Final(digest, &ctx->ctx);

  memset(k_opad, 0, AMD_SHA1_BLOCK_LENGTH);
  memcpy(k_opad, ctx->key, ctx->key_len);
  for (i = 0; i < AMD_SHA1_BLOCK_LENGTH; i++)
    k_opad[i] ^= 0x5c;

  AMD_SHA1Init(&ctx->ctx);
  AMD_SHA1Update(&ctx->ctx, k_opad, AMD_SHA1_BLOCK_LENGTH);
  AMD_SHA1Update(&ctx->ctx, digest, AMD_SHA1_DIGEST_LENGTH);
  AMD_SHA1Final(digest, &ctx->ctx);

  memset(k_opad, 0, sizeof k_opad);
}

#if 0

void
  HMAC_SHA256_Init(HMAC_SHA256_CTX *ctx, const uint8_t *key, unsigned int key_len)
{
  uint8_t k_ipad[SHA256_BLOCK_LENGTH];
  int i;

  if (key_len > SHA256_BLOCK_LENGTH) {
    SHA256Init(&ctx->ctx);
    SHA256Update(&ctx->ctx, key, key_len);
    SHA256Final(ctx->key, &ctx->ctx);
    ctx->key_len = SHA256_DIGEST_LENGTH;
  } else {
    memcpy(ctx->key, key, key_len);
    ctx->key_len = key_len;
  }

  memset(k_ipad, 0, SHA256_BLOCK_LENGTH);
  memcpy(k_ipad, ctx->key, ctx->key_len);
  for (i = 0; i < SHA256_BLOCK_LENGTH; i++)
    k_ipad[i] ^= 0x36;

  SHA256Init(&ctx->ctx);
  SHA256Update(&ctx->ctx, k_ipad, SHA256_BLOCK_LENGTH);

  memset(k_ipad, 0, sizeof k_ipad);
}

void
  HMAC_SHA256_Update(HMAC_SHA256_CTX *ctx, const uint8_t *data, unsigned int len)
{
  SHA256Update(&ctx->ctx, data, len);
}

void
  HMAC_SHA256_Final(uint8_t digest[SHA256_DIGEST_LENGTH], HMAC_SHA256_CTX *ctx)
{
  uint8_t k_opad[SHA256_BLOCK_LENGTH];
  int i;

  SHA256Final(digest, &ctx->ctx);

  memset(k_opad, 0, SHA256_BLOCK_LENGTH);
  memcpy(k_opad, ctx->key, ctx->key_len);
  for (i = 0; i < SHA256_BLOCK_LENGTH; i++)
    k_opad[i] ^= 0x5c;

  SHA256Init(&ctx->ctx);
  SHA256Update(&ctx->ctx, k_opad, SHA256_BLOCK_LENGTH);
  SHA256Update(&ctx->ctx, digest, SHA256_DIGEST_LENGTH);
  SHA256Final(digest, &ctx->ctx);

  memset(k_opad, 0, sizeof k_opad);
}

#endif
