/*
  The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/***************************************************************************
 *            jitterctl.c
 *
 *  Mon Nov  8 11:53:21 2004
 *  Copyright  2004  Simon MORLAT
 *  Email simon.morlat@linphone.org
 ****************************************************************************/

#include "ortp/rtpsession.h"
#include "ortp/payloadtype.h"
#include "ortp/ortp.h"
#include "utils.h"
#include "rtpsession_priv.h"
#include <math.h>

#define JC_BETA_SLOW 0.001
#define JC_BETA 1

#define JC_GAMMA (0.03)

#include "jitterctl.h"

void jitter_control_init(JitterControl *ctl, int base_jiitt_time, PayloadType *payload){
	ctl->count=0;
	ctl->slide=0;
	ctl->prev_slide=0;
	ctl->jitter=0;
	ctl->inter_jitter=0;
	ctl->slide=0;
	if (base_jiitt_time!=-1) ctl->jitt_comp = base_jiitt_time;
	/* convert in timestamp unit: */
	if (payload!=NULL){
		jitter_control_set_payload(ctl,payload);
	}
	ctl->adapt_jitt_comp_ts=ctl->jitt_comp_ts;
	ctl->corrective_slide=0;
	ctl->burst_factor=0;
	ctl->average_buffer_size=0;
	ctl->average_buffer_size_count=0;
}

void jitter_control_enable_adaptive(JitterControl *ctl, bool_t val){
	ctl->adaptive=val;
}

void jitter_control_set_payload(JitterControl *ctl, PayloadType *pt){
  if (ctl->clock_rate != pt->clock_rate) {
    ctl->count=0;
    ctl->inter_jitter=0;
    ctl->jitter=0;
    ctl->max_jitter=0;
  }
	ctl->clock_rate = pt->clock_rate;
	ctl->jitt_comp_ts =
			(int) (((double) ctl->jitt_comp / 1000.0) * (pt->clock_rate));
	ctl->corrective_step=(int)((0.01)*pt->clock_rate); /* This formula got to me after some beers */
	ctl->adapt_jitt_comp_ts=ctl->jitt_comp_ts;
	ctl->count=0;
	/* ortp_message("DJB nominal: %i", ctl->adapt_jitt_comp_ts*1000/ctl->clock_rate); */
}

#if 0
static void jitter_control_dump_stats(JitterControl *ctl){
	ortp_message("JitterControl:\n\tslide=%f,jitter=%g,inter_jitter=%f, adapt_jitt_comp_ts=%i, count=%i",
			(double)ctl->slide,ctl->jitter, ctl->inter_jitter, ctl->adapt_jitt_comp_ts, ctl->count);
}
#endif

/*the goal of this method is to compute "corrective_slide": a timestamp unit'd value to be added
 to recv timestamp to make them reflect the instant they are delivered by the jitter buffer. */
void jitter_control_update_corrective_slide(JitterControl *ctl){
	int tmp;
	tmp=(int)(ctl->slide-ctl->prev_slide);
	if (tmp>ctl->corrective_step) {
		ctl->corrective_slide+=ctl->corrective_step;
		ctl->prev_slide=ctl->slide+ctl->corrective_step;
		//ortp_message("clock_rate=%i, corrective_slide is updated: %ims", ctl->clock_rate, ctl->corrective_slide*1000/ctl->clock_rate);
	}
	else if (tmp<-ctl->corrective_step) {
		ctl->corrective_slide-=ctl->corrective_step;
		ctl->prev_slide=ctl->slide-ctl->corrective_step;
		//ortp_message("clock_rate=%i, corrective_slide is updated: %ims", ctl->clock_rate, ctl->corrective_slide*1000/ctl->clock_rate);
	}
}

void jitter_control_update_average_buffer_size(JitterControl *ctl, mblk_t *first, mblk_t *last){
	uint32_t first_ts,last_ts;
	if (first==NULL) return;
	first_ts=rtp_get_timestamp(first);
	last_ts=rtp_get_timestamp(last);
	ctl->average_buffer_size+=(uint32_t)(first_ts-last_ts);
	ctl->average_buffer_size_count++;
}

/*
The algorithm computes two values:
slide: an average of difference between the expected and the socket-received timestamp
jitter: an average of the absolute value of the difference between socket-received timestamp and slide.
slide is used to make clock-slide detection and correction.
jitter is added to the initial jitt_comp_time value. It compensates bursty packets arrival (packets
not arriving at regular interval ).
*/
void jitter_control_new_packet(JitterControl *ctl, uint32_t packet_ts, uint32_t cur_str_ts){
  int32_t diff=packet_ts - cur_str_ts;
  int32_t slide;
  int32_t gap;
  int32_t d;
  int32_t d_slide;

  /* control looping packet_ts */
  if (packet_ts<ctl->prev_packet_ts && ctl->prev_packet_ts>0xFFFF0000) {
    /* loop is happening */
    ortp_warning("rotation detected for remote timestamp (packet_ts=%u cur_str_ts:%u)", packet_ts, cur_str_ts);
  }
  ctl->prev_packet_ts = packet_ts;
  /* control looping cur_str_ts */
  if (cur_str_ts<ctl->prev_app_ts && ctl->prev_app_ts>0xFFFF0000) {
    ortp_warning("clock_rate=%i: rotation detected for local timestamp (packet_ts=%u cur_str_ts:%u)", ctl->clock_rate, packet_ts, cur_str_ts);
  }
  ctl->prev_app_ts = cur_str_ts;

  d=(int32_t)(diff-ctl->olddiff);
  d_slide=(int32_t)(diff-ctl->slide);

  if (d>ctl->clock_rate || d<-ctl->clock_rate) {
    ortp_warning("clock_rate=%i: unexpected difference: packet_ts=%u cur_str_ts=%u ctl->slide=%u, d=%i (%ims) diff=%u olddiff:%u", ctl->clock_rate, packet_ts, cur_str_ts, ctl->slide, d, d/(ctl->clock_rate/1000), diff, ctl->olddiff);
  }

#if 0
  ortp_warning("packet_ts=%u cur_str_ts=%u ctl->slide=%i, d=%i diff=%i olddiff:%i (diff-ctl->olddiff=%i) (ctl->slide-ctl->olddiff=%i)", packet_ts, cur_str_ts, ctl->slide, d, diff, ctl->olddiff, diff - ctl->olddiff, ctl->slide - ctl->olddiff);
#endif
  if (((ctl->count<6 && ctl->clock_rate!=90000)) || ((ctl->count<20 && ctl->clock_rate==90000))) {
    /* if count=0 && payload change with sample rate change, the current value for cur_str_ts
    (session->rtp.rcv_last_app_ts) is wrong because the original value is coming from an
    uint64_t. Waiting for count=20 allow to make sure the values are accurate according to
    newer samplerate. */
    slide=ctl->slide=ctl->prev_slide=diff;
    ctl->olddiff=diff;
    ctl->jitter=0;
  }else{
    /* slide will increase slowly, but if we detect early packets, we want to revert faster to a better slide. */
    if (ctl->slide - ctl->olddiff<0)
      slide = (double)ctl->slide*(1 - JC_BETA) + (double)diff*(JC_BETA);
    else
      slide = (double)ctl->slide*(1 - JC_BETA_SLOW) + (double)diff*(JC_BETA_SLOW);
  }
  gap=(int32_t)(diff - slide);
  gap=gap<0 ? -gap : 0; /*compute only for late packets*/
  if (gap>ctl->adapt_jitt_comp_ts+ctl->corrective_step)
    ctl->cum_late_packet++;
  if (gap>ctl->cur_max_late_packet)
    ctl->cur_max_late_packet=(float)gap;
  ctl->jitter=(float) ((ctl->jitter*(1-JC_GAMMA)) + (gap*JC_GAMMA));
  ctl->inter_jitter=(float) (ctl->inter_jitter+ (( (float)abs(d) - ctl->inter_jitter)*(1/16.0)));

  ctl->olddiff=diff;
  ctl->count++;
  if (ctl->adaptive){
#if 0
    if (gap>ctl->adapt_jitt_comp_ts) {
      ctl->burst_factor++;
      ortp_warning("late packet detected %i", ctl->burst_factor);
      ctl->adapt_jitt_comp_ts+=ctl->corrective_step;
    }
#endif
    if (ctl->max_jitter< (float)gap) {
      ctl->max_jitter= (float)gap;
    }
    if (ctl->count%50==0) {
      ctl->cum_late_packet=0;
      ctl->adapt_jitt_comp_ts=(int) MAX(ctl->jitt_comp_ts,2*ctl->jitter);
      ctl->adapt_jitt_comp_ts=(int)MAX(ctl->adapt_jitt_comp_ts, ctl->max_jitter*1.2f);
      if (ctl->adapt_jitt_comp_ts * 10 / ctl->clock_rate > 15)
        ctl->adapt_jitt_comp_ts = (15 * ctl->clock_rate) / 10;

      if (ctl->clock_rate!=90000) {
        if (ctl->cur_max_late_packet>10)
          ctl->adapt_jitt_comp_ts=ctl->adapt_jitt_comp_ts+ctl->corrective_step*2;
      }
      ctl->max_jitter=ctl->max_jitter*98/100;
      ctl->cur_max_late_packet=ctl->cur_max_late_packet*90/100;
#if 0
      ctl->adapt_jitt_comp_ts+=ctl->burst_factor*ctl->corrective_step;
      jitter_control_dump_stats(ctl);
#endif
      /* ortp_message("DJB nominal: %ims + corrective_slide=%ims", ctl->adapt_jitt_comp_ts*1000/ctl->clock_rate, ctl->corrective_slide*1000/ctl->clock_rate); */
#define ENABLE_JITTER_LOG
#ifdef ENABLE_JITTER_LOG
      ortp_warning("adaptive jitter set to = %ims // jitter = %f // inter_jitter = %f", ctl->adapt_jitt_comp_ts*1000/ctl->clock_rate, ctl->jitter, ctl->inter_jitter);
#endif
    }
    /* ortp_warning("jitter = %f // inter_jitter = %f // slide = %f // gap = %f", ctl->jitter, ctl->inter_jitter, slide, gap); */
    ctl->slide=slide;
  }else {
    /*ctl->slide and jitter size are not updated*/
  }
  return ;
}





/**
 *rtp_session_set_jitter_compensation:
 *@session: a RtpSession
 *@milisec: the time interval in milisec to be jitter compensed.
 *
 * Sets the time interval for which packet are buffered instead of being delivered to the 
 * application.
 **/
void
rtp_session_set_jitter_compensation (RtpSession * session, int milisec)
{
	PayloadType *payload=NULL;
	if (session->rcv.pt!=-1) {
		payload = rtp_profile_get_payload (session->rcv.profile,session->rcv.pt);
	}/*else not set yet */
	jitter_control_init(&session->rtp.jittctl,milisec,payload);
}

void rtp_session_enable_adaptive_jitter_compensation(RtpSession *session, bool_t val){
	jitter_control_enable_adaptive(&session->rtp.jittctl,val);
}

bool_t rtp_session_adaptive_jitter_compensation_enabled(RtpSession *session){
	return session->rtp.jittctl.adaptive;
}

void rtp_session_enable_jitter_buffer(RtpSession *session, bool_t enabled){
	session->rtp.jittctl.enabled=enabled;
	session->flags|=RTP_SESSION_RECV_SYNC;
}

bool_t rtp_session_jitter_buffer_enabled(const RtpSession *session){
	return session->rtp.jittctl.enabled;
}

void rtp_session_set_jitter_buffer_params(RtpSession *session, const JBParameters *par){
	/* FIXME min_size and max_size to be implemented */
	rtp_session_set_jitter_compensation(session,par->nom_size);
	jitter_control_enable_adaptive(&session->rtp.jittctl,par->adaptive);
	session->rtp.max_rq_size=par->max_packets;
}

void rtp_session_get_jitter_buffer_params(RtpSession *session, JBParameters *par){
	int nom_size=session->rtp.jittctl.jitt_comp;
	par->min_size=nom_size;
	par->nom_size=nom_size;
	par->max_size=-1;
	par->adaptive=session->rtp.jittctl.adaptive;
	par->max_packets=session->rtp.max_rq_size;
}
