/*
  The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#if (_MSC_VER >= 1700) && !defined(_USING_V110_SDK71_)
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#define DISABLE_RTP_SCHEDULER
#endif
#endif

#if defined(WIN32) || defined(_WIN32_WCE)
#include "ortp-config-win32.h"
#elif HAVE_CONFIG_H
#include "ortp-config.h"
#endif
#include "ortp/ortp.h"

#if defined(__linux) || defined(__APPLE__)
#include <inttypes.h>
#endif

#if defined(HAVE_OPENSSL_SSL_H) && defined(HAVE_SRTP)
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/conf.h>
#endif

#ifdef TSC_SUPPORT
//#define TSC_REDUNDANCY

#include "tsc_socket_api.h"
#include "tsc_control_api.h"
#endif

int rtp_transport_setup(void);
int turn_transport_setup(void);

int srtp_module_setup(void);
int dtls_srtp_module_setup(void);
int rtprewrite_module_setup(void);

rtp_stats_t ortp_global_stats;

#ifdef ENABLE_MEMCHECK
int ortp_allocations=0;
#endif


extern void av_profile_init(RtpProfile *profile);

void init_random_number_generator(void);

void init_random_number_generator(void) {
	uint64_t tick;
	int seed;

#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
	tick = GetTickCount64 ();
#elif defined(_WIN32_WCE) || defined(_WIN64)
	tick = GetTickCount ();
#elif defined(_MSC_VER)
	{
		volatile unsigned int lowtick=0,hightick=0;
		__asm
		{
			rdtsc 
				mov lowtick, eax
				mov hightick, edx
		}
		tick = hightick;
		tick <<= 32;
		tick |= lowtick;
	}
#elif defined(__MACH__) 
	{
		int fd=open("/dev/random",O_RDONLY);
		read(fd,&tick,sizeof(tick));
		close(fd);
	}
#elif defined(__GNUC__) && ( defined(__i686__) || defined(__i386__) )
	asm("rdtsc" : "=A" (tick));
#elif defined(__GNUC__) && defined(__amd64__)
	asm("rdtsc" : "=A" (tick));
#elif defined (__SUNPRO_CC) && defined( __sparc__ )	
	tick = gethrtime();
#elif defined(__linux) || defined(__linux__) || defined(HAVE_DEV_RANDOM) 
	{
		fd_set fdSet;
		int maxFd=0;
		struct timeval tv;
		int e;

		int fd=open("/dev/random",O_RDONLY);

		if (fd<0)
		{
			struct timeval t;
			gettimeofday(&t,NULL);
			srandom((unsigned int)(t.tv_usec+t.tv_sec));
			ortp_message("init_random_number_generator: Failed to open random device");
			return;
		}
		FD_ZERO(&fdSet);
		FD_SET(fd,&fdSet);
		maxFd=fd+1;

		tv.tv_sec = 0;
		tv.tv_usec = 500;

		e = select( maxFd, &fdSet, NULL,NULL, &tv );
		if (e <= 0)
		{
			struct timeval t;
			gettimeofday(&t,NULL);
			srandom((unsigned int)(t.tv_usec+t.tv_sec));
			ortp_error("init_random_number_generator: Failed to get data from random device");
			close(fd);
			return;
		}
		read(fd,&tick,sizeof(tick));
		close(fd);
	}
#else
#     error Need some way to seed the random number generator 
#endif 
	seed = (int)(tick);
#if	defined(_WIN32) || defined(_WIN32_WCE)
	srand(seed);
#else
	srandom(seed);
#endif
}

#ifdef WIN32
#define WIN32_USE_CRYPTO
#endif

#if (_MSC_VER >= 1700) && !defined(_USING_V110_SDK71_)
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#undef WIN32_USE_CRYPTO
#elif defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PC_APP)
#undef WIN32_USE_CRYPTO
#endif
#endif

#ifdef WIN32_USE_CRYPTO
#include <Wincrypt.h>
#endif

uint32_t ortp_random(void){
#if defined(WIN32_USE_CRYPTO) && (_MSC_VER >= 1300)
	HCRYPTPROV crypto;
	BOOL err;
	uint32_t num = 0xAAAABBBB;
	err = CryptAcquireContext(&crypto, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT);
	if (err) {
		err = CryptGenRandom(crypto, sizeof(num), (BYTE *) & num);
		CryptReleaseContext(crypto, 0);
	}
	if (!err || num==0xAAAABBBB) {
		return (uint32_t)random();
	}
	return num;
#else
	return (uint32_t)random();
#endif
}

#ifdef WIN32
static bool_t win32_init_sockets(void){
	WORD wVersionRequested;
	WSADATA wsaData;
	int i;
	
	wVersionRequested = MAKEWORD(2,0);
	if( (i = WSAStartup(wVersionRequested,  &wsaData))!=0)
	{
		ortp_error("Unable to initialize windows socket api, reason: %d (%s)",i,getWinSocketError(i));
		return FALSE;
	}
	return TRUE;
}
#endif

/**
 *	Initialize the oRTP library. You should call this function first before using
 *	oRTP API.
**/
void ortp_init()
{
	static bool_t initialized=FALSE;
	if (initialized) {
		init_random_number_generator();
		return;
	}

	initialized=TRUE;

#ifdef WIN32
	win32_init_sockets();
#endif

	av_profile_init(&av_profile);
	ortp_global_stats_reset();
	init_random_number_generator();
	ortp_message("oRTP-" ORTP_VERSION " initialized.");

	rtp_transport_setup();
	turn_transport_setup();

  srtp_module_setup();
  dtls_srtp_module_setup();
  rtprewrite_module_setup();
}

/**
 * Gracefully uninitialize the library, including shutdowning the scheduler if it was started.
 *	
**/
void ortp_exit()
{
  ortp_transport_remove_all();
  ortp_module_remove_all();

#if defined(HAVE_OPENSSL_SSL_H) && defined(HAVE_SRTP)
  FIPS_mode_set(0);
  CONF_modules_unload(1);
#if !(OPENSSL_VERSION_NUMBER < 0x10002000L)
  SSL_COMP_free_compression_methods();
#endif
  EVP_cleanup();
  CRYPTO_cleanup_all_ex_data();
  ERR_free_strings();
#endif
}

static FILE *__log_file=0;

/**
 *@param file a FILE pointer where to output the ortp logs.
 *
**/
void ortp_set_log_file(FILE *file)
{
	__log_file=file;
}

static void __ortp_logv_out(OrtpLogLevel lev, const char *fmt, va_list args);

OrtpLogFunc ortp_logv_out=__ortp_logv_out;

/**
 *@param func: your logging function, compatible with the OrtpLogFunc prototype.
 *
**/
void ortp_set_log_handler(OrtpLogFunc func){
	ortp_logv_out=func;
}


unsigned int __ortp_log_mask=ORTP_WARNING|ORTP_ERROR|ORTP_FATAL;

/**
 * @ param levelmask a mask of ORTP_DEBUG, ORTP_MESSAGE, ORTP_WARNING, ORTP_ERROR
 * ORTP_FATAL .
**/
void ortp_set_log_level_mask(int levelmask){
	__ortp_log_mask=levelmask;
}

char * ortp_strdup_vprintf(const char *fmt, va_list ap)
{
	/* Guess we need no more than 100 bytes. */
	int n, size = 200;
	char *p,*np;
#if defined(__linux) || defined(__APPLE__)
	va_list cap;/*copy of our argument list: a va_list cannot be re-used (SIGSEGV on linux 64 bits)*/
#endif
	if ((p = (char *) ortp_malloc (size)) == NULL)
		return NULL;
	while (1)
	{
		/* Try to print in the allocated space. */
#if defined(__linux) || defined(__APPLE__)
		va_copy(cap,ap);
		n = vsnprintf (p, size, fmt, cap);
		va_end(cap);
#else
		/*this works on 32 bits, luckily*/
		n = vsnprintf (p, size, fmt, ap);
#endif
		/* If that worked, return the string. */
		if (n > -1 && n < size)
			return p;
		/* Else try again with more space. */
		if (n > -1)	/* glibc 2.1 */
			size = n + 1;	/* precisely what is needed */
		else		/* glibc 2.0 */
			size *= 2;	/* twice the old size */
		if ((np = (char *) ortp_realloc (p, size)) == NULL)
		  {
		    free(p);
		    return NULL;
		  }
		else
		  {
		    p = np;
		  }
	}
	return NULL;
}

char *ortp_strdup_printf(const char *fmt,...){
	char *ret;
	va_list args;
	va_start (args, fmt);
	ret=ortp_strdup_vprintf(fmt, args);
 	va_end (args);
	return ret;
}

#if	defined(WIN32) || defined(_WIN32_WCE)
#define ENDLINE "\r\n"
#else
#define ENDLINE "\n"
#endif

void ortp_logv(int level, const char *fmt, va_list args)
{
	if (ortp_logv_out!=NULL && ortp_log_level_enabled(level))
		ortp_logv_out(level,fmt,args);
#if !defined(_WIN32_WCE)
	if ((level)==ORTP_FATAL) abort();
#endif
}

static void __ortp_logv_out(OrtpLogLevel lev, const char *fmt, va_list args){
	const char *lname="undef";
	char *msg;
	if (__log_file==NULL) __log_file=stderr;
	switch(lev){
		case ORTP_DEBUG:
			lname="debug";
			break;
		case ORTP_MESSAGE:
			lname="message";
			break;
		case ORTP_WARNING:
			lname="warning";
			break;
		case ORTP_ERROR:
			lname="error";
			break;
		case ORTP_FATAL:
			lname="fatal";
			break;
		default:
			ortp_fatal("Bad level !");
			break;
	}
	msg=ortp_strdup_vprintf(fmt,args);
#if defined(_MSC_VER) && !defined(_WIN32_WCE)
#ifdef UNICODE
    {
#ifndef MAX_LENGTH_TR
#define MAX_LENGTH_TR 512       /* NEVER DEFINE MORE THAN 2024 */
#endif
      WCHAR wUnicode[MAX_LENGTH_TR * 2];

      MultiByteToWideChar (CP_UTF8, 0, msg, -1, wUnicode, MAX_LENGTH_TR * 2);
      OutputDebugString (wUnicode);
      OutputDebugString(L"\r\n");
    }
#else
 	OutputDebugStringA(msg);
  	OutputDebugStringA("\r\n");
#endif
#else
	fprintf(__log_file,"ortp-%s-%s" ENDLINE,lname,msg);
	fflush(__log_file);
#endif

	ortp_free(msg);
}

/**
 * Display global statistics (cumulative for all RtpSession)
**/
void ortp_global_stats_display()
{
	rtp_stats_display(&ortp_global_stats,"Global statistics");
#ifdef ENABLE_MEMCHECK	
	printf("Unfreed allocations: %i\n",ortp_allocations);
#endif
}

/**
 * Print RTP statistics.
**/
void rtp_stats_display(const rtp_stats_t *stats, const char *header)
{
#if defined(__linux)
  ortp_log(ORTP_MESSAGE, "%s // sent=%" PRIu64 " received=%" PRIu64 " lost=%" PRIu64 ", late=%" PRIu64 ", bad=%" PRIu64 ", discarded=%" PRIu64 ", av.jitter=%.2f",
    header,
    (uint64_t)stats->packet_sent,
    (uint64_t)stats->packet_recv,
    (uint64_t)stats->cum_packet_loss,
    (uint64_t)stats->outoftime,
    (uint64_t)stats->bad,
    (uint64_t)stats->discarded,
    stats->average_jitter);
#elif defined(__APPLE__)
  ortp_log(ORTP_MESSAGE, "%s // sent=%" PRIu64 " received=%" PRIu64 " lost=%" PRIu64 ", late=%" PRIu64 ", bad=%" PRIu64 ", discarded=%" PRIu64 ", av.jitter=%.2f",
           header,
           (uint64_t)stats->packet_sent,
           (uint64_t)stats->packet_recv,
           (uint64_t)stats->cum_packet_loss,
           (uint64_t)stats->outoftime,
           (uint64_t)stats->bad,
           (uint64_t)stats->discarded,
           stats->average_jitter);
#elif !defined(WIN32)  
  ortp_log(ORTP_MESSAGE,
	   "oRTP-stats:\n   %s :",
	   header);
  ortp_log(ORTP_MESSAGE,
	   " number of rtp packet sent=%lld",
	   (long long)stats->packet_sent);
  ortp_log(ORTP_MESSAGE,
	   " number of rtp bytes sent=%lld bytes",
	   (long long)stats->sent);
  ortp_log(ORTP_MESSAGE,
	   " number of rtp packet received=%lld",
	   (long long)stats->packet_recv);
  ortp_log(ORTP_MESSAGE,
	   " number of rtp bytes received=%lld bytes",
	   (long long)stats->hw_recv);
  ortp_log(ORTP_MESSAGE,
	   " number of incoming rtp bytes successfully delivered to the application=%lld ",
	   (long long)stats->recv);
  ortp_log(ORTP_MESSAGE,
	   " number of times the application queried a packet that didn't exist=%lld ",
	   (long long)stats->unavaillable);
  ortp_log(ORTP_MESSAGE,
	   " number of rtp packet lost=%lld",
	   (long long) stats->cum_packet_loss);
  ortp_log(ORTP_MESSAGE,
	   " number of rtp packets received too late=%lld",
	   (long long)stats->outoftime);
  ortp_log(ORTP_MESSAGE,
	   " number of bad formatted rtp packets=%lld",
	   (long long)stats->bad);
  ortp_log(ORTP_MESSAGE,
	   " number of packet discarded because of queue overflow=%lld",
	   (long long)stats->discarded);
  ortp_log(ORTP_MESSAGE,
	   " current local average jitter buffer size=%.2f",
	   stats->average_jitter);
#else
  ortp_log(ORTP_MESSAGE, "%s // sent=%I64d received=%I64d lost=%I64d, late=%I64d, bad=%I64d, discarded=%I64d, av.jitter=%.2f",
    header,
    (uint64_t)stats->packet_sent,
    (uint64_t)stats->packet_recv,
    (uint64_t)stats->cum_packet_loss,
    (uint64_t)stats->outoftime,
    (uint64_t)stats->bad,
    (uint64_t)stats->discarded,
    stats->average_jitter);
#endif
}

void ortp_global_stats_reset(){
	memset(&ortp_global_stats,0,sizeof(rtp_stats_t));
}

rtp_stats_t *ortp_get_global_stats(){
	return &ortp_global_stats;
}

void rtp_stats_reset(rtp_stats_t *stats){
	memset((void*)stats,0,sizeof(rtp_stats_t));
}


/**
 * This function give the opportunity to programs to check if the libortp they link to
 * has the minimum version number they need.
 *
 * Returns: true if ortp has a version number greater or equal than the required one.
**/
bool_t ortp_min_version_required(int major, int minor, int micro){
	return ((major*1000000) + (minor*1000) + micro) <= 
		   ((ORTP_MAJOR_VERSION*1000000) + (ORTP_MINOR_VERSION*1000) + ORTP_MICRO_VERSION);
}
