
/*
  The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#if defined(WIN32) || defined(_WIN32_WCE)
#include "ortp-config-win32.h"
#elif HAVE_CONFIG_H
#include "ortp-config.h"
#endif
#include "ortp/port.h"
#include "utils.h"

#if	defined(_WIN32) && !defined(_WIN32_WCE)
#include <process.h>
#endif

//#define ORTP_MEMORY_DEBUG
#ifdef ORTP_MEMORY_DEBUG
/* this is to get file and line for malloc */
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>
#include "ortp/ortp.h"
#endif

static void *ortp_libc_malloc(size_t sz){
#ifdef ORTP_MEMORY_DEBUG
  void *ptr = malloc(sz);
  if (ptr == NULL)
    ortp_error("a NULL ptr is detected");
  return ptr; //malloc(sz);
#else
  return malloc(sz);
#endif
}

static void *ortp_libc_realloc(void *ptr, size_t sz){
#ifdef ORTP_MEMORY_DEBUG
  void *ptr_new = realloc(ptr, sz);
  if (ptr_new == NULL)
    ortp_error("a NULL ptr is detected");
  return ptr_new; //realloc(ptr, sz);
#else
  return realloc(ptr, sz);
#endif
}

static void ortp_libc_free(void*ptr){
	free(ptr);
}

static bool_t allocator_used=FALSE;

static OrtpMemoryFunctions ortp_allocator={
	ortp_libc_malloc,
	ortp_libc_realloc,
	ortp_libc_free
};

void ortp_set_memory_functions(OrtpMemoryFunctions *functions){
	if (allocator_used){
		return;
	}
	ortp_allocator=*functions;
}

void* ortp_malloc(size_t sz){
	allocator_used=TRUE;
	return ortp_allocator.malloc_fun(sz);
}

void* ortp_realloc(void *ptr, size_t sz){
	allocator_used=TRUE;
	return ortp_allocator.realloc_fun(ptr,sz);
}

void ortp_free(void* ptr){
	ortp_allocator.free_fun(ptr);
}

void * ortp_malloc0(size_t size){
	void *ptr=ortp_malloc(size);
	memset(ptr,0,size);
	return ptr;
}

char * ortp_strdup(const char *tmp){
	size_t sz;
	char *ret;
	if (tmp==NULL)
	  return NULL;
	sz=strlen(tmp)+1;
	ret=(char*)ortp_malloc(sz);
	strcpy(ret,tmp);
	ret[sz-1]='\0';
	return ret;
}

/*
 * this method is an utility method that calls fnctl() on UNIX or
 * ioctlsocket on Win32.
 * int retrun the result of the system method
 */
int set_non_blocking_socket (ortp_socket_t sock)
{


#if	!defined(_WIN32) && !defined(_WIN32_WCE)
	return fcntl (sock, F_SETFL, O_NONBLOCK);
#else
	unsigned long nonBlock = 1;
	return ioctlsocket(sock, FIONBIO , &nonBlock);
#endif
}

#if defined (_WIN32_WCE) || defined(_MSC_VER)
int ortp_file_exist(const char *pathname) {
	FILE* fd;
	if (pathname==NULL) return -1;
	fd=fopen(pathname,"r");
	if (fd==NULL) {
		return -1;
	} else {
		fclose(fd);
		return 0;
	}
}
#else
int ortp_file_exist(const char *pathname) {
	return access(pathname,F_OK);
}
#endif /*_WIN32_WCE*/

#if	!defined(_WIN32) && !defined(_WIN32_WCE)
	/* Use UNIX inet_aton method */
#else
	int inet_aton (const char * cp, struct in_addr * addr)
	{
		unsigned long retval;

		retval = inet_addr (cp);

		if (retval == INADDR_NONE)
		{
			return -1;
		}
		else
		{
			addr->S_un.S_addr = retval;
			return 1;
		}
	}
#endif

char *ortp_strndup(const char *str,int n){
	int min=MIN((int)strlen(str),n)+1;
	char *ret=(char*)ortp_malloc(min);
	strncpy(ret,str,n);
	ret[min-1]='\0';
	return ret;
}

#if	!defined(_WIN32) && !defined(_WIN32_WCE)
int __ortp_thread_join(ortp_thread_t thread, void **ptr){
	int err=pthread_join(thread,ptr);
	if (err!=0) {
		//ortp_error("pthread_join error: %s",ortp_strerror(err));
	}
	return err;
}

int __ortp_thread_create(pthread_t *thread, pthread_attr_t *attr, void * (*routine)(void*), void *arg){
	pthread_attr_t my_attr;
	pthread_attr_init(&my_attr);
	if (attr)
		my_attr = *attr;
#ifdef ORTP_DEFAULT_THREAD_STACK_SIZE
	if (ORTP_DEFAULT_THREAD_STACK_SIZE!=0)
		pthread_attr_setstacksize(&my_attr, ORTP_DEFAULT_THREAD_STACK_SIZE);
#endif
	return pthread_create(thread, &my_attr, routine, arg);
}

int __ortp_thread_set_name(pthread_t thread, const char *name){
#ifndef __APPLE__
  return pthread_setname_np(thread, name);
#else
  return 0;
#endif
}
#endif
#if	defined(_WIN32) || defined(_WIN32_WCE)

#include <limits.h>

#if (_WIN32_WINNT >= 0x0600) && !defined(_WIN32_WCE)
//amsip use recursive locks. InitializeSRWLock don't support this.
//#define USE_WIN32_CONDITION_VARIABLE
#endif

#if defined(USE_WIN32_CONDITION_VARIABLE)
struct _private_mutex {
  SRWLOCK h;
};

int WIN_mutex_init(ortp_mutex_t *mutex, void *attr)
{
  struct _private_mutex *mut = (struct _private_mutex *) ortp_malloc(sizeof(struct _private_mutex));

  if (mut == NULL)
    return -1;

  InitializeSRWLock(&mut->h);
  *mutex = mut;
  return 0;
}

int WIN_mutex_lock(ortp_mutex_t * hMutex)
{
  struct _private_mutex *mut = (struct _private_mutex *) *hMutex;

  if (mut == NULL)
    return -1;
  AcquireSRWLockExclusive(&mut->h);

  return 0;
}

int WIN_mutex_unlock(ortp_mutex_t * hMutex)
{
  struct _private_mutex *mut = (struct _private_mutex *) *hMutex;

  if (mut == NULL)
    return -1;
  ReleaseSRWLockExclusive(&mut->h);
  return 0;
}

int WIN_mutex_destroy(ortp_mutex_t * hMutex)
{
  struct _private_mutex *mut = (struct _private_mutex *) *hMutex;

  if (mut == NULL)
    return -1;
  ortp_free(mut);
  return 0;
}

#elif (_WIN32_WINNT >= 0x0403) && !defined(_WIN32_WCE)

#define OSIP_CRITICALSECTION_SPIN  4000
struct _private_mutex {
  CRITICAL_SECTION h;
};

int WIN_mutex_init(ortp_mutex_t *mutex, void *attr)
{
  struct _private_mutex *mut = (struct _private_mutex *) ortp_malloc (sizeof (struct _private_mutex));

  if (mut == NULL)
    return -1;
  
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
  if (InitializeCriticalSectionEx (&mut->h, OSIP_CRITICALSECTION_SPIN, CRITICAL_SECTION_NO_DEBUG_INFO) != 0) {
    *mutex = mut;
    return 0;
  }
#else
  /* look again at documentation for return code:
  Return value
	This function always returns a nonzero value.
	
	Windows Server 2003 and Windows XP:  If the function succeeds,
	the return value is nonzero. If the function fails, the return
	value is zero (0). To get extended error information, call
	GetLastError. This behavior was changed starting with Windows Vista.
*/
  if (InitializeCriticalSectionAndSpinCount (&mut->h, OSIP_CRITICALSECTION_SPIN) != 0) {
    *mutex = mut;
    return 0;
  }
#endif
  ortp_free (mut);
  return -1;
}

int WIN_mutex_lock(ortp_mutex_t * hMutex)
{
 struct _private_mutex *mut = (struct _private_mutex *) *hMutex;

  if (mut == NULL)
    return -1;
  EnterCriticalSection (&mut->h);

  return 0;
}

int WIN_mutex_unlock(ortp_mutex_t * hMutex)
{
  struct _private_mutex *mut = (struct _private_mutex *) *hMutex;

  if (mut == NULL)
    return -1;
  LeaveCriticalSection (&mut->h);
  return 0;
}

int WIN_mutex_destroy(ortp_mutex_t * hMutex)
{
  struct _private_mutex *mut = (struct _private_mutex *) *hMutex;

  if (mut == NULL)
    return -1;
  DeleteCriticalSection (&mut->h);
  ortp_free (mut);
	return 0;
}

#else

int WIN_mutex_init(ortp_mutex_t *mutex, void *attr)
{
	*mutex=CreateMutex(NULL, FALSE, NULL);
	return 0;
}

int WIN_mutex_lock(ortp_mutex_t * hMutex)
{
	WaitForSingleObject(*hMutex, INFINITE); /* == WAIT_TIMEOUT; */
	return 0;
}

int WIN_mutex_unlock(ortp_mutex_t * hMutex)
{
	ReleaseMutex(*hMutex);
	return 0;
}

int WIN_mutex_destroy(ortp_mutex_t * hMutex)
{
	CloseHandle(*hMutex);
	return 0;
}
#endif

/* Windows without Pthreads for Win32 */
#if (_MSC_VER >= 1700) && !defined(_USING_V110_SDK71_)
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#define HAVE_CPP11_THREAD
#define HAVE_RUNASYNC
#elif defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PC_APP)
#define HAVE_CPP11_THREAD
#endif
#endif

#ifdef HAVE_RUNASYNC

using namespace std;
using namespace Platform;
using namespace Windows::Foundation;
using namespace Windows::System::Threading;

typedef struct thread_param{
	void * (*func)(void *);
	void * arg;
	HANDLE threadHandle;
}thread_param_t;

static unsigned long WINAPI thread_starter(void *data){
	thread_param_t *params=(thread_param_t*)data;
	void *ret=params->func(params->arg);
	return (DWORD)ret;
}

static void StartThread(LPTHREAD_START_ROUTINE lpStartAddress, LPVOID lpParameter)
{
	thread_param_t *params=(thread_param_t*)lpParameter;
	auto workItemHandler = ref new WorkItemHandler([=](IAsyncAction^)
	{
		try
		{
			lpStartAddress(lpParameter);
		}
		catch (...) { }

		/* Signal that the thread has completed. */
		SetEvent(params->threadHandle);
		ortp_free(params);
	}, CallbackContext::Any);

	ThreadPool::RunAsync(workItemHandler, WorkItemPriority::High, WorkItemOptions::TimeSliced);
}

int WIN_thread_create(ortp_thread_t *th, void *attr, void * (*func)(void *), void *data)
{
	thread_param_t *params=ortp_new(thread_param_t,1);
	params->func=func;
	params->arg=data;
	params->threadHandle = CreateEventEx(nullptr, nullptr, CREATE_EVENT_MANUAL_RESET, EVENT_ALL_ACCESS);
	*th=params->threadHandle;
	StartThread(thread_starter, params);
	return 0;
}

int WIN_thread_join(ortp_thread_t thread_h, void **unused)
{
	if (thread_h!=NULL)
	{
		WaitForSingleObjectEx(thread_h, INFINITE, FALSE);
		CloseHandle(thread_h);
	}
	return 0;
}

int WIN_thread_set_name(ortp_thread_t thread, const char *name) {
  return 0;
}

#elif defined(HAVE_CPP11_THREAD)
#include <iostream>
#include <thread>

typedef struct thread_param{
	void * (*func)(void *);
	void * arg;
}thread_param_t;

static unsigned WINAPI thread_starter(void *data){
	thread_param_t *params=(thread_param_t*)data;
	void *ret=params->func(params->arg);
	ortp_free(data);
	return (DWORD)ret;
}

int WIN_thread_create(ortp_thread_t *th, void *attr, void * (*func)(void *), void *data)
{
    thread_param_t *params=ortp_new(thread_param_t,1);
    params->func=func;
    params->arg=data;
	*th=new std::thread(thread_starter, params);
	return 0;
}

int WIN_thread_join(ortp_thread_t thread_h, void **unused)
{
	std::thread *th=(std::thread*)thread_h;
	if (th!=NULL)
	{
		th->join();
		delete th;
	}
	return 0;
}

int WIN_thread_set_name(ortp_thread_t thread, const char *name) {
  return 0;
}

#else


typedef struct thread_param{
	void * (*func)(void *);
	void * arg;
}thread_param_t;

static unsigned WINAPI thread_starter(void *data){
	thread_param_t *params=(thread_param_t*)data;
	void *ret=params->func(params->arg);
	ortp_free(data);
	return (DWORD)ret;
}

#if defined _WIN32_WCE
#    define _beginthreadex	CreateThread
#    define	_endthreadex	ExitThread
#endif

int WIN_thread_create(ortp_thread_t *th, void *attr, void * (*func)(void *), void *data)
{
    thread_param_t *params=ortp_new(thread_param_t,1);
    params->func=func;
    params->arg=data;
	*th=(HANDLE)_beginthreadex( NULL, 0, thread_starter, params, 0, NULL);
	return 0;
}

int WIN_thread_join(ortp_thread_t thread_h, void **unused)
{
	if (thread_h!=NULL)
	{
		WaitForSingleObject(thread_h, INFINITE);
		CloseHandle(thread_h);
	}
	return 0;
}

HINSTANCE m_kernelInst = NULL;

typedef HRESULT(WINAPI *KERNEL32SETTHREADDESCRIPTION)(HANDLE IcmpHandle, PCWSTR lpThreadDescription);

KERNEL32SETTHREADDESCRIPTION pSetThreadDescription = NULL;

int WIN_thread_set_name(ortp_thread_t thread_h, const char *name) {
  WCHAR wUnicode[16];
  static int loaded_once=0;
  if (pSetThreadDescription == NULL) {
    if (loaded_once != 0)
      return 0;
    loaded_once++;
    m_kernelInst = LoadLibrary(L"kernel32.dll");
    if (m_kernelInst != NULL)
    {
      pSetThreadDescription = (KERNEL32SETTHREADDESCRIPTION) GetProcAddress(m_kernelInst, "SetThreadDescription");
    }
  }
  if (pSetThreadDescription != NULL) {
    MultiByteToWideChar(CP_UTF8, 0, name, -1, wUnicode, 16);
    pSetThreadDescription(thread_h, wUnicode);
  }
  return 0;
}

#endif

#if (_MSC_VER >= 1700) && !defined(_USING_V110_SDK71_)
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#define DISABLE_CONDITIONVARIABLE
#elif defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PC_APP)
#define DISABLE_CONDITIONVARIABLE
#endif
#endif

#ifdef DISABLE_CONDITIONVARIABLE

int WIN_cond_init(ortp_cond_t *cond, void *attr)
{
	return 0;
}

int WIN_cond_wait(ortp_cond_t* hCond, ortp_mutex_t * hMutex)
{
	return 0;
}

int WIN_cond_signal(ortp_cond_t * hCond)
{
	return 0;
}

int WIN_cond_broadcast(ortp_cond_t * hCond)
{
	return 0;
}

int WIN_cond_destroy(ortp_cond_t * hCond)
{
	return 0;
}

#elif defined(USE_WIN32_CONDITION_VARIABLE)

int WIN_cond_init(ortp_cond_t *cond, void *attr)
{
  /* HANDLE and CONDITION_VARIABLE are the same size */
  InitializeConditionVariable((PCONDITION_VARIABLE)cond);
  return 0;
}

int WIN_cond_wait(ortp_cond_t* hCond, ortp_mutex_t * hMutex)
{
  struct _private_mutex *mut = (struct _private_mutex *) *hMutex;
  SleepConditionVariableSRW((PCONDITION_VARIABLE)hCond, &mut->h, INFINITE, 0);
  return 0;
}

int WIN_cond_signal(ortp_cond_t * hCond)
{
  WakeConditionVariable((PCONDITION_VARIABLE)hCond);
  return 0;
}

int WIN_cond_broadcast(ortp_cond_t * hCond)
{
  WIN_cond_signal(hCond);
  return 0;
}

int WIN_cond_destroy(ortp_cond_t * hCond)
{
  return 0;
}

#else

int WIN_cond_init(ortp_cond_t *cond, void *attr)
{
	*cond=CreateEvent(NULL, FALSE, FALSE, NULL);
	return 0;
}

int WIN_cond_wait(ortp_cond_t* hCond, ortp_mutex_t * hMutex)
{
  WIN_mutex_unlock(hMutex);
  WaitForSingleObject(*hCond, INFINITE);
	WIN_mutex_lock(hMutex);
  return 0;
}

int WIN_cond_signal(ortp_cond_t * hCond)
{
  SetEvent(*hCond);
  return 0;
}

int WIN_cond_broadcast(ortp_cond_t * hCond)
{
  WIN_cond_signal(hCond);
	return 0;
}

int WIN_cond_destroy(ortp_cond_t * hCond)
{
  CloseHandle(*hCond);
  return 0;
}

#endif

#if defined(_WIN32_WCE)
#include <time.h>

int
gettimeofday (struct timeval *tv, void *tz)
{
  DWORD timemillis = GetTickCount();
  tv->tv_sec  = timemillis/1000;
  tv->tv_usec = (timemillis - (tv->tv_sec*1000)) * 1000;
  return 0;
}

#else

int gettimeofday (struct timeval *tv, void* tz)
{
	union
	{
		__int64 ns100; /*time since 1 Jan 1601 in 100ns units */
		FILETIME fileTime;
	} now;

	GetSystemTimeAsFileTime (&now.fileTime);
	tv->tv_usec = (long) ((now.ns100 / 10LL) % 1000000LL);
	tv->tv_sec = (long) ((now.ns100 - 116444736000000000LL) / 10000000LL);
	return (0);
}

#endif

const char *getWinSocketError(int error)
{
	static char buf[80];

	switch (error)
	{
		case WSANOTINITIALISED: return "Windows sockets not initialized : call WSAStartup";
		case WSAEADDRINUSE:		return "Local Address already in use";
		case WSAEADDRNOTAVAIL:	return "The specified address is not a valid address for this machine";
		case WSAEINVAL:			return "The socket is already bound to an address.";
		case WSAENOBUFS:		return "Not enough buffers available, too many connections.";
		case WSAENOTSOCK:		return "The descriptor is not a socket.";
		case WSAECONNRESET:		return "Connection reset by peer";
		case WSAEWOULDBLOCK:	return "Operation would block";
		case WSAEINPROGRESS:	return "Operation now in progress";
		case WSAENOTCONN:		return "Socket is not connected";
		case WSAETIMEDOUT:		return "Connection timed out";
		case WSAECONNREFUSED:	return "Connection refused";

		default :
			sprintf(buf, "Error code : %d", error);
			return buf;
		break;
	}

	return buf;
}

#ifdef _WORKAROUND_MINGW32_BUGS
char * WSAAPI gai_strerror(int errnum){
	 return (char*)getWinSocketError(errnum);
}
#endif

#endif

#if 0 //ndef WIN32

#include <sys/socket.h>
#include <netdb.h>
#include <sys/un.h>
#include <sys/stat.h>

static char *make_pipe_name(const char *name){
	return ortp_strdup_printf("/tmp/%s",name);
}

/* portable named pipes */
ortp_socket_t ortp_server_pipe_create(const char *name){
	struct sockaddr_un sa;
	char *pipename=make_pipe_name(name);
	ortp_socket_t sock;
	sock=socket(AF_UNIX,SOCK_STREAM,0);
	sa.sun_family=AF_UNIX;
	strncpy(sa.sun_path,pipename,sizeof(sa.sun_path)-1);
	unlink(pipename);/*in case we didn't finished properly previous time */
	ortp_free(pipename);
	fchmod(sock,S_IRUSR|S_IWUSR);
	if (bind(sock,(struct sockaddr*)&sa,sizeof(sa))!=0){
		//ortp_error("Failed to bind command unix socket: %s",ortp_strerror(errno));
		return -1;
	}
	listen(sock,1);
	return sock;
}

ortp_socket_t ortp_server_pipe_accept_client(ortp_socket_t server){
	struct sockaddr_un su;
	socklen_t ssize=sizeof(su);
	ortp_socket_t client_sock=accept(server,(struct sockaddr*)&su,&ssize);
	return client_sock;
}

int ortp_server_pipe_close_client(ortp_socket_t client){
	return close(client);
}

int ortp_server_pipe_close(ortp_socket_t spipe){
	return close(spipe);
}

ortp_socket_t ortp_client_pipe_connect(const char *name){
	struct sockaddr_un sa;
	char *pipename=make_pipe_name(name);
	ortp_socket_t sock=socket(AF_UNIX,SOCK_STREAM,0);
	sa.sun_family=AF_UNIX;
	strncpy(sa.sun_path,pipename,sizeof(sa.sun_path)-1);
	ortp_free(pipename);
	if (connect(sock,(struct sockaddr*)&sa,sizeof(sa))!=0){
		close(sock);
		return -1;
	}
	return sock;
}

int ortp_pipe_read(ortp_socket_t p, uint8_t *buf, int len){
	return (int)read(p,buf,len);
}

int ortp_pipe_write(ortp_socket_t p, const uint8_t *buf, int len){
	return (int)write(p,buf,len);
}

int ortp_client_pipe_close(ortp_socket_t sock){
	return close(sock);
}


#elif 0 //defined(WIN32) && !defined(_WIN32_WCE)

static char *make_pipe_name(const char *name){
	return ortp_strdup_printf("\\\\.\\pipe\\%s",name);
}

static HANDLE event=NULL;

/* portable named pipes */
ortp_pipe_t ortp_server_pipe_create(const char *name){
	ortp_pipe_t h;
	TCHAR pipename[256];
	_snwprintf(pipename, 256, TEXT("\\\\.\\pipe\\%s"), name);
	//char *pipename=make_pipe_name(name);
	h=CreateNamedPipe(pipename,PIPE_ACCESS_DUPLEX|FILE_FLAG_OVERLAPPED,PIPE_TYPE_MESSAGE|PIPE_WAIT,1,
						32768,32768,0,NULL);
	ortp_free(pipename);
	if (h==INVALID_HANDLE_VALUE){
		//ortp_error("Fail to create named pipe %s",pipename);
	}
	if (event==NULL) event=CreateEvent(NULL,TRUE,FALSE,NULL);
	return h;
}


/*this function is a bit complex because we need to wakeup someday
even if nobody connects to the pipe.
ortp_server_pipe_close() makes this function to exit.
*/
ortp_pipe_t ortp_server_pipe_accept_client(ortp_pipe_t server){
	OVERLAPPED ol;
	DWORD undef;
	HANDLE handles[2];
	memset(&ol,0,sizeof(ol));
	ol.hEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
	ConnectNamedPipe(server,&ol);
	handles[0]=ol.hEvent;
	handles[1]=event;
	WaitForMultipleObjects(2,handles,FALSE,INFINITE);
	if (GetOverlappedResult(server,&ol,&undef,FALSE)){
		CloseHandle(ol.hEvent);
		return server;
	}
	CloseHandle(ol.hEvent);
	return INVALID_HANDLE_VALUE;
}

int ortp_server_pipe_close_client(ortp_pipe_t server){
	return DisconnectNamedPipe(server)==TRUE ? 0 : -1;
}

int ortp_server_pipe_close(ortp_pipe_t spipe){
	SetEvent(event);
	//CancelIoEx(spipe,NULL); /*vista only*/
	return CloseHandle(spipe);
}

ortp_pipe_t ortp_client_pipe_connect(const char *name){
	ortp_pipe_t hpipe;
	TCHAR pipename[256];
	_snwprintf(pipename, 256, TEXT("\\\\.\\pipe\\%s"), name);
	//char *pipename=make_pipe_name(name);
	hpipe = CreateFile(
         pipename,   // pipe name
         GENERIC_READ |  // read and write access
         GENERIC_WRITE,
         0,              // no sharing
         NULL,           // default security attributes
         OPEN_EXISTING,  // opens existing pipe
         0,              // default attributes
         NULL);          // no template file
	ortp_free(pipename);
	return hpipe;
}

int ortp_pipe_read(ortp_pipe_t p, uint8_t *buf, int len){
	DWORD ret=0;
	if (ReadFile(p,buf,len,&ret,NULL))
		return ret;
	/*ortp_error("Could not read from pipe: %s",ortp_strerror(GetLastError()));*/
	return -1;
}

int ortp_pipe_write(ortp_pipe_t p, const uint8_t *buf, int len){
	DWORD ret=0;
	if (WriteFile(p,buf,len,&ret,NULL))
		return ret;
	/*ortp_error("Could not write to pipe: %s",ortp_strerror(GetLastError()));*/
	return -1;
}


int ortp_client_pipe_close(ortp_pipe_t sock){
	return CloseHandle(sock);
}


#endif

#if defined(_WIN32_WCE)

char * ortp_strerror(DWORD value) {
	static TCHAR msgBuf[256];
	FormatMessage(
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			value,
			0, // Default language
			(LPTSTR) &msgBuf,
			0,
			NULL
	);
	return (const char *)msgBuf;
}

#else

char * ortp_strerror(int value) {
	return strerror(value);
}

#endif

#if defined(WIN32) || defined(_WIN32_WCE)

#if !defined(_WIN32_WCE)
#include <errno.h>
#endif

#define	ORTP_IN6ADDRSZ	16
#define	ORTP_INT16SZ		 2

#ifndef	AF_INET6
#define	AF_INET6	AF_MAX+1        /* just to let this compile */
#endif

/*
 * WARNING: Don't even consider trying to compile this on a system where
 * sizeof(int) < 4.  sizeof(int) > 4 is fine; all the world's not a VAX.
 */

static const char *_inet_ntop4 (const u_char * src, char *dst, size_t size);
static const char *_inet_ntop6 (const u_char * src, char *dst, size_t size);

/* char *
 * inet_ntop(af, src, dst, size)
 *	convert a network format address to presentation format.
 * return:
 *	pointer to presentation format address (`dst'), or NULL (see errno).
 * author:
 *	Paul Vixie, 1996.
 */
const char *
ortp_inet_ntop (af, src, dst, size)
     int af;
     const void *src;
     char *dst;
     size_t size;
{
  switch (af) {
  case AF_INET:
    return (_inet_ntop4 (src, dst, size));
  case AF_INET6:
    return (_inet_ntop6 (src, dst, size));
  default:
#ifndef _WIN32_WCE
    errno = WSAEAFNOSUPPORT;
#endif
    return (NULL);
  }
  /* NOTREACHED */
}

/* const char *
 * inet_ntop4(src, dst, size)
 *	format an IPv4 address, more or less like inet_ntoa()
 * return:
 *	`dst' (as a const)
 * notes:
 *	(1) uses no statics
 *	(2) takes a u_char* not an in_addr as input
 * author:
 *	Paul Vixie, 1996.
 */
static const char *
_inet_ntop4 (src, dst, size)
     const u_char *src;
     char *dst;
     size_t size;
{
  static const char fmt[] = "%u.%u.%u.%u";
  char tmp[sizeof "255.255.255.255"];

  sprintf (tmp, fmt, src[0], src[1], src[2], src[3]);
  if ((size_t) strlen (tmp) >= size) {
#ifndef _WIN32_WCE
    errno = ENOSPC;
#endif
    return (NULL);
  }
  strcpy (dst, tmp);
  return (dst);
}

/* const char *
 * inet_ntop6(src, dst, size)
 *	convert IPv6 binary address into presentation (printable) format
 * author:
 *	Paul Vixie, 1996.
 */
static const char *
_inet_ntop6 (src, dst, size)
     const u_char *src;
     char *dst;
     size_t size;
{
  /*
   * Note that int32_t and int16_t need only be "at least" large enough
   * to contain a value of the specified size.  On some systems, like
   * Crays, there is no such thing as an integer variable with 16 bits.
   * Keep this in mind if you think this function should have been coded
   * to use pointer overlays.  All the world's not a VAX.
   */
  char tmp[sizeof "ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255"], *tp;
  struct {
    int base, len;
  } best, cur;
  uint32_t words[ORTP_IN6ADDRSZ / ORTP_INT16SZ];
  int i;

  /*
   * Preprocess:
   *      Copy the input (bytewise) array into a wordwise array.
   *      Find the longest run of 0x00's in src[] for :: shorthanding.
   */
  memset (words, 0, sizeof words);
  for (i = 0; i < ORTP_IN6ADDRSZ; i++)
    words[i / 2] |= (src[i] << ((1 - (i % 2)) << 3));
  best.base = -1;
  best.len = 0;
  cur.base = -1;
  cur.len = 0;
  for (i = 0; i < (ORTP_IN6ADDRSZ / ORTP_INT16SZ); i++) {
    if (words[i] == 0) {
      if (cur.base == -1)
        cur.base = i, cur.len = 1;
      else
        cur.len++;
    }
    else {
      if (cur.base != -1) {
        if (best.base == -1 || cur.len > best.len)
          best = cur;
        cur.base = -1;
      }
    }
  }
  if (cur.base != -1) {
    if (best.base == -1 || cur.len > best.len)
      best = cur;
  }
  if (best.base != -1 && best.len < 2)
    best.base = -1;

  /*
   * Format the result.
   */
  tp = tmp;
  for (i = 0; i < (ORTP_IN6ADDRSZ / ORTP_INT16SZ); i++) {
    /* Are we inside the best run of 0x00's? */
    if (best.base != -1 && i >= best.base && i < (best.base + best.len)) {
      if (i == best.base)
        *tp++ = ':';
      continue;
    }
    /* Are we following an initial run of 0x00s or any real hex? */
    if (i != 0)
      *tp++ = ':';
    /* Is this address an encapsulated IPv4? */
    if (i == 6 && best.base == 0 && (best.len == 6 || (best.len == 5 && words[5] == 0xffff))) {
      if (!_inet_ntop4 (src + 12, tp, sizeof tmp - (tp - tmp)))
        return (NULL);
      tp += strlen (tp);
      break;
    }
    sprintf (tp, "%x", words[i]);
    tp += strlen (tp);
  }
  /* Was it a trailing run of 0x00's? */
  if (best.base != -1 && (best.base + best.len) == (ORTP_IN6ADDRSZ / ORTP_INT16SZ))
    *tp++ = ':';
  *tp++ = '\0';

  /*
   * Check for overflow, copy, and we're done.
   */
  if ((size_t) (tp - tmp) > size) {
#if !defined (_WIN32_WCE)
    errno = ENOSPC;
#endif
    return (NULL);
  }
  strcpy (dst, tmp);
  return (dst);
}

#endif
