/*
The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/***************************************************************************
*            rtcp.c
*
*  Wed Dec  1 11:45:30 2004
*  Copyright  2004  Simon Morlat
*  Email simon dot morlat at linphone dot org
****************************************************************************/
#include "ortp/ortp.h"
#include "ortp/rtpsession.h"
#include "ortp/rtcp.h"
#include "utils.h"
#include "rtpsession_priv.h"
#include <math.h>

#define rtcp_bye_set_ssrc(b,pos,ssrc)	(b)->ssrc[pos]=htonl(ssrc)
#define rtcp_bye_get_ssrc(b,pos)		ntohl((b)->ssrc[pos])

mblk_t *rtcp_create_simple_bye_packet(uint32_t ssrc, const char *reason);

static void rtcp_common_header_init(rtcp_common_header_t *ch, RtpSession *s,int type, int rc, int bytes_len){
  rtcp_common_header_set_version(ch,2);
  rtcp_common_header_set_padbit(ch,0);
  rtcp_common_header_set_packet_type(ch,type);
  rtcp_common_header_set_rc(ch,rc);	/* as we don't yet support multi source receiving */
  rtcp_common_header_set_length(ch,(bytes_len/4)-1);
}

static mblk_t *sdes_chunk_new(uint32_t ssrc){
  mblk_t *m=allocb(RTCP_SDES_CHUNK_DEFAULT_SIZE,0);
  sdes_chunk_t *sc=(sdes_chunk_t*)m->b_rptr;
  sc->csrc=htonl(ssrc);
  m->b_wptr+=sizeof(sc->csrc);
  return m;
}


static mblk_t * sdes_chunk_append_item(mblk_t *m, rtcp_sdes_type_t sdes_type, const char *content)
{	
  if ( content && content[0]!='\0')
  {
    sdes_item_t si;
    si.item_type=sdes_type;
    si.len=(uint8_t) MIN(strlen(content),RTCP_SDES_MAX_STRING_SIZE);
    m=appendb(m,(char*)&si,RTCP_SDES_ITEM_HEADER_SIZE,FALSE);
    m=appendb(m,content,si.len,FALSE);
  }
  return m;
}

static void sdes_chunk_set_ssrc(mblk_t *m, uint32_t ssrc){
  sdes_chunk_t *sc=(sdes_chunk_t*)m->b_rptr;
  sc->csrc=htonl(ssrc);
}

#define sdes_chunk_get_ssrc(m) ntohl(((sdes_chunk_t*)((m)->b_rptr))->csrc)

static mblk_t * sdes_chunk_pad(mblk_t *m){
  return appendb(m,"",1,TRUE);
}

/**
* Set session's SDES item for automatic sending of RTCP compound packets.
* If some items are not specified, use NULL.
**/
void rtp_session_set_source_description(RtpSession *session, 
                                        const char *cname, const char *name, const char *email, const char *phone, 
                                        const char *loc, const char *tool, const char *note){
                                          mblk_t *chunk = sdes_chunk_new(session->snd.ssrc);
                                          mblk_t *m=chunk;
                                          const char *_cname=cname;
                                          if (_cname==NULL)
                                          {
                                            _cname="unknown@unknown.com";
                                          }
                                          chunk=sdes_chunk_append_item(chunk, RTCP_SDES_CNAME, _cname);
                                          chunk=sdes_chunk_append_item(chunk, RTCP_SDES_NAME, name);
                                          chunk=sdes_chunk_append_item(chunk, RTCP_SDES_EMAIL, email);
                                          chunk=sdes_chunk_append_item(chunk, RTCP_SDES_PHONE, phone);
                                          chunk=sdes_chunk_append_item(chunk, RTCP_SDES_LOC, loc);
                                          chunk=sdes_chunk_append_item(chunk, RTCP_SDES_TOOL, tool);
                                          chunk=sdes_chunk_append_item(chunk, RTCP_SDES_NOTE, note);
                                          chunk=sdes_chunk_pad(chunk);
                                          if (session->sd!=NULL) freemsg(session->sd);
                                          session->sd=m;
}

void
  rtp_session_add_contributing_source(RtpSession *session, uint32_t csrc, 
  const char *cname, const char *name, const char *email, const char *phone, 
  const char *loc, const char *tool, const char *note)
{
  mblk_t *chunk = sdes_chunk_new(csrc);
  mblk_t *m=chunk;
  char *_cname=(char*)cname;
  if (_cname==NULL)
  {
    _cname="toto";
  }
  chunk=sdes_chunk_append_item(chunk, RTCP_SDES_CNAME, cname);
  chunk=sdes_chunk_append_item(chunk, RTCP_SDES_NAME, name);
  chunk=sdes_chunk_append_item(chunk, RTCP_SDES_EMAIL, email);
  chunk=sdes_chunk_append_item(chunk, RTCP_SDES_PHONE, phone);
  chunk=sdes_chunk_append_item(chunk, RTCP_SDES_LOC, loc);
  chunk=sdes_chunk_append_item(chunk, RTCP_SDES_TOOL, tool);
  chunk=sdes_chunk_append_item(chunk, RTCP_SDES_NOTE, note);
  chunk=sdes_chunk_pad(chunk);
  putq(&session->contributing_sources,m);
}



mblk_t* rtp_session_create_rtcp_sdes_packet(RtpSession *session)
{
  mblk_t *mp=allocb(sizeof(rtcp_common_header_t),0);
  rtcp_common_header_t *rtcp;
  mblk_t *tmp,*m=mp;
  queue_t *q;
  int rc=0;
  rtcp = (rtcp_common_header_t*)mp->b_wptr;
  mp->b_wptr+=sizeof(rtcp_common_header_t);

  /* concatenate all sdes chunks */
  sdes_chunk_set_ssrc(session->sd,session->snd.ssrc);
  m=concatb(m,dupmsg(session->sd));
  rc++;

  q=&session->contributing_sources;
  for (tmp=qbegin(q); !qend(q,tmp); tmp=qnext(q,tmp)){
    m=concatb(m,dupmsg(tmp));
    rc++;
  }
  rtcp_common_header_init(rtcp,session,RTCP_SDES,rc,msgdsize(mp));
  return mp;
}

mblk_t* rtp_session_create_rtcp_xr_vm_packet(RtpSession *session)
{
  mblk_t *mp=allocb(sizeof(rtcp_xr_t)+sizeof(rtcp_xr_vm_t),0);
  rtcp_xr_t *rtcp_xr;
  rtcp_xr_vm_t *_rtcp_xr_vm;
  rtcp_xr = (rtcp_xr_t*)mp->b_wptr;
  mp->b_wptr+=sizeof(rtcp_xr_t)+sizeof(rtcp_xr_vm_t);
  memset(mp->b_rptr, 0, msgdsize(mp));
  
  rtcp_xr->ssrc=htonl(session->snd.ssrc);
  rtcp_xr->rb[0].BT=RTCP_XR_VM;
  rtcp_xr->rb[0].type_specific=0;
  rtcp_xr->rb[0].block_len=htons(sizeof(rtcp_xr_vm_t)/4);
  
  _rtcp_xr_vm = (rtcp_xr_vm_t *)(rtcp_xr+1);
  _rtcp_xr_vm->ssrc=htonl(session->rcv.ssrc);

  /* This value is calculated by dividing the total number of packets lost (after
   the effects of applying any error protection such as FEC) by
   the total number of packets expected, multiplying the result of
   the division by 256, limiting the maximum value to 255 (to
   avoid overflow), and taking the integer part */
  _rtcp_xr_vm->loss_rate = (uint8_t)((session->rtp.stats.cum_packet_loss*256.0)/(session->rtp.stats.cum_packet_loss+session->rtp.stats.packet_recv));

  _rtcp_xr_vm->discard_rate = (uint8_t)((session->rtp.stats.discarded*256.0)/(session->rtp.stats.cum_packet_loss+session->rtp.stats.packet_recv));

  _rtcp_xr_vm->signal_level=0x7f;
  _rtcp_xr_vm->noise_level=0x7f;
  _rtcp_xr_vm->RERL=0x7f;

  _rtcp_xr_vm->R_factor=0x7f;
  _rtcp_xr_vm->ext_R_factor=0x7f;
  _rtcp_xr_vm->MOS_CQ=0x7f;
  _rtcp_xr_vm->MOS_LQ=0x7f;
  
  /* PLC: Standard (11) / enhanced (10) */
  /* JBA: Adaptive (11) / non-adaptive (10) */
  _rtcp_xr_vm->RX_config=0x8<<4;
  _rtcp_xr_vm->RX_config+=0x3<<4;

  _rtcp_xr_vm->JB_nominal=htons(session->rtp.stats.average_jitter); /* plus time in msconf + time in soundcard */
  _rtcp_xr_vm->JB_maximum=htons(session->rtp.jittctl.max_jitter); /* first packet that would be discarded */
  _rtcp_xr_vm->JB_abs_max=htons(session->rtp.jittctl.max_jitter);
  
  rtcp_common_header_init(&rtcp_xr->ch,session,RTCP_XR,0,msgdsize(mp));
  return mp;
}

mblk_t* rtp_session_create_rtcp_xr_rr_packet(RtpSession *session) {
  mblk_t *mp = allocb(sizeof(rtcp_xr_t) + sizeof(rtcp_xr_rrt_t), 0);
  rtcp_xr_t *rtcp_xr;
  rtcp_xr_rrt_t *_rtcp_xr_rrt;
  rtcp_xr = (rtcp_xr_t*)mp->b_wptr;
  mp->b_wptr += sizeof(rtcp_xr_t) + sizeof(rtcp_xr_rrt_t);
  memset(mp->b_rptr, 0, msgdsize(mp));

  rtcp_xr->ssrc = htonl(session->snd.ssrc);
  rtcp_xr->rb[0].BT = RTCP_XR_RR;
  rtcp_xr->rb[0].type_specific = 0;
  rtcp_xr->rb[0].block_len = htons(sizeof(rtcp_xr_rrt_t) / 4);

  _rtcp_xr_rrt = (rtcp_xr_rrt_t *)(rtcp_xr + 1);

  {
    struct timeval tv;
    uint32_t tmp;
    gettimeofday(&tv, NULL);
    _rtcp_xr_rrt->ntp_timestamp_msw = htonl(tv.tv_sec + 0x83AA7E80); /* 0x83AA7E80 is the number of seconds from 1900 to 1970 */
#if defined(_WIN32_WCE)
    tmp = (uint32_t)((double)tv.tv_usec*(double)(((uint64_t)1) << 32)*1.0e-6);
#else
    tmp = (uint32_t)((double)tv.tv_usec*(double)(1LL << 32)*1.0e-6);
    tmp = (uint32_t)((double)tv.tv_usec*(double)(((uint64_t)1) << 32)*1.0e-6);
#endif
    _rtcp_xr_rrt->ntp_timestamp_lsw = htonl(tmp);
  }

  rtcp_common_header_init(&rtcp_xr->ch,session,RTCP_XR,0,msgdsize(mp));
  return mp;
}

mblk_t *rtp_session_create_rtcp_xr_dlrr_packet(RtpSession *session) {
  mblk_t *mp = allocb(sizeof(rtcp_xr_t) + sizeof(rtcp_xr_dlrr_t), 0);
  rtcp_xr_t *rtcp_xr;
  rtcp_xr_dlrr_t *_rtcp_xr_dlrr;
  rtcp_xr = (rtcp_xr_t*)mp->b_wptr;
  mp->b_wptr += sizeof(rtcp_xr_t) + sizeof(rtcp_xr_dlrr_t);
  memset(mp->b_rptr, 0, msgdsize(mp));

  rtcp_xr->ssrc = htonl(session->snd.ssrc);
  rtcp_xr->rb[0].BT = RTCP_XR_DLRR;
  rtcp_xr->rb[0].type_specific = 0;
  rtcp_xr->rb[0].block_len = htons(sizeof(rtcp_xr_dlrr_t) / 4);

  _rtcp_xr_dlrr = (rtcp_xr_dlrr_t *)(rtcp_xr + 1);
  _rtcp_xr_dlrr->ssrc = htonl(session->rcv.ssrc);

  _rtcp_xr_dlrr->last_rr = htonl(session->rtcp_xr_stats.last_rcvr_rtt_ts);

  _rtcp_xr_dlrr->delaysince_last_rr = 0;
  if (session->rtcp_xr_stats.last_rcvr_rtt_time.tv_sec != 0) {
    struct timeval tv;
    double delay;
    uint32_t dlrr;
    gettimeofday(&tv, NULL);
    delay = ((tv.tv_sec - session->rtcp_xr_stats.last_rcvr_rtt_time.tv_sec)
      + ((tv.tv_usec - session->rtcp_xr_stats.last_rcvr_rtt_time.tv_usec) * 1e-6)) * 65536;
    dlrr = (uint32_t)delay;

    _rtcp_xr_dlrr->delaysince_last_rr = htonl(dlrr);
  }

  rtcp_common_header_init(&rtcp_xr->ch, session, RTCP_XR, 0, msgdsize(mp));
  return mp;
}

mblk_t *rtcp_create_simple_bye_packet(uint32_t ssrc, const char *reason)
{	
  int packet_size;
  int strsize = 0;
  int strpadding = 0;
  mblk_t *mp;
  rtcp_bye_t *rtcp;

  packet_size	= RTCP_BYE_HEADER_SIZE;
  if (reason!=NULL) {
    strsize=(int)MIN(strlen(reason),RTCP_BYE_REASON_MAX_STRING_SIZE);
    if (strsize > 0) {
      strpadding = 3 - (strsize % 4);
      packet_size += 1 + strsize + strpadding;
    }
  }
  mp	= allocb(packet_size, 0);

  rtcp = (rtcp_bye_t*)mp->b_rptr;
  rtcp_common_header_init(&rtcp->ch,NULL,RTCP_BYE,1,packet_size);
  rtcp->ssrc[0] = htonl(ssrc);
  mp->b_wptr += RTCP_BYE_HEADER_SIZE;
  /* append the reason if any*/
  if (reason!=NULL) {
    const char pad[] = {0, 0, 0};
    unsigned char strsize_octet = (unsigned char)strsize;

    appendb(mp, (const char*)&strsize_octet, 1, FALSE);
    appendb(mp, reason,strsize, FALSE);
    appendb(mp, pad,strpadding, FALSE);
  }
  return mp;
}

void rtp_session_remove_contributing_sources(RtpSession *session, uint32_t ssrc)
{
  queue_t *q=&session->contributing_sources;
  mblk_t *tmp;
  for (tmp=qbegin(q); !qend(q,tmp); tmp=qnext(q,tmp)){
    uint32_t csrc=sdes_chunk_get_ssrc(tmp);
    if (csrc==ssrc) {
      remq(q,tmp);
      break;
    }
  }
  tmp=rtcp_create_simple_bye_packet(ssrc, NULL);
  rtp_session_rtcp_send(session,tmp);
}

static void sender_info_init(sender_info_t *info, RtpSession *session){
  struct timeval tv;
  uint32_t tmp;
  gettimeofday(&tv,NULL);
  info->ntp_timestamp_msw=htonl(tv.tv_sec + 0x83AA7E80); /* 0x83AA7E80 is the number of seconds from 1900 to 1970 */
#if defined(_WIN32_WCE)
  tmp=(uint32_t)((double)tv.tv_usec*(double)(((uint64_t)1)<<32)*1.0e-6);
#else
  tmp=(uint32_t)((double)tv.tv_usec*(double)(1LL<<32)*1.0e-6);
#endif
  info->ntp_timestamp_lsw=htonl(tmp);
  info->rtp_timestamp=htonl(session->rtp.snd_last_ts);
  info->senders_packet_count=(uint32_t) htonl((u_long) session->rtp.stats.packet_sent);
  info->senders_octet_count=(uint32_t) htonl((u_long) session->rtp.sent_payload_bytes);
  session->rtp.last_rtcp_packet_count=(uint32_t)session->rtp.stats.packet_sent;
}



static void report_block_init(report_block_t *b, RtpSession *session){
  int packet_loss=0;
  uint8_t loss_fraction=0;
  RtpStream *stream=&session->rtp;
  uint32_t delay_snc_last_sr=0;
  uint32_t fl_cnpl;

  /* compute the statistics */
  /*printf("hwrcv_extseq.one=%u, hwrcv_seq_at_last_SR=%u hwrcv_since_last_SR=%u\n",
  stream->hwrcv_extseq.one,
  stream->hwrcv_seq_at_last_SR,
  stream->hwrcv_since_last_SR
  );*/
  if (stream->hwrcv_since_last_SR!=0){
    packet_loss=(stream->hwrcv_extseq - stream->hwrcv_seq_at_last_SR) - stream->hwrcv_since_last_SR;
    if (packet_loss<0)
      packet_loss=0;
    stream->stats.cum_packet_loss+=packet_loss;
    if (stream->hwrcv_extseq - stream->hwrcv_seq_at_last_SR>0) {
      stream->extended_stats.cur_packet_loss_fraction = (float)packet_loss/(float)(stream->hwrcv_extseq - stream->hwrcv_seq_at_last_SR);
      loss_fraction=(int)(256.0*stream->extended_stats.cur_packet_loss_fraction);
    } else {
      stream->extended_stats.cur_packet_loss_fraction = 0;
      loss_fraction=0;
    }
  }
  /* reset them */
  stream->hwrcv_since_last_SR=0;
  stream->hwrcv_seq_at_last_SR=stream->hwrcv_extseq;

  if (stream->last_rcv_SR_time.tv_sec!=0){
    struct timeval now;
    float delay;
    gettimeofday(&now,NULL);
    delay=(float) ((now.tv_sec-stream->last_rcv_SR_time.tv_sec)*1e6 ) + (now.tv_usec-stream->last_rcv_SR_time.tv_usec);
    delay=(float) (delay*65536*1e-6);
    delay_snc_last_sr=(uint32_t) delay;
  }

  b->ssrc=htonl(session->rcv.ssrc);
  fl_cnpl=((loss_fraction&0xFF)<<24) | (stream->stats.cum_packet_loss & 0xFFFFFF);
  b->fl_cnpl=htonl(fl_cnpl);
  b->interarrival_jitter=htonl((uint32_t) stream->jittctl.inter_jitter);
  b->ext_high_seq_num_rec=htonl(stream->hwrcv_extseq);
  b->lsr=htonl(stream->last_rcv_SR_ts);
  b->delay_snc_last_sr=htonl(delay_snc_last_sr);
}



static int rtcp_sr_init(RtpSession *session, uint8_t *buf, int size){
  rtcp_sr_t *sr=(rtcp_sr_t*)buf;
  int rr=(session->rtp.stats.packet_recv>0);
  int sr_size=sizeof(rtcp_sr_t)-sizeof(report_block_t)+(rr*sizeof(report_block_t));
  if (size<sr_size) return 0;
  rtcp_common_header_init(&sr->ch,session,RTCP_SR,rr,sr_size);
  sr->ssrc=htonl(session->snd.ssrc);
  sender_info_init(&sr->si,session);
  /*only include a report block if packets were received*/
  if (rr)
    report_block_init(&sr->rb[0],session);
  return sr_size;
}

static int rtcp_rr_init(RtpSession *session, uint8_t *buf, int size){
  rtcp_rr_t *rr=(rtcp_rr_t*)buf;
  if (size<sizeof(rtcp_rr_t)) return 0;
  rtcp_common_header_init(&rr->ch,session,RTCP_RR,1,sizeof(rtcp_rr_t));
  rr->ssrc=htonl(session->snd.ssrc);
  report_block_init(&rr->rb[0],session);
  return sizeof(rtcp_rr_t);
}

static int rtcp_app_init(RtpSession *session, uint8_t *buf, uint8_t subtype, const char *name, int size){
  rtcp_app_t *app=(rtcp_app_t*)buf;
  if (size<sizeof(rtcp_app_t)) return 0;
  rtcp_common_header_init(&app->ch,session,RTCP_APP,subtype,size);
  app->ssrc=htonl(session->snd.ssrc);
  memset(app->name,0,4);
  strncpy(app->name,name,4);
  return sizeof(rtcp_app_t);
}

static mblk_t * make_rr(RtpSession *session){
  mblk_t *cm=NULL;
  mblk_t *sdes=NULL;
  mblk_t *rtcpxr_rr = NULL;
  mblk_t *rtcpxr_vm = NULL;

  cm=allocb(sizeof(rtcp_sr_t),0);
  cm->b_wptr+=rtcp_rr_init(session,cm->b_wptr,sizeof(rtcp_rr_t));
  /* make a SDES packet */
  if (session->sd!=NULL)
    sdes=rtp_session_create_rtcp_sdes_packet(session);
  rtcpxr_rr = rtp_session_create_rtcp_xr_rr_packet(session);
  if (session->rtcpxr_voip_metrics)
    rtcpxr_vm=rtp_session_create_rtcp_xr_vm_packet(session);
  /* link them */
  cm->b_cont=sdes;
  if (rtcpxr_rr != NULL)
    concatb(cm, rtcpxr_rr);
  if (rtcpxr_vm!=NULL)
    concatb(cm, rtcpxr_vm);
  return cm;
}


static mblk_t * make_sr(RtpSession *session){
  mblk_t *cm=NULL;
  mblk_t *sdes=NULL;
  mblk_t *rtcpxr_dlrr = NULL;
  mblk_t *rtcpxr_vm = NULL;

  cm=allocb(sizeof(rtcp_sr_t),0);
  cm->b_wptr+=rtcp_sr_init(session,cm->b_wptr,sizeof(rtcp_sr_t));
  /* make a SDES packet */
  if (session->sd!=NULL)
    sdes=rtp_session_create_rtcp_sdes_packet(session);
  if (session->rtcpxr_send_one_dlrr) {
    session->rtcpxr_send_one_dlrr=FALSE;
    rtcpxr_dlrr=rtp_session_create_rtcp_xr_dlrr_packet(session);
  }
  if (session->rtcpxr_voip_metrics)
    rtcpxr_vm=rtp_session_create_rtcp_xr_vm_packet(session);
    
  /* link them */
  cm->b_cont=sdes;
  if (rtcpxr_dlrr != NULL)
    concatb(cm, rtcpxr_dlrr);
  if (rtcpxr_vm != NULL)
    concatb(cm, rtcpxr_vm);
  return cm;
}

void rtp_session_rtcp_process_send(RtpSession *session){
  RtpStream *st=&session->rtp;
  mblk_t *m;
  if (st->last_rtcp_report_snt_r==0) {
    /* initialize to initial value */
    st->last_rtcp_report_snt_r=st->rcv_last_app_ts;
  }
  if (st->last_rtcp_report_snt_s==0) {
    /* initialize to initial value */
    st->last_rtcp_report_snt_s=st->snd_last_ts;
  }
  if (st->rcv_last_app_ts - st->last_rtcp_report_snt_r > st->rtcp_report_snt_interval 
    || st->snd_last_ts - st->last_rtcp_report_snt_s > st->rtcp_report_snt_interval){
      st->last_rtcp_report_snt_r=st->rcv_last_app_ts;
      st->last_rtcp_report_snt_s=st->snd_last_ts;
      m=make_sr(session);
      /* send the compound packet */
      rtp_session_rtcp_send(session,m);
      ortp_debug("RTCP compound message sent. (SR)");
  }
}

/*
 rfc5348:
 NOTE: I receive loss and rtt, I calculate my Tfrc=max bandwidth to send
 and i limit the value to twice the received rate.
 -> thus, i adjust the send bitrate.
 
 NOTE2: If I have RTCP feedback enable, I calculate local loss & RTT, this
 give the Tfrc bandwidth I can receive. I adjust the value to not be above
 twice the send bitrate, and I send a TMMBR or GOOG-REMB.
 
 o   The receiver measures the loss event rate and feeds this
 information back to the sender.
 
 o   The sender also uses these feedback messages to measure the
 round-trip time (RTT).
 
 o   The loss event rate and RTT are then fed into TFRC's throughput
 equation, and the resulting sending rate is limited to at most
 twice the receive rate to give the allowed transmit rate X.
 
 o   The sender then adjusts its transmit rate to match the allowed
 transmit rate X.
 */
// Calculate the rate that TCP-Friendly Rate Control (TFRC) would apply.
// The formula in RFC 3448, Section 3.1, is used.
uint32_t CalcTfrcBps(double rtt, double loss) {
  int b;
  double t_RTO;
  double s;
  double X;
  if (rtt == 0 || loss == 0) {
    // Input variables out of range.
    return 0;
  }
  //double R = (double)(rtt) / 1000;  // RTT in seconds.
  b = 1;  // Number of packets acknowledged by a single TCP acknowledgement:
  // recommended = 1.
  t_RTO = 4.0 * rtt;  // TCP retransmission timeout value in seconds
  // recommended = 4*R.
  //double p = (double)(loss) / 255;  // Packet loss rate in [0, 1).
  s = (double)(1000); //kAvgPacketSizeBytes=1000;
  // Calculate send rate in bytes/second.
  X =  s / (rtt * sqrt(2 * b * loss / 3) +
       (t_RTO * (3 * sqrt(3 * b * loss / 8) * loss * (1 + 32 * loss * loss))));
  // Convert to bits/second.
  return ((uint32_t)(X * 8));
}

static int _rtp_session_update_zigzagtable(RtpSession *session, uint32_t current_total_bitrate) {
  RtpStream *st=&session->rtp;
  if (st->cur_bitrate_table.bitrates_table[0]==0) {
    int idx;
    int cur_bitrate=0;
    for (idx=0;idx<10;idx++) {
      cur_bitrate=cur_bitrate+64000;
      st->cur_bitrate_table.bitrates_table[idx]=cur_bitrate;
      st->cur_bitrate_table.bitrates_tables_loss[idx]=0;
      ortp_message("bitrates_table[%i]=%i", idx, cur_bitrate);
    }
    /* index=9 -> 640000 */
    for (idx=10;idx<15;idx++) {
      cur_bitrate=cur_bitrate+128000;
      st->cur_bitrate_table.bitrates_table[idx]=cur_bitrate;
      st->cur_bitrate_table.bitrates_tables_loss[idx]=0;
      ortp_message("bitrates_table[%i]=%i", idx, cur_bitrate);
    }
    /* index=14 -> ? */
    for (idx=15;idx<20;idx++) {
      cur_bitrate=cur_bitrate+256000;
      st->cur_bitrate_table.bitrates_table[idx]=cur_bitrate;
      st->cur_bitrate_table.bitrates_tables_loss[idx]=0;
      ortp_message("bitrates_table[%i]=%i", idx, cur_bitrate);
    }
    /* index=19 -> ? */
  }
  if (session->rtp.stats.packet_recv>0 && current_total_bitrate>0) {
    int new_bitrate=st->cur_bitrate_table.previous_tmmbr_value;
    if (st->cur_bitrate_table.previous_tmmbr_value==0) {
      /* preset the value */
      int idx;
      st->cur_bitrate_table.previous_tmmbr_value=st->cur_bitrate_table.bitrates_table[0];
      for (idx=0;idx<20;idx++) {
        if (st->cur_bitrate_table.bitrates_table[idx]>current_total_bitrate)
          break;
        st->cur_bitrate_table.previous_tmmbr_value=st->cur_bitrate_table.bitrates_table[idx];
      }
      new_bitrate=st->cur_bitrate_table.previous_tmmbr_value;
    }
    if (session->rtp.extended_stats.rtt==0 && session->rtp.extended_stats.cur_packet_loss_fraction==0) {
      ortp_message("zigzag: rtt=0 // loss=0 (not increasing. wait for RTT>0)");
      return new_bitrate;
    }

    if (session->rtp.extended_stats.cur_packet_loss_fraction>=0.10) {
      int idx;
      int idx_cur_bitrate=0;

      if (current_total_bitrate>st->cur_bitrate_table.previous_tmmbr_value)
        current_total_bitrate=st->cur_bitrate_table.previous_tmmbr_value;

      for (idx=0;idx<20;idx++) {
        if (st->cur_bitrate_table.bitrates_table[idx]<current_total_bitrate/2)
          continue; /* skip */
        st->cur_bitrate_table.bitrates_tables_loss[idx]=6.0f+idx_cur_bitrate++;
      }
    } else if (session->rtp.extended_stats.cur_packet_loss_fraction>=0.02) {
      int idx;
      int idx_cur_bitrate=0;

      if (current_total_bitrate>st->cur_bitrate_table.previous_tmmbr_value)
        current_total_bitrate=st->cur_bitrate_table.previous_tmmbr_value;

      for (idx=0;idx<20;idx++) {
        if (st->cur_bitrate_table.bitrates_table[idx]<current_total_bitrate)
          continue; /* skip */
        st->cur_bitrate_table.bitrates_tables_loss[idx]=2.0f+idx_cur_bitrate++;
      }
    } else if (session->rtp.extended_stats.cur_packet_loss_fraction==0) {
      int idx;
      int cmp_bitrate=current_total_bitrate;
      int doitonce=0;
      if (current_total_bitrate<st->cur_bitrate_table.previous_tmmbr_value)
        cmp_bitrate=st->cur_bitrate_table.previous_tmmbr_value;

      for (idx=0;idx<20;idx++) {
        if (st->cur_bitrate_table.bitrates_table[idx]>cmp_bitrate) {
          if (doitonce==0 && st->cur_bitrate_table.bitrates_tables_loss[idx]==0) {
            doitonce=1;
            st->cur_bitrate_table.bitrates_tables_loss[idx]=1.0f; /* new level allowed for first try */
          }
          st->cur_bitrate_table.bitrates_tables_loss[idx]*=0.9f;
          continue; /* skip */
        }
        if (doitonce==0 && st->cur_bitrate_table.bitrates_tables_loss[idx]==0)
          doitonce=1;  /* new level allowed for first try */
        st->cur_bitrate_table.bitrates_tables_loss[idx]=1.0f;
      }
    }

    /* select a new bitrate? */
    if (session->rtp.extended_stats.cur_packet_loss_fraction==0) {
      int idx;

      if (current_total_bitrate+512000<=session->rtp.cur_bitrate_table.previous_tmmbr_value)
      {
        ortp_message("zigzag: current received bitrate is very low compared to previous tmmbr value (not increasing)");
        return 0;
      }

      for (idx=0;idx<19;idx++) {
        if (st->cur_bitrate_table.bitrates_table[idx]==session->rtp.cur_bitrate_table.previous_tmmbr_value) {
          if (st->cur_bitrate_table.bitrates_tables_loss[idx+1]<=1.0f) {
            new_bitrate = st->cur_bitrate_table.bitrates_table[idx+1];
            ortp_message("zigzag: selecting higher bitrate for ccm bitrate[idx=%i]=%i", idx+1, new_bitrate);
          }
          else {
            ortp_message("zigzag: NOT selecting higher bitrate for ccm bitrate[idx=%i]=%i (proba>1 -> %f)", idx+1, new_bitrate, st->cur_bitrate_table.bitrates_tables_loss[idx+1]);
          }
          break;
        }
      }
    } else if (session->rtp.extended_stats.cur_packet_loss_fraction>=0.10) {
      int idx;

      new_bitrate = st->cur_bitrate_table.bitrates_table[0];
      for (idx=0;idx<19;idx++) {
          if (st->cur_bitrate_table.bitrates_tables_loss[idx]<=1 && st->cur_bitrate_table.bitrates_table[idx]<session->rtp.cur_bitrate_table.previous_tmmbr_value/2) {
            new_bitrate = st->cur_bitrate_table.bitrates_table[idx];
            continue;
          }
          break;
      }
      ortp_message("zigzag: selecting lower bitrate for ccm bitrate[idx=%i]=%i", idx, new_bitrate);

    } else if (session->rtp.extended_stats.cur_packet_loss_fraction>=0.02) {
      int idx;

      new_bitrate = st->cur_bitrate_table.bitrates_table[0];
      for (idx=0;idx<19;idx++) {
          if (st->cur_bitrate_table.bitrates_tables_loss[idx]<=1 && st->cur_bitrate_table.bitrates_table[idx]<session->rtp.cur_bitrate_table.previous_tmmbr_value*4/5) {
            new_bitrate = st->cur_bitrate_table.bitrates_table[idx];
            continue;
          }
          break;
      }
      ortp_message("zigzag: selecting lower bitrate for ccm bitrate[idx=%i]=%i", idx, new_bitrate);
    }


    if (session->rtp.extended_stats.rtt>0.5f) {
      if (st->cur_bitrate_table.max_bitrate==0 || st->cur_bitrate_table.max_bitrate>512000) {
        st->cur_bitrate_table.max_bitrate=512000;
        ortp_message("zigzag: rtt=%f is too high. set max bitrate limit=%i", session->rtp.extended_stats.rtt, st->cur_bitrate_table.max_bitrate);
      }
    }

    if (session->rtp.extended_stats.rtt>3.0f) {
      new_bitrate=12000;
      ortp_message("zigzag: rtt=%f is too high. reverting to bitrate=%i", session->rtp.extended_stats.rtt, new_bitrate);
    } else if (session->rtp.extended_stats.rtt>1.0f) {
      new_bitrate=30000;
      ortp_message("zigzag: rtt=%f is too high. reverting to bitrate=%i", session->rtp.extended_stats.rtt, new_bitrate);
    }

    if (st->cur_bitrate_table.max_bitrate!=0 && new_bitrate>st->cur_bitrate_table.max_bitrate)
      new_bitrate = st->cur_bitrate_table.max_bitrate;

    if (st->cur_bitrate_table.previous_tmmbr_value==30000 || st->cur_bitrate_table.previous_tmmbr_value==12000) {
      return new_bitrate; /* even if it's the same value, we resend the request -in case it was lost- */
    }

    if (new_bitrate!=st->cur_bitrate_table.previous_tmmbr_value)
      return new_bitrate;
  }
  return 0;
}

void rtp_session_rtcp_process_recv(RtpSession *session){
  RtpStream *st=&session->rtp;
  mblk_t *m=NULL;
  if (st->last_rtcp_report_snt_r==0) {
    /* initialize to initial value */
    st->last_rtcp_report_snt_r=st->rcv_last_app_ts;
  }
  if (st->last_rtcp_report_snt_s==0) {
    /* initialize to initial value */
    st->last_rtcp_report_snt_s=st->snd_last_ts;
  }
  if (st->rcv_last_app_ts - st->last_rtcp_report_snt_r > st->rtcp_report_snt_interval 
    || st->snd_last_ts - st->last_rtcp_report_snt_s > st->rtcp_report_snt_interval){
      st->last_rtcp_report_snt_r=st->rcv_last_app_ts;
      st->last_rtcp_report_snt_s=st->snd_last_ts;
      if (session->rtp.last_rtcp_packet_count<session->rtp.stats.packet_sent){
        m=make_sr(session);
        session->rtp.last_rtcp_packet_count=(uint32_t)session->rtp.stats.packet_sent;
      }else if (session->rtp.stats.packet_recv>0){
        /*don't send RR when no packet are received yet*/
        m=make_rr(session);
      }
      if (m!=NULL){
        if (session->rtcpfeedback_ccm_tmmbr==TRUE || session->rtcpfeedback_goog_remb==TRUE) {
          uint32_t current_total_bitrate = (uint32_t)rtp_session_get_recv_rtppayload_bandwidth(session, NULL);
#if 0
          /*
           This value seems to provide a good hint, but testing is showing that:
           we can do better than tfrc for high rtt (ex: rtt=0,323s loss=1,3% -> 230Kbps // but 512Kbps is running perfectly)
           under 5% loss and high rtt, the value looks low? (ex: rtt=0.647736 loss=0.067265   ->  trfc=34Kbit/s)
           we can't do as much as tfrc for low rtt (ex: rtt=0,010s loss=1%   -> 8996Kbps // would be too much for CPU?)
           */
          uint32_t trfc = CalcTfrcBps(session->rtp.extended_stats.rtt, session->rtp.extended_stats.cur_packet_loss_fraction);
          ortp_message("CalcTfrcBps rtt=%f loss=%f   ->  trfc=%iKbit/s",
                     session->rtp.extended_stats.rtt, session->rtp.extended_stats.cur_packet_loss_fraction, trfc/1000);
#endif
          if (session->rtp.stats.packet_recv>0 && current_total_bitrate>0) {
            int new_bitrate = _rtp_session_update_zigzagtable(session, current_total_bitrate);

            if (new_bitrate!=0) {
              if (session->rtcpfeedback_goog_remb==TRUE)
                rtp_session_send_rtcp_fb_GOOGREMB(session, session->rcv.ssrc, new_bitrate);
              else
                rtp_session_send_rtcp_fb_TMMB(session, 3, session->rcv.ssrc, new_bitrate, session->rtp.cur_bitrate_table.recv_overhead_realtime_bytes);
              session->rtp.cur_bitrate_table.previous_tmmbr_value = new_bitrate;
            }
          }
        }

        /* send the compound packet */
        rtp_session_rtcp_send(session,m);
        ortp_debug("RTCP compound message sent. (RR or SR)");
      }
  }
}

void rtp_session_send_rtcp_APP(RtpSession *session, uint8_t subtype, const char *name, const uint8_t *data, int datalen){
  mblk_t *h=allocb(sizeof(rtcp_app_t),0);
  mblk_t *d;
  h->b_wptr+=rtcp_app_init(session,h->b_wptr,subtype,name,datalen+sizeof(rtcp_app_t));
  d=esballoc((uint8_t*)data,datalen,0,NULL);
  d->b_wptr+=datalen;
  h->b_cont=d;
  rtp_session_rtcp_send(session,h);
}

/**
* Sends a RTCP bye packet.
*@param session RtpSession
*@param reason the reason phrase.
**/
int
  rtp_session_bye(RtpSession *session, const char *reason)
{
  mblk_t *cm;
  mblk_t *sdes = NULL;
  mblk_t *bye = NULL;
  int ret;

  /* Make a BYE packet (will be on the end of the compund packet). */
  bye = rtcp_create_simple_bye_packet(session->snd.ssrc, reason);

  /* SR or RR is determined by the fact whether stream was sent*/
  if (session->rtp.stats.packet_sent>0)
  {
    cm = allocb(sizeof(rtcp_sr_t), 0);
    cm->b_wptr += rtcp_sr_init(session,cm->b_wptr, sizeof(rtcp_sr_t));
    /* make a SDES packet */
    sdes = rtp_session_create_rtcp_sdes_packet(session);
    /* link them */
    concatb(concatb(cm, sdes), bye);
  } else if (session->rtp.stats.packet_recv>0){
    /* make a RR packet */
    cm = allocb(sizeof(rtcp_rr_t), 0);
    cm->b_wptr += rtcp_rr_init(session, cm->b_wptr, sizeof(rtcp_rr_t));
    /* link them */
    cm->b_cont = bye;
  }else cm=bye;

  /* Send compound packet. */
  ret = rtp_session_rtcp_send(session, cm);
  ortp_message("RTCP BYE message sent.");
  return ret;
}

static int rtcp_fb_init(RtpSession *session, uint8_t *buf, rtcp_type_t type, uint8_t subtype, int size){
  rtcp_fb_t *fb=(rtcp_fb_t*)buf;
  if (size<sizeof(rtcp_fb_t)) return 0;
  rtcp_common_header_init(&fb->ch,session,type,subtype,size);
  fb->ssrc_sender=htonl(session->snd.ssrc);
  fb->ssrc_local=htonl(session->rcv.ssrc);
  return sizeof(rtcp_fb_t);
}

void rtp_session_send_rtcp_fb_NACK(RtpSession *session, const uint16_t cseq, const uint16_t bits) {
  struct fci_nack {
    uint16_t pid:16;
    uint16_t BLP:16;
  };
  mblk_t *h;
  struct fci_nack *nack;

  h=allocb(sizeof(rtcp_fb_t)+sizeof(struct fci_nack),0);
  h->b_wptr+=rtcp_fb_init(session,h->b_wptr,RTCP_RTPFB,1,sizeof(rtcp_fb_t)+sizeof(struct fci_nack));

  nack = (struct fci_nack*)h->b_wptr;
  nack->pid = htons(cseq);
  nack->BLP = htons(bits);

  h->b_wptr+=sizeof(struct fci_nack);

  rtp_session_rtcp_send(session,h);
}

void rtp_session_send_rtcp_fb_RPSI(RtpSession *session, const uint8_t *bitstring, int bitstring_len){
  struct rtcp_rpsi {
    uint8_t PB;
    uint8_t PT;
    uint16_t bitstring[1];
  };

  mblk_t *h;
  struct rtcp_rpsi *rpsi;
  int pos;
  int extrabytes = (bitstring_len / 8)-2;
  if (bitstring_len % 8>0)
    extrabytes++;
  if (extrabytes < 0) extrabytes = 0;
  /* we currently only support bitstring_len=8 or bitstring_len=16, so size of FCI is always 32 bits */
  h=allocb(sizeof(rtcp_fb_t)+sizeof(struct rtcp_rpsi)+extrabytes,0);
  h->b_wptr+=rtcp_fb_init(session,h->b_wptr,RTCP_PSFB,3,sizeof(rtcp_fb_t)+sizeof(struct rtcp_rpsi)+extrabytes);

  rpsi = (struct rtcp_rpsi*)h->b_wptr;
  memset(rpsi, 0, sizeof(struct rtcp_rpsi)+extrabytes);
  if (extrabytes==0)
    rpsi->PB = 16-bitstring_len;
  else
    rpsi->PB = (bitstring_len-16) % 32;

  rpsi->PT = session->rcv.pt & 0x7F;

  memcpy(&rpsi->bitstring, bitstring, bitstring_len / 8);
  for (pos = 0; pos < (bitstring_len % 8); pos++) {
    rpsi->bitstring[extrabytes - 1] |= (bitstring[extrabytes - 1] & (1 << (7 - pos)));
  }

  h->b_wptr+=sizeof(struct rtcp_rpsi)+extrabytes;

  rtp_session_rtcp_send(session,h);
}


void rtp_session_send_rtcp_fb_PLI(RtpSession *session) {
  mblk_t *h;

  h=allocb(sizeof(rtcp_fb_t),0);
  h->b_wptr+=rtcp_fb_init(session,h->b_wptr,RTCP_PSFB,1,sizeof(rtcp_fb_t));

  rtp_session_rtcp_send(session,h);
}

void rtp_session_send_rtcp_fb_TMMB(RtpSession *session, const uint8_t subtype, const uint32_t ssrc, const uint32_t bitrate, const uint32_t overhead) {
  struct fci_tmmb {
    uint32_t ssrc_affected; /* The SSRC value of the media sender that is requested to obey the new maximum bit rate */
    uint32_t other32_bits; /* 6bit: MxTBR Exp  - 17bit: MxTBR Mantissa   - 9bit: overhead */
  };

  mblk_t *h;
  struct fci_tmmb *tmmbr;
  rtcp_fb_t *fb;
  h=allocb(sizeof(rtcp_fb_t)+sizeof(struct fci_tmmb),0);
  fb=(rtcp_fb_t*)h->b_wptr;
  h->b_wptr+=rtcp_fb_init(session,h->b_wptr,RTCP_RTPFB,subtype,sizeof(rtcp_fb_t)+sizeof(struct fci_tmmb));
  fb->ssrc_local=htonl(0);

  tmmbr = (struct fci_tmmb *)h->b_wptr;
  tmmbr->ssrc_affected=htonl(ssrc); /* the owner */

  {
    uint32_t nbit = 1<<17;
    uint32_t MxTBR_Exp = 0;
    uint32_t MxTBR_Mantissa = bitrate;
    while (MxTBR_Mantissa >= nbit) {
      MxTBR_Mantissa >>= 1;
      ++MxTBR_Exp;
    }
    tmmbr->other32_bits = htonl(overhead | (MxTBR_Mantissa << 9) | (MxTBR_Exp << 26));
  }

  h->b_wptr+=sizeof(struct fci_tmmb);

  rtp_session_rtcp_send(session,h);
}

void rtp_session_send_rtcp_fb_GOOGREMB(RtpSession *session, const uint32_t ssrc, const uint32_t bitrate){
  struct fci_googremb {
    char name[4];
    uint32_t other32_bits; /* 8bit: Num of SSRC  - 6bit: MxTBR Exp   - 18bit: MxTBR Mantissa */
    uint32_t ssrc_affected;
  };

  mblk_t *h;
  struct fci_googremb *googremb;
  rtcp_fb_t *fb;
  h=allocb(sizeof(rtcp_fb_t)+sizeof(struct fci_googremb),0);
  fb=(rtcp_fb_t*)h->b_wptr;
  h->b_wptr+=rtcp_fb_init(session,h->b_wptr,RTCP_PSFB,15,sizeof(rtcp_fb_t)+sizeof(struct fci_googremb));
  fb->ssrc_local=htonl(0);

  googremb = (struct fci_googremb *)h->b_wptr;
  strncpy(googremb->name,"REMB",4);
  googremb->ssrc_affected=htonl(ssrc); /* the owner */

  {
    uint32_t nbit = 1<<18;
    uint32_t MxTBR_Exp = 0;
    uint32_t MxTBR_Mantissa = bitrate;
    while (MxTBR_Mantissa >= nbit) {
      MxTBR_Mantissa >>= 1;
      ++MxTBR_Exp;
    }
    googremb->other32_bits = htonl(MxTBR_Mantissa | (MxTBR_Exp << 18) | (1 << 24));
  }

  h->b_wptr+=sizeof(struct fci_googremb); /* only one SSRC included in */

  rtp_session_rtcp_send(session,h);
}
