/*
The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "ortp/ortp.h"
#include "utils.h"
#include "rtpsession_priv.h"


/*in case of coumpound packet, set read pointer of m to the beginning of the next RTCP
packet */
bool_t rtcp_next_packet(mblk_t *m){
  const rtcp_common_header_t *ch=rtcp_get_common_header(m);
  if (ch){
    int nextlen=sizeof(rtcp_common_header_t)+
      (rtcp_common_header_get_length(ch)*4);
    if (m->b_rptr+nextlen<m->b_wptr){
      m->b_rptr+=nextlen;
      return TRUE;
    }
  }
  return FALSE;
}

void rtcp_rewind(mblk_t *m){
  m->b_rptr=m->b_datap->db_base;
}

/* get common header; this function will also check the sanity of the packet*/
const rtcp_common_header_t * rtcp_get_common_header(const mblk_t *m){
  int size=msgdsize(m);
  rtcp_common_header_t *ch;
  if (m->b_cont!=NULL){
    ortp_fatal("RTCP parser does not work on fragmented mblk_t. Use msgpullup() before to re-assemble the packet.");
    return NULL;
  }
  if (size<sizeof(rtcp_common_header_t)){
    ortp_warning("Bad RTCP packet, too short.");
    return NULL;
  }
  ch=(rtcp_common_header_t*)m->b_rptr;
  return ch;
}

bool_t rtcp_is_SR(const mblk_t *m){
  const rtcp_common_header_t *ch=rtcp_get_common_header(m);
  if (ch!=NULL && rtcp_common_header_get_packet_type(ch)==RTCP_SR){
    if (msgdsize(m)<sizeof(rtcp_sr_t)){
      ortp_warning("Too short RTCP SR packet.");
      return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

/*Sender Report accessors */
uint32_t rtcp_SR_get_ssrc(const mblk_t *m){
  rtcp_sr_t *sr=(rtcp_sr_t*)m->b_rptr;
  return ntohl(sr->ssrc);
}

const sender_info_t * rtcp_SR_get_sender_info(const mblk_t *m){
  rtcp_sr_t *sr=(rtcp_sr_t*)m->b_rptr;
  return &sr->si;
}

const report_block_t * rtcp_SR_get_report_block(const mblk_t *m, int idx){
  rtcp_sr_t *sr=(rtcp_sr_t*)m->b_rptr;
  report_block_t *rb=&sr->rb[idx];
  int size=sizeof(rtcp_common_header_t)+(4*rtcp_common_header_get_length(&sr->ch));
  if ( ( (uint8_t*)rb)+sizeof(report_block_t) <= m->b_rptr + size ) {
    return rb;
  }else{
    if (idx<rtcp_common_header_get_rc(&sr->ch)){
      ortp_warning("RTCP packet should include a report_block_t at pos %i but has no space for it.",idx);
    }
  }
  return NULL;
}

/*Receiver report accessors*/
bool_t rtcp_is_RR(const mblk_t *m){
  const rtcp_common_header_t *ch=rtcp_get_common_header(m);
  if (ch!=NULL && rtcp_common_header_get_packet_type(ch)==RTCP_RR){
    if (msgdsize(m)<sizeof(rtcp_rr_t)){
      ortp_warning("Too short RTCP RR packet.");
      return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

uint32_t rtcp_RR_get_ssrc(const mblk_t *m){
  rtcp_rr_t *rr=(rtcp_rr_t*)m->b_rptr;
  return ntohl(rr->ssrc);
}

const report_block_t * rtcp_RR_get_report_block(const mblk_t *m,int idx){
  rtcp_rr_t *rr=(rtcp_rr_t*)m->b_rptr;
  report_block_t *rb=&rr->rb[idx];
  int size=sizeof(rtcp_common_header_t)+(4*rtcp_common_header_get_length(&rr->ch));
  if ( ( (uint8_t*)rb)+sizeof(report_block_t) <= (m->b_rptr + size ) ){
    return rb;
  }else{
    if (idx<rtcp_common_header_get_rc(&rr->ch)){
      ortp_warning("RTCP packet should include a report_block_t at pos %i but has no space for it.",idx);
    }
  }
  return NULL;
}

/*SDES accessors */
bool_t rtcp_is_SDES(const mblk_t *m){
  const rtcp_common_header_t *ch=rtcp_get_common_header(m);
  if (ch && rtcp_common_header_get_packet_type(ch)==RTCP_SDES){
    if (msgdsize(m)<sizeof(rtcp_common_header_t)+
      (4*rtcp_common_header_get_length(ch))){
        ortp_warning("Too short RTCP SDES packet.");
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

void rtcp_sdes_parse(const mblk_t *m, SdesItemFoundCallback cb, void *user_data){
  uint8_t *rptr=(uint8_t*)m->b_rptr+sizeof(rtcp_common_header_t);
  const rtcp_common_header_t *ch=(rtcp_common_header_t*)m->b_rptr;
  uint8_t *end=rptr+sizeof(rtcp_common_header_t)+
    (4*rtcp_common_header_get_length(ch));
  uint32_t ssrc=0;
  int nchunk=0;
  bool_t chunk_start=TRUE;

  if (end>(uint8_t*)m->b_wptr) end=(uint8_t*)m->b_wptr;

  while(rptr<end){
    if (chunk_start){
      if (rptr+4<=end){
        ssrc=ntohl(*(uint32_t*)rptr);
        rptr+=4;
      }else{
        ortp_warning("incorrect chunk start in RTCP SDES");
        break;
      }
      chunk_start=FALSE;
    }else{
      if (rptr+2<=end){
        uint8_t type=rptr[0];
        uint8_t len=rptr[1];

        if (type==RTCP_SDES_END){
          /* pad to next 32bit boundary*/
          //CHECKME
          rptr=(uint8_t*)(((intptr_t)rptr+4) & ~0x3);
          nchunk++;
          if (nchunk<rtcp_common_header_get_rc(ch)){
            chunk_start=TRUE;
            continue;
          }else break;
        }
        rptr+=2;
        if (rptr+len<=end){
          cb(user_data,ssrc,type,(char*)rptr,len);
          rptr+=len;
        }else{
          ortp_warning("bad item length in RTCP SDES");
          break;
        }
      }else{
        /*end of packet */
        break;
      }
    }
  }
}

/*BYE accessors */
bool_t rtcp_is_BYE(const mblk_t *m){
  const rtcp_common_header_t *ch=rtcp_get_common_header(m);
  if (ch && rtcp_common_header_get_packet_type(ch)==RTCP_BYE){
    if (msgdsize(m)<sizeof(rtcp_common_header_t)+
      rtcp_common_header_get_length(ch)){
        ortp_warning("Too short RTCP BYE packet.");
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

bool_t rtcp_BYE_get_ssrc(const mblk_t *m, int idx, uint32_t *ssrc){
  rtcp_bye_t *bye=(rtcp_bye_t*)m->b_rptr;
  int rc=rtcp_common_header_get_rc(&bye->ch);
  int len=rtcp_common_header_get_length(&bye->ch);
  if (idx<rc){
    if ((uint8_t*)&bye->ssrc[idx]<=(m->b_rptr
      +sizeof(rtcp_common_header_t)+len-4)) {
        *ssrc=ntohl(bye->ssrc[idx]);
        return TRUE;
    }else{
      ortp_warning("RTCP BYE should contain %i ssrc, but there is not enough room for it.");
    }
  }
  return FALSE;
}

bool_t rtcp_BYE_get_reason(const mblk_t *m, const char **reason, int *reason_len){
  rtcp_bye_t *bye=(rtcp_bye_t*)m->b_rptr;
  int rc=rtcp_common_header_get_rc(&bye->ch);
  int len=rtcp_common_header_get_length(&bye->ch);
  uint8_t *rptr=(uint8_t*)m->b_rptr+sizeof(rtcp_common_header_t)+rc*4;
  uint8_t *end=(uint8_t*)(m->b_rptr+sizeof(rtcp_common_header_t)+len);
  if (rptr<end){
    uint8_t content_len=rptr[0];
    if (rptr+1+content_len<=end){
      *reason=(char*)rptr+1;
      *reason_len=content_len;
      return TRUE;
    }else{
      ortp_warning("RTCP BYE has not enough space for reason phrase.");
      return FALSE;
    }
  }
  return FALSE;
}

/*APP accessors */
bool_t rtcp_is_APP(const mblk_t *m){
  const rtcp_common_header_t *ch=rtcp_get_common_header(m);
  if (ch!=NULL && rtcp_common_header_get_packet_type(ch)==RTCP_APP){
    if (msgdsize(m)<sizeof(rtcp_common_header_t)+
      rtcp_common_header_get_length(ch)){
        ortp_warning("Too short RTCP APP packet.");
        return FALSE;
    }
    if (sizeof(rtcp_common_header_t)+rtcp_common_header_get_length(ch)
      < sizeof(rtcp_app_t)){
        ortp_warning("Bad RTCP APP packet.");
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

int rtcp_APP_get_subtype(const mblk_t *m){
  rtcp_app_t *app=(rtcp_app_t*)m->b_rptr;
  return rtcp_common_header_get_rc(&app->ch);
}

uint32_t rtcp_APP_get_ssrc(const mblk_t *m){
  rtcp_app_t *app=(rtcp_app_t*)m->b_rptr;
  return ntohl(app->ssrc);
}
/* name argument is supposed to be at least 4 characters (note: no '\0' written)*/
void rtcp_APP_get_name(const mblk_t *m, char *name){
  rtcp_app_t *app=(rtcp_app_t*)m->b_rptr;
  memcpy(name,app->name,4);
}
/* retrieve the data. when returning, data points directly into the mblk_t */
void rtcp_APP_get_data(const mblk_t *m, uint8_t **data, int *len){
  rtcp_app_t *app=(rtcp_app_t*)m->b_rptr;
  int datalen=sizeof(rtcp_common_header_t)+rtcp_common_header_get_length(&app->ch)-8;
  if (datalen>0){
    *data=(uint8_t*)m->b_rptr+sizeof(rtcp_app_t);
    *len=datalen;
  }else{
    *len=0;
    *data=NULL;
  }
}

bool_t rtcp_is_RTPFB(const mblk_t *m){
  const rtcp_common_header_t *ch=rtcp_get_common_header(m);
  if (ch!=NULL && rtcp_common_header_get_packet_type(ch)==RTCP_RTPFB){
    if (msgdsize(m)<sizeof(rtcp_fb_t)){
      ortp_warning("Too short RTCP RTPFB packet.");
      return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

bool_t rtcp_is_PSFB(const mblk_t *m){
  const rtcp_common_header_t *ch=rtcp_get_common_header(m);
  if (ch!=NULL && rtcp_common_header_get_packet_type(ch)==RTCP_PSFB){
    if (msgdsize(m)<sizeof(rtcp_fb_t)){
      ortp_warning("Too short RTCP PSFB packet.");
      return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}


static uint32_t ortp_get_compact_ntp(struct timeval *rcv_time_tv) {
  uint32_t rcv_time_msw;
  uint32_t rcv_time_lsw;
  rcv_time_msw = (uint32_t)rcv_time_tv->tv_sec + 0x83AA7E80;
#if defined(_WIN32_WCE)
  rcv_time_lsw = (uint32_t) ((double)rcv_time_tv->tv_usec*(double)(((uint64_t)1)<<32)*1.0e-6);
#else
  rcv_time_lsw = (uint32_t) ((double)rcv_time_tv->tv_usec*(double)(1LL<<32)*1.0e-6);
#endif
  return (rcv_time_msw<<16) | (rcv_time_lsw >> 16);
}

#include <math.h>

static double convert_to_most (const double rf)
{
  /*
  * R<0    :   MOS = 1
  * R>100  :   MOS = 4.5
  * 0<R<100:   MOS = 1+0.035*R + R*(R-60)*(100-R)*7/1000000;
  * |R-factor       MOS
  * |90<R<100  4.34  -   4.5
  * |80<R<90   4.03  -   4.34
  * |70<R<80   3.06  -   4.03
  * |60<R<70   3.10  -   3.60
  * |50<R<60   2.58  -   3.10
  */
  if (rf <0) {
    return 1.0;
  }
  else {
    if (rf >100 ){
      return 4.5;
    }
    else {
      return (1+0.035*rf + rf*(rf-60)*(100-rf)*7/1000000);
    }
  }
}

#define HFUNC(X) ((X<0)?0.0:1.0)

/* e is total loss probability */
static double display_rfactor(RtpSession *session, double lfr, double r) {
  double d = session->rtp.extended_stats.rtt*1000.0*0.029;

  double I_d, I_ef, _Rf;
  /* compute the equipment impairment factor <I_ef> */
  I_ef = 30*log(1+15*lfr)*HFUNC(0.04-lfr) + 19*log(1+70*lfr)*HFUNC(lfr-0.04);

  I_d = 0.024*d;
  if (d>177.3) I_d = I_d + 0.11*(d-177.3);
  _Rf = r - I_d - I_ef;
  return _Rf;
}

static void compute_rtt(RtpSession *session, uint32_t lsr, uint32_t delay_snc_last_sr, struct timeval *rcv_time_tv) {
  uint32_t rcv_time;
  rcv_time = ortp_get_compact_ntp(rcv_time_tv);

  /* calculating Round Trip Time */
  if (lsr != 0) {
    double _Rf;
    session->rtp.extended_stats.rtt = (double)(rcv_time - delay_snc_last_sr - lsr);
    if (rcv_time - lsr <= delay_snc_last_sr) {
      session->rtp.extended_stats.rtt = 0.001;
    } else {
      session->rtp.extended_stats.rtt /= 65536.0;
    }

    _Rf = display_rfactor(session, session->rtp.extended_stats.cur_packet_loss_fraction, 94.2);
    ortp_message("RTCP (ssrc=%X) RTT=%fs rfactor=%.2f MOS=%.2f loss=%.0f%%\n", session->snd.ssrc, session->rtp.extended_stats.rtt, _Rf, convert_to_most(_Rf), session->rtp.extended_stats.cur_packet_loss_fraction * 100);
  }
}

void rtp_session_rtcp_parse(RtpSession *session, mblk_t *mp)
{
  rtcp_common_header_t *rtcp;
  int msgsize;
  int rtcp_pk_size;
  struct timeval rcv_time_tv;


  gettimeofday(&rcv_time_tv,NULL);

  return_if_fail(mp!=NULL);

  msgsize=(int) (mp->b_wptr-mp->b_rptr);

  if (msgsize < RTCP_COMMON_HEADER_SIZE)
  {
    ortp_warning("Receiving too short RTCP packet... discarded");
    return;
  }

  rtcp=(rtcp_common_header_t *)mp->b_rptr;

  /* compound rtcp packet can be composed by more than one rtcp message */
  while (msgsize >= RTCP_COMMON_HEADER_SIZE)
  {

    if (rtcp->version!=2)
    {
      ortp_warning("Receiving RTCP packet with version number !=2...discarded");
      return;
    }

    /* convert header data from network order to host order */
    rtcp->length = ntohs(rtcp->length);

    /* compute length */
    rtcp_pk_size = (rtcp->length + 1) * 4;
    /* Sanity check of simple RTCP packet length. */
    if (rtcp_pk_size > msgsize)
    {
      ortp_warning("Receiving RTCP packet shorter than the specified length.. discarded");
      return;
    }

    switch (rtcp->packet_type)
    {
    case RTCP_RTPFB:
    case RTCP_PSFB:

#if 0
      {
        rtcp_fb_t *fb = (rtcp_fb_t *) rtcp;
        uint32_t *cur;
        uint32_t *end;
        uint32_t *fci;

        OList *elem;
        RtpSession *final_session = session;
        RtpStream *rtpstream=&session->rtp;
        if (session->bundle!=NULL && session->use_bundle) { /* part of the bundle */
          for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
            RtpMedia *rtpMedia=(RtpMedia*) elem->data;
            if (ntohl(fb->ssrc_sender) == rtpMedia->session.rcv.ssrc) {
              final_session = &rtpMedia->session;
              rtpstream=&final_session->rtp;
              break;
            }
          }
        }

        if ( ntohl(fb->ssrc_sender) != final_session->rcv.ssrc )
        {
          ortp_warning("Receiving RTCP fb packet from unknown ssrc.. discarded");
          return;
        }

        if (msgsize < RTCP_COMMON_HEADER_SIZE + RTCP_SSRC_FIELD_SIZE + RTCP_SSRC_FIELD_SIZE)
        {
          ortp_warning("Receiving too short RTCP fb packet... discarded");
          return;
        }

        end=((uint32_t*)fb) + (1+fb->ch.length);
        cur=((uint32_t*)fb) + 3;
        fci=(uint32_t*)cur;

        if (fb->ch.packet_type==RTCP_PSFB && fb->ch.rc==1) {
          ortp_message("RTCP fb: pt:RTCP_PSFB fmt:%i FCI: Picture Loss Indication (PLI)", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_PSFB && fb->ch.rc==2) {
          ortp_message("RTCP fb: pt:RTCP_PSFB fmt:%i FCI: Slice Loss Indication (SLI)", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_PSFB && fb->ch.rc==3) {
          ortp_message("RTCP fb: pt:RTCP_PSFB fmt:%i FCI: Reference Picture Selection Indication (RPSI) ", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_PSFB && fb->ch.rc==4) {
          ortp_message("RTCP fb: pt:RTCP_PSFB fmt:%i FCI: Full Intra Request (FIR)", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_PSFB && fb->ch.rc==5) {
          ortp_message("RTCP fb: pt:RTCP_PSFB fmt:%i FCI: Temporal-Spatial Trade-off Request (TSTR)", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_PSFB && fb->ch.rc==6) {
          ortp_message("RTCP fb: pt:RTCP_PSFB fmt:%i FCI: TemporalTemporal-Spatial Trade-off Notification (TSTN)", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_PSFB && fb->ch.rc==7) {
          ortp_message("RTCP fb: pt:RTCP_PSFB fmt:%i FCI: Video Back Channel Message (VBCM)", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_PSFB && fb->ch.rc==15) {
          ortp_message("RTCP fb: pt:RTCP_PSFB fmt:%i FCI: Application layer FB (AFB) message", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_RTPFB && fb->ch.rc==1) {
          ortp_message("RTCP fb: pt:RTCP_RTPFB fmt:%i FCI: Generic NACK", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_RTPFB && fb->ch.rc==3) {
          ortp_message("RTCP fb: pt:RTCP_RTPFB fmt:%i FCI: Temporary Maximum Media Stream Bit Rate Request (TMMBR)", fb->ch.rc);
        }
        else if (fb->ch.packet_type==RTCP_RTPFB && fb->ch.rc==3) {
          ortp_message("RTCP fb: pt:RTCP_RTPFB fmt:%i FCI: Temporary Maximum Media Stream Bit Rate Notification (TMMBN)", fb->ch.rc);
        }
        else
          ortp_message("RTCP fb: pt:%i fmt:%i FCI: Unknown ", fb->ch.packet_type, fb->ch.rc);
        /* parsing all RTCP report blocks */
        for (; cur<end; )
        {
          break;
        }
      }
#endif
      break;
    case RTCP_XR:

#if 0
      ortp_message("Receiving RTCP XR");
#else
      {
        rtcp_xr_t *xr = (rtcp_xr_t *) rtcp;
        xreport_block_t *xrb;
        uint32_t *cur;
        uint32_t *end;

        OList *elem;
        RtpSession *final_session = session;

        ortp_message("Receiving RTCP_XR (ssrc=%X)", ntohl(xr->ssrc));

        if (session->bundle != NULL && session->use_bundle) { /* part of the bundle */
          for (elem = session->bundle->rtpMedias; elem != NULL; elem = o_list_next(elem)) {
            RtpMedia *rtpMedia = (RtpMedia*)elem->data;
            if (ntohl(xr->ssrc) == rtpMedia->session.rcv.ssrc) {
              final_session = &rtpMedia->session;
              break;
            }
          }
        }

        if (final_session->rcv.ssrc == 0) {
          /* rcv.ssrc is not set, so we adopt the incoming one */
          final_session->rcv.ssrc = ntohl(xr->ssrc);
          ortp_warning("Receiving RTCP xr packet from unknown ssrc.. adopting %X", final_session->rcv.ssrc);
        } else if (ntohl(xr->ssrc) != final_session->rcv.ssrc) {
          if (final_session->mode == RTP_SESSION_SENDONLY) /* final_session->snd.ssrc==0xAC288726) */
          {
            final_session->rcv.ssrc = ntohl(xr->ssrc);
            ortp_warning("Receiving RTCP xr packet from unknown ssrc.. fix/adopting %X", final_session->rcv.ssrc);
          } else {
            ortp_warning("Receiving RTCP xr packet from unknown ssrc.. discarded %X %X", ntohl(xr->ssrc), final_session->rcv.ssrc);
            return;
          }
        }

        if ( ntohl(xr->ssrc) != final_session->rcv.ssrc )
        {
          ortp_warning("Receiving RTCP xr packet from unknown ssrc.. discarded");		
          return;
        }	    

        if (msgsize < RTCP_COMMON_HEADER_SIZE + RTCP_SSRC_FIELD_SIZE + RTCP_XREPORT_BLOCK_SIZE)
        {
          ortp_warning("Receiving too short RTCP xr packet... discarded");
          return;
        }

        end=((uint32_t*)xr) + (1+xr->ch.length);
        cur=((uint32_t*)xr) + 2;
        xrb=(xreport_block_t*)cur;
        /* parsing all RTCP report blocks */    	  
        for (; cur<end; )
        {
          uint16_t block_len;
          /* parsing RTCP Sender Info */ 
          block_len = ntohs(xrb->block_len);
          switch (xrb->BT) {
          case RTCP_XR_LRLE: /* Loss RLE Report Block */
            ortp_message("XR block: RTCP_XR_LRLE %u %u", xrb->type_specific, block_len);
            break;
          case RTCP_XR_DRLE: /* Duplicate RLE Report Block */
            ortp_message("XR block: RTCP_XR_DRLE %u %u", xrb->type_specific, block_len);
            break;
          case RTCP_XR_PRT: /* Packet Receipt Times Report Block */
            ortp_message("XR block: RTCP_XR_PRT %u %u", xrb->type_specific, block_len);
            break;
          case RTCP_XR_RR: /* Receiver Reference Time Report Block */
            {
              rtcp_xr_rrt_t *xrrrt = (rtcp_xr_rrt_t*)(xrb + 1);

              uint32_t msw = ntohl(xrrrt->ntp_timestamp_msw);
              uint32_t lsw = ntohl(xrrrt->ntp_timestamp_lsw);
              ortp_message("XR block: RTCP_XR_RR %u %u", xrb->type_specific, block_len);

              final_session->rtcp_xr_stats.last_rcvr_rtt_ts = (msw << 16) | (lsw >> 16);

              final_session->rtcp_xr_stats.last_rcvr_rtt_time.tv_usec = rcv_time_tv.tv_usec;
              final_session->rtcp_xr_stats.last_rcvr_rtt_time.tv_sec = rcv_time_tv.tv_sec;
              final_session->rtcpxr_send_one_dlrr = TRUE; /* send one dlrr in next SR report */
            }
            break;
          case RTCP_XR_DLRR: /* DLRR Report Block */
            {
              rtcp_xr_dlrr_t *xrdlrr= (rtcp_xr_dlrr_t*)(xrb + 1);
              uint32_t lrr = ntohl(xrdlrr->last_rr);
              uint32_t dlrr = ntohl(xrdlrr->delaysince_last_rr);
              ortp_message("XR block: RTCP_XR_DLRR %u %u", xrb->type_specific, block_len);
              ortp_message("XR block: RTCP_XR_DLRR source ssrc: %X", ntohl(xrdlrr->ssrc));

              if (ntohl(xrdlrr->ssrc) != final_session->snd.ssrc) {
                ortp_warning("Received rtcp report block related to unknown ssrc (not from us)... discarded");
              } else {
                compute_rtt(final_session, lrr, dlrr, &rcv_time_tv);
              }
            }
            break;
          case RTCP_XR_SS: /* Statistics Summary Report Block */
            ortp_message("XR block: RTCP_XR_SS %u %u", xrb->type_specific, block_len);
            break;
          case RTCP_XR_VM: /* VoIP Metrics Report Block */
            {
              rtcp_xr_vm_t *xrvm=(rtcp_xr_vm_t*)(xrb + 1);
              ortp_message("XR block: RTCP_XR_VM %u %u", xrb->type_specific, block_len);

              ortp_message("XR block: RTCP_XR_VM source ssrc: %X", ntohl(xrvm->ssrc));
              ortp_message("XR block: RTCP_XR_VM loss=%u/256 discard=%u/256 burstdensity=%u gapdensity=%u", xrvm->loss_rate, xrvm->discard_rate, xrvm->burst_density, xrvm->gap_density);
              ortp_message("XR block: RTCP_XR_VM burstduration=%u gapduration=%u", ntohs(xrvm->burst_duration), ntohs(xrvm->gap_duration));
              ortp_message("XR block: RTCP_XR_VM roundtripdelay=%u endsystemdelay=%u", ntohs(xrvm->round_trip_delay), ntohs(xrvm->end_system_delay));
              ortp_message("XR block: RTCP_XR_VM signallevel=%u noiselevel=%u RERL=%u GMIN=%u", xrvm->signal_level, xrvm->noise_level, xrvm->RERL, xrvm->Gmin);
              ortp_message("XR block: RTCP_XR_VM Rfactor=%u extRfactor=%u MOS_LQ=%u MOS_CQ=%u", xrvm->R_factor, xrvm->ext_R_factor, xrvm->MOS_LQ, xrvm->MOS_CQ);
              ortp_message("XR block: RTCP_XR_VM RX_config=(PLC=%2X JBA=%2X JBrate=%4X) reserverd=%u JB_nominal=%u",
                           ((xrvm->RX_config>>6)&0x03),
                           ((xrvm->RX_config>>4)&0x03),
                           (xrvm->RX_config&0x0F),
                           xrvm->reserved, ntohs(xrvm->JB_nominal));
              ortp_message("XR block: RTCP_XR_VM JB_maximum=%u JB_abs_max=%u", ntohs(xrvm->JB_maximum), ntohs(xrvm->JB_abs_max));
            }
            break;
          case RTCP_XR_DM: /* Delay metrics block */
            ortp_message("XR block: RTCP_XR_DM %u %u", xrb->type_specific, block_len);
            break;
          case RTCP_XR_DJB: /* De-Jitter Buffer Metrics block */
            ortp_message("XR block: RTCP_XR_DJB %u %u", xrb->type_specific, block_len);
            break;
          default:
            ortp_message("XR block: %u %u %u", xrb->BT, xrb->type_specific, block_len);
            break;
          }
          cur+=(1+block_len);
          xrb=(xreport_block_t*)cur;
        }
      }
#endif
      break;
    case RTCP_SR:

      {
        rtcp_sr_t *sr = (rtcp_sr_t *) rtcp;
        report_block_t *rb;
        int i;
        OList *elem;
        RtpSession *final_session = session;
        RtpStream *rtpstream=&session->rtp;
        ortp_message("Receiving RTCP SR (ssrc=%X)", ntohl(sr->ssrc));
        if (session->bundle!=NULL && session->use_bundle) { /* part of the bundle */
          for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
            RtpMedia *rtpMedia=(RtpMedia*) elem->data;
            if (ntohl(sr->ssrc) == rtpMedia->session.rcv.ssrc) {
              final_session = &rtpMedia->session;
              rtpstream=&final_session->rtp;
              break;
            }
          }
        }

        if (msgsize < RTCP_COMMON_HEADER_SIZE + RTCP_SSRC_FIELD_SIZE + RTCP_SENDER_INFO_SIZE + (RTCP_REPORT_BLOCK_SIZE*sr->ch.rc))
        {
          ortp_warning("Receiving too short RTCP sr packet... discarded");
          return;
        }

        if ( ntohl(sr->ssrc) != final_session->rcv.ssrc )
        {
          ortp_warning("Receiving RTCP sr packet from unknown ssrc.. discarded");		
          break;
        }
        rtpstream->last_rcv_SR_ts = 0;
        rtpstream->last_rcv_SR_time.tv_usec = 0;
        rtpstream->last_rcv_SR_time.tv_sec = 0;

        /* parsing RTCP Sender Info */ 
        sr->si.ntp_timestamp_msw = ntohl(sr->si.ntp_timestamp_msw);
        sr->si.ntp_timestamp_lsw = ntohl(sr->si.ntp_timestamp_lsw);
        sr->si.rtp_timestamp = ntohl(sr->si.rtp_timestamp);
        sr->si.senders_packet_count = ntohl(sr->si.senders_packet_count);
        sr->si.senders_octet_count = ntohl(sr->si.senders_octet_count);	    

        /* saving data to fill LSR and DLSR field in next RTCP report to be transmitted  */
        rtpstream->last_rcv_SR_ts = (sr->si.ntp_timestamp_msw << 16) | (sr->si.ntp_timestamp_lsw >> 16);  
        rtpstream->last_rcv_SR_time.tv_usec = rcv_time_tv.tv_usec;
        rtpstream->last_rcv_SR_time.tv_sec = rcv_time_tv.tv_sec;


        /* parsing all RTCP report blocks */    	  
        for (i=0; i<sr->ch.rc; i++)
        { 
          rb = &(sr->rb[i]);

          if (ntohl(rb->ssrc) != final_session->snd.ssrc) {
            ortp_warning("Received rtcp report block related to unknown ssrc (not from us)... discarded");
          } else {
            uint32_t lsr = report_block_get_last_SR_time(rb);
            uint32_t delay_snc_last_sr = report_block_get_last_SR_delay(rb);
            compute_rtt(final_session, lsr, delay_snc_last_sr, &rcv_time_tv);
          }
        }

      }
      break;



    case RTCP_RR:

      {
        rtcp_rr_t *rr = (rtcp_rr_t *) rtcp;
        report_block_t *rb;
        int i;
        OList *elem;
        RtpSession *final_session = session;
        ortp_message("Receiving RTCP RR (ssrc=%X)", ntohl(rr->ssrc));
        if (session->bundle!=NULL && session->use_bundle) { /* part of the bundle */
          for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
            RtpMedia *rtpMedia=(RtpMedia*) elem->data;
            if (ntohl(rr->ssrc) == rtpMedia->session.rcv.ssrc) {
              final_session = &rtpMedia->session;
              break;
            }
          }
        }

        if (final_session->rcv.ssrc == 0)
        {
          /* rcv.ssrc is not set, so we adopt the incoming one */
          final_session->rcv.ssrc = ntohl(rr->ssrc);
          ortp_warning("Receiving RTCP rr packet from unknown ssrc.. adopting %X", final_session->rcv.ssrc);
        }
        else if ( ntohl(rr->ssrc) != final_session->rcv.ssrc )
        {
          if (final_session->mode==RTP_SESSION_SENDONLY) /* final_session->snd.ssrc==0xAC288726) */
          {
            final_session->rcv.ssrc = ntohl(rr->ssrc);
            ortp_warning("Receiving RTCP rr packet from unknown ssrc.. fix/adopting %X", final_session->rcv.ssrc);
          }
          else
          {
            ortp_warning("Receiving RTCP rr packet from unknown ssrc.. discarded %X %X", ntohl(rr->ssrc), final_session->rcv.ssrc);
            return;
          }
        }

        if (msgsize < RTCP_COMMON_HEADER_SIZE + RTCP_SSRC_FIELD_SIZE + (RTCP_REPORT_BLOCK_SIZE*rr->ch.rc))
        {
          ortp_warning("Receiving too short RTCP sr packet... discarded");
          return;
        }	    

        /* parsing all RTCP report blocks */    	  
        for (i=0; i<rr->ch.rc; i++)
        {
          rb = &(rr->rb[i]);

          if (ntohl(rb->ssrc) != final_session->snd.ssrc) {
            ortp_warning("Received rtcp report block related to unknown ssrc (not from us)... discarded");
          } else {
            uint32_t lsr = report_block_get_last_SR_time(rb);
            uint32_t delay_snc_last_sr = report_block_get_last_SR_delay(rb);
            compute_rtt(final_session, lsr, delay_snc_last_sr, &rcv_time_tv);
          }
        }

      }
      break;


    case RTCP_SDES: 
      /* to be implemented */

      {
        RtpSession *final_session = session;
        uint8_t *rptr=(uint8_t*)rtcp+sizeof(rtcp_common_header_t);
        uint8_t *end=(uint8_t*)rtcp + rtcp_pk_size;
        uint32_t csrc=0;
        int nchunk=0;
        bool_t chunk_start=TRUE;

        ortp_debug("Receiving RTCP SDES");
        while(rptr<end){

          if (chunk_start){
            final_session = session;
            if (rptr+4<=end){
              csrc=ntohl(*(uint32_t*)rptr);
              rptr+=4;

              if (session->bundle!=NULL && session->use_bundle) { /* part of the bundle */
                OList *elem;
                for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
                  RtpMedia *rtpMedia=(RtpMedia*) elem->data;
                  if (ntohl(csrc) == rtpMedia->session.rcv.ssrc) {
                    final_session = &rtpMedia->session;
                    break;
                  }
                }
              }

            }else{
              ortp_warning("incorrect chunk start in RTCP SDES");
              break;
            }
            chunk_start=FALSE;
          }else{
            if (rptr+2<=end){
              uint8_t type=rptr[0];
              uint8_t len=rptr[1];

              if (type==RTCP_SDES_END){
                /* pad to next 32bit boundary*/
                //CHECKME
                rptr=(uint8_t*)(((intptr_t)rptr+4) & ~0x3);
                nchunk++;
                if (nchunk<rtcp_common_header_get_rc(rtcp)){
                  chunk_start=TRUE;
                  continue;
                }else break;
              }
              rptr+=2;
              if (rptr+len<=end){
                if (type==7 && strncmp((const char *)rptr, "video stream", len)==0) {
                  /* NOTE for antisip app: do not change */
                  final_session->rtcpfeedback_pli=TRUE;
                  /* do we want this always?  TODO! */
                  /* final_session->rtcpfeedback_ccm_tmmbr=TRUE; */
                }
                rptr+=len;
              }else{
                ortp_warning("bad item length in RTCP SDES");
                break;
              }
            }else{
              /*end of packet */
              break;
            }
          }
        }
      }
      break;


    case RTCP_BYE:
      {
        rtcp_bye_t *bye = (rtcp_bye_t *) rtcp;
        unsigned sclen = bye->ch.rc * 4;
        int reason_space_len = rtcp_pk_size
          - sizeof (rtcp_common_header_t)
          - sclen;
        int i;
        char *reason = NULL;
        bool_t rcv_ssrc_match = FALSE;

        ortp_message("Receiving RTCP BYE");
        if (reason_space_len < 0) {
          ortp_warning("Receiving too short RTCP BYE packet... discarded");
          return;
        }
        for (i = 0; i < bye->ch.rc; i++) {
          if (ntohl(bye->ssrc[i]) == session->rcv.ssrc) {
            rcv_ssrc_match = TRUE;
            break;
          }
        }
        if (rcv_ssrc_match) {
          if (session->on_rtcp_bye.count > 0) {
            /* Get reason. */
            if (reason_space_len > 1) {
              uint8_t *reasonbuf = (uint8_t *) rtcp
                + sizeof (rtcp_common_header_t)
                + sclen;
              if (reasonbuf[0] <= reason_space_len-1)
                reason = ortp_strndup((char *)(reasonbuf+1), reasonbuf[0]);
              else
                ortp_warning("Incorrect RTCP BYE reason length");
            }
            rtp_signal_table_emit2(&session->on_rtcp_bye, (uintptr_t)reason);
            if (reason)
              ortp_free(reason);
          } else {
            ortp_warning("Got RTCP BYE without RTCP BYE handler");
          }
        } else {
          ortp_warning("No SSRC in the RTCP BYE packet matched our rcv.ssrc.");
        }
        break;
      }

    case RTCP_APP:
      /* to be implemented */
      ortp_debug("Receiving RTCP app");
      break;


    default:

      ortp_warning("Receiving unknown RTCP packet type... discarded");
      return;

    }


    msgsize -= rtcp_pk_size;              /* size of unparsed portion of UDP packet, in octets */
    rtcp = (rtcp_common_header_t *) (rtcp_pk_size + (char *) rtcp);  /* pointer to next RTCP packet in current UDP packet */

  }

  /* The function did not failed sanity checks, write down the RTPC/RTCP
  reception time. */
  session->last_recv_time = rcv_time_tv;
}
