/*
The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
Copyright (C) 2010-2014  Aymeric moizard amoizard@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#if defined(WIN32) || defined(_WIN32_WCE)
#include "ortp-config-win32.h"
#elif HAVE_CONFIG_H
#include "ortp-config.h"
#endif

#include "ortp/ortp.h"
#include "rtpsession_priv.h"
#include "utils.h"

RtpMedia *
rtp_media_new (int session_mode, bool_t use_bundle)
{
  RtpMedia *media;
  media = (RtpMedia *) ortp_malloc0 (sizeof (RtpMedia));
  if (media == NULL)
  {
    ortp_error("rtp_media_new: Memory allocation failed");
    return NULL;
  }
  rtp_session_init (&media->session, session_mode);
  media->session.use_bundle = use_bundle;
  return media;
}

void
rtp_media_destroy (RtpMedia *media)
{
  rtp_session_uninit (&media->session);
  ortp_free(media);
}

RtpBundle *
rtp_bundle_new (void)
{
  RtpBundle *bundle;
  bundle = (RtpBundle *) ortp_malloc0 (sizeof (RtpBundle));
  if (bundle == NULL)
  {
    ortp_error("rtp_bundle_new: Memory allocation failed");
    return NULL;
  }

#if !defined(WIN32) && !defined(_WIN32_WCE)
  {
    pthread_mutexattr_t Attr;

    pthread_mutexattr_init(&Attr);
    pthread_mutexattr_settype(&Attr, PTHREAD_MUTEX_RECURSIVE);
    ortp_mutex_init(&bundle->lock, &Attr);
  }
#else
  ortp_mutex_init(&bundle->lock, NULL);
#endif
  return bundle;
}

void
rtp_bundle_destroy (RtpBundle *bundle)
{
  if (bundle==NULL)
    return;
  rtp_bundle_stop(bundle);
  if (bundle->rtpMedias!=NULL)
    o_list_free(bundle->rtpMedias);
  ortp_mutex_destroy(&bundle->lock);
  ortp_free(bundle);
}

void
rtp_bundle_lock (RtpBundle *bundle) {
    if (bundle==NULL)
      return;
    ortp_mutex_lock(&bundle->lock);
}

void
rtp_bundle_unlock (RtpBundle *bundle) {
    if (bundle==NULL)
      return;
    ortp_mutex_unlock(&bundle->lock);
}

void
rtp_bundle_media_add (RtpBundle *bundle, RtpMedia *_RtpMedia, int pos)
{
  rtp_bundle_lock(bundle);
  if (pos==0)
    bundle->rtpMedias = o_list_insert(bundle->rtpMedias, _RtpMedia);
  else
    bundle->rtpMedias = o_list_append(bundle->rtpMedias, _RtpMedia);
  rtp_session_set_bundle(&_RtpMedia->session, bundle);
  rtp_bundle_unlock(bundle);
}

void
rtp_bundle_media_remove (RtpBundle *bundle, RtpMedia *_RtpMedia)
{
  rtp_bundle_lock(bundle);
  bundle->rtpMedias = o_list_remove(bundle->rtpMedias, _RtpMedia);
  rtp_session_set_bundle(&_RtpMedia->session, NULL);
  rtp_bundle_unlock(bundle);
}

static void _internal_usleep(uint64_t usec){
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
  struct timeval delay;
  int sec;

  sec = (int) usec / 1000000;
  if (sec > 0) {
    delay.tv_sec = sec;
    delay.tv_usec = 0;
  }
  else {
    delay.tv_sec = 0;
    delay.tv_usec = (long)usec;
  }
  select (0, 0, 0, 0, &delay);
#elif defined(WIN32)
	Sleep((DWORD)(usec/1000));
#else
	struct timespec ts,rem;
	int err;
	ts.tv_sec=usec/1000000LL;
	ts.tv_nsec=(usec%1000000LL)*1000;
	do {
		err=nanosleep(&ts,&rem);
		ts=rem;
	}while(err==-1 && errno==EINTR);
#endif
}

static void *_rtp_bundle_run(void *arg)
{
  RtpBundle *bundle=(RtpBundle*)arg;
  while (bundle->run==TRUE) {
    OList *elem;

    struct timeval tv;
    ortp_socket_t max = 0;
    fd_set fdset;
    int i;

    FD_ZERO (&fdset);

    rtp_bundle_lock(bundle);

    for (elem=bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
      RtpMedia *rtpMedia=(RtpMedia*) elem->data;

      if (!(rtpMedia->session.flags & RTP_SESSION_ENABLE_SENDRECV))
        continue;
      if (rtpMedia->session.flags & RTP_SESSION_RECV_NOT_STARTED)
        continue; /* timestamp for session not initialized */

      for (i=0;i<10;i++) {
        if (sockisvalid(rtpMedia->session.rtp.socketv4[i])) {
          FD_SET (rtpMedia->session.rtp.socketv4[i], &fdset);
          max = MAX(max, rtpMedia->session.rtp.socketv4[i]+ 1);
        }
        if (sockisvalid(rtpMedia->session.rtcp.socketv4[i])) {
          FD_SET (rtpMedia->session.rtcp.socketv4[i], &fdset);
          max = MAX(max, rtpMedia->session.rtcp.socketv4[i]+ 1);
        }
      }

      if (rtp_session_using_transport((&rtpMedia->session), rtp)) {
        ortp_socket_t rtp_socket = rtpMedia->session.rtp.tr->t_rtp_getsocket(&rtpMedia->session);
        ortp_socket_t rtcp_socket = rtpMedia->session.rtp.tr->t_rtcp_getsocket(&rtpMedia->session);
        if (sockisvalid(rtp_socket)) {
          FD_SET (rtp_socket, &fdset);
          max = MAX(max, rtp_socket+ 1);
        }
        if (sockisvalid(rtcp_socket)) {
          FD_SET (rtcp_socket, &fdset);
          max = MAX(max, rtcp_socket+ 1);
        }
      }
    }

    rtp_bundle_unlock(bundle);

    tv.tv_sec = 0;
    tv.tv_usec = 20000; /* each 20ms */

    if (bundle->run!=TRUE)
      break;

    if (max==0) {
      _internal_usleep(tv.tv_usec);
      continue;
    }

    i = select((int)max, &fdset, NULL, NULL, &tv);
    if (bundle->run!=TRUE)
      break;
    if ((i == -1) && (errno == EINTR || errno == EAGAIN))
      continue;
    if (i == -1)
      continue; /* error!! */

    rtp_bundle_lock(bundle);

    for (elem=bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
      RtpMedia *rtpMedia=(RtpMedia*) elem->data;

      if (!(rtpMedia->session.flags & RTP_SESSION_ENABLE_SENDRECV))
        continue;

      if (rtp_session_using_transport((&rtpMedia->session), rtp)) {
        /* this will do the ICE connection even if no activity exist */
        if (rtpMedia->session.rtp.tr->t_rtp_refresh!=NULL)
         rtpMedia->session.rtp.tr->t_rtp_refresh(&rtpMedia->session);
        if (rtpMedia->session.rtp.tr->t_rtcp_refresh!=NULL)
          rtpMedia->session.rtp.tr->t_rtcp_refresh(&rtpMedia->session);
      }

      for (i=0;i<10;i++) {
        if (sockisvalid(rtpMedia->session.rtp.socketv4[i])) {
          if (FD_ISSET (rtpMedia->session.rtp.socketv4[i], &fdset)) {
            rtp_session_rtp_recv (&rtpMedia->session, i);
            rtp_session_rtcp_recv(&rtpMedia->session, i);
          }
        }

        if (sockisvalid(rtpMedia->session.rtcp.socketv4[i])) {
          if (FD_ISSET (rtpMedia->session.rtcp.socketv4[i], &fdset)) {
            rtp_session_rtp_recv (&rtpMedia->session, i);
            rtp_session_rtcp_recv(&rtpMedia->session, i);
            continue;
          }
        }
      }

      if (rtp_session_using_transport((&rtpMedia->session), rtp)) {
        ortp_socket_t rtp_socket = rtpMedia->session.rtp.tr->t_rtp_getsocket(&rtpMedia->session);
        ortp_socket_t rtcp_socket = rtpMedia->session.rtp.tr->t_rtcp_getsocket(&rtpMedia->session);
        if (sockisvalid(rtp_socket)) {
          if (FD_ISSET (rtp_socket, &fdset)) {
            rtp_session_rtp_recv (&rtpMedia->session, 0);
            rtp_session_rtcp_recv(&rtpMedia->session, 0);
            continue;
          }
        }
        if (sockisvalid(rtcp_socket)) {
          if (FD_ISSET (rtcp_socket, &fdset)) {
            rtp_session_rtp_recv (&rtpMedia->session, 0);
            rtp_session_rtcp_recv(&rtpMedia->session, 0);
            continue;
          }
        }
      }
    }

    rtp_bundle_unlock(bundle);

  }
  return NULL;
}

void
rtp_bundle_start(RtpBundle *bundle)
{
  if (bundle==NULL)
    return;
  if (bundle->run==TRUE)
    return;
  bundle->run=TRUE;
  ortp_thread_create(&bundle->thread,NULL,_rtp_bundle_run,bundle);
}

void
rtp_bundle_stop(RtpBundle *bundle)
{
  if (bundle==NULL)
    return;
  if (bundle->run==FALSE)
    return;
  bundle->run=FALSE;
  ortp_thread_join(bundle->thread,NULL);
}
