/*
The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "ortp/ortp.h"
#include "jitterctl.h"
#include "utils.h"
#include "rtpsession_priv.h"

static bool_t queue_packet(queue_t *q, int maxrqsz, mblk_t *mp, rtp_header_t *rtp, int *discarded)
{
  mblk_t *tmp;
  int header_size;
  *discarded=0;
  header_size=RTP_FIXED_HEADER_SIZE+ (4*rtp->cc)+(4*rtp->extbit);

  if (rtp->extbit)
  {
    uint16_t *ext_buf = (uint16_t *)(mp->b_rptr+header_size-2);
    header_size += (ntohs(ext_buf[0])*4);
  }

  if ((mp->b_wptr - mp->b_rptr)==header_size){
    ortp_debug("Rtp packet contains no data.");
    (*discarded)++;
    freemsg(mp);
    return FALSE;
  }
  /* and then add the packet to the queue */

  rtp_putq(q,mp);
  /* make some checks: q size must not exceed RtpStream::max_rq_size */
  while (q->q_mcount > maxrqsz)
  {
    /* remove the oldest mblk_t */
    tmp=getq(q);
    if (mp!=NULL)
    {
      ortp_debug("rtp_putq: Queue is full. Discarding message with ts=%i",((rtp_header_t*)mp->b_rptr)->timestamp);
      freemsg(tmp);
      (*discarded)++;
    }
  }
  return TRUE;
}

static RtpSession *rtp_session_rtcp_demux(RtpSession *session, mblk_t *mp)
{
  rtcp_common_header_t *rtcp;
  int msgsize;
  int rtcp_pk_size;

  if (session->bundle==NULL || !session->use_bundle) {
    return session;
  }

  msgsize=(int) (mp->b_wptr-mp->b_rptr);

  if (msgsize < RTCP_COMMON_HEADER_SIZE)
  {
    ortp_warning("Receiving too short RTCP packet... discarded");
    return NULL;
  }

  rtcp=(rtcp_common_header_t *)mp->b_rptr;

  /* compound rtcp packet can be composed by more than one rtcp message */
  while (msgsize >= RTCP_COMMON_HEADER_SIZE)
  {
    int rtcp_length;
    if (rtcp->version!=2)
    {
      ortp_warning("Receiving RTCP packet with version number !=2...discarded");
      return NULL;
    }

    /* convert header data from network order to host order */
    rtcp_length = ntohs(rtcp->length);

    /* compute length */
    rtcp_pk_size = (rtcp_length + 1) * 4;
    /* Sanity check of simple RTCP packet length. */
    if (rtcp_pk_size > msgsize)
    {
      ortp_warning("Receiving RTCP packet shorter than the specified length.. discarded");
      return NULL;
    }

    switch (rtcp->packet_type)
    {
    case RTCP_RTPFB:
    case RTCP_PSFB:
      {
        rtcp_fb_t *fb = (rtcp_fb_t *) rtcp;
        OList *elem;

        if (msgsize < RTCP_COMMON_HEADER_SIZE + RTCP_SSRC_FIELD_SIZE + RTCP_SSRC_FIELD_SIZE)
        {
          ortp_warning("Receiving too short RTCP fb packet... discarded");
          return NULL;
        }

        for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
          RtpMedia *rtpMedia=(RtpMedia*) elem->data;
          if (ntohl(fb->ssrc_sender) == rtpMedia->session.rcv.ssrc) {
            return &rtpMedia->session;
          }
        }

        return NULL;
      }

    case RTCP_XR:
      {
        rtcp_xr_t *xr = (rtcp_xr_t *) rtcp;
        OList *elem;

        if (msgsize < RTCP_COMMON_HEADER_SIZE + RTCP_SSRC_FIELD_SIZE + RTCP_XREPORT_BLOCK_SIZE)
        {
          ortp_warning("Receiving too short RTCP xr packet... discarded");
          return NULL;
        }
        for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
          RtpMedia *rtpMedia=(RtpMedia*) elem->data;
          if (ntohl(xr->ssrc) == rtpMedia->session.rcv.ssrc) {
            return &rtpMedia->session;
          }
        }

        return NULL;
      }

    case RTCP_SR:
      {
        rtcp_sr_t *sr = (rtcp_sr_t *) rtcp;
        OList *elem;

        if (msgsize < RTCP_COMMON_HEADER_SIZE + RTCP_SSRC_FIELD_SIZE + RTCP_SENDER_INFO_SIZE + (RTCP_REPORT_BLOCK_SIZE*sr->ch.rc))
        {
          ortp_warning("Receiving too short RTCP sr packet... discarded");
          return NULL;
        }

        for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
          RtpMedia *rtpMedia=(RtpMedia*) elem->data;
          if (ntohl(sr->ssrc) == rtpMedia->session.rcv.ssrc) {
            return &rtpMedia->session;
          }
        }

        return NULL;
      }

    case RTCP_RR:
      {
        rtcp_rr_t *rr = (rtcp_rr_t *) rtcp;
        OList *elem;

        if (msgsize < RTCP_COMMON_HEADER_SIZE + RTCP_SSRC_FIELD_SIZE + (RTCP_REPORT_BLOCK_SIZE*rr->ch.rc))
        {
          ortp_warning("Receiving too short RTCP rr packet... discarded");
          return NULL;
        }

        for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
          RtpMedia *rtpMedia=(RtpMedia*) elem->data;
          if (ntohl(rr->ssrc) == rtpMedia->session.rcv.ssrc) {
            return &rtpMedia->session;
          }
        }
        return NULL;
      }

    case RTCP_SDES:
      {
        uint8_t *rptr=(uint8_t*)rtcp+sizeof(rtcp_common_header_t);
        uint8_t *end=(uint8_t*)rtcp + rtcp_pk_size;
        uint32_t csrc=0;
        int nchunk=0;
        bool_t chunk_start=TRUE;

        while(rptr<end){

          if (chunk_start){
            if (rptr+4<=end){
              OList *elem;
              csrc=ntohl(*(uint32_t*)rptr);
              rptr+=4;

              for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
                RtpMedia *rtpMedia=(RtpMedia*) elem->data;
                if (ntohl(csrc) == rtpMedia->session.rcv.ssrc) {
                  return &rtpMedia->session;
                }
              }

            }else{
              return NULL;
            }
            chunk_start=FALSE;
          }else{
            if (rptr+2<=end){
              uint8_t type=rptr[0];
              uint8_t len=rptr[1];

              if (type==RTCP_SDES_END){
                /* pad to next 32bit boundary*/
                //CHECKME
                rptr=(uint8_t*)(((intptr_t)rptr+4) & ~0x3);
                nchunk++;
                if (nchunk<rtcp_common_header_get_rc(rtcp)){
                  chunk_start=TRUE;
                  continue;
                }else break;
              }
              rptr+=2;
              if (rptr+len<=end){
                rptr+=len;
              }else{
                return NULL;
              }
            }else{
              /*end of packet */
              break;
            }
          }
        }
        return NULL;
      }

    case RTCP_BYE:
      {
        rtcp_bye_t *bye = (rtcp_bye_t *) rtcp;
        unsigned sclen = bye->ch.rc * 4;
        int reason_space_len = rtcp_pk_size
          - sizeof (rtcp_common_header_t)
          - sclen;
        int i;

        if (reason_space_len < 0) {
          ortp_warning("Receiving too short RTCP BYE packet... discarded");
          return NULL;
        }
        for (i = 0; i < bye->ch.rc; i++) {
          OList *elem;
          for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
            RtpMedia *rtpMedia=(RtpMedia*) elem->data;
            if (ntohl(bye->ssrc[i]) == rtpMedia->session.rcv.ssrc) {
              return &rtpMedia->session;
            }
          }
        }
        return NULL;
      }

    case RTCP_APP:
      {
        rtcp_app_t *app = (rtcp_app_t *) rtcp;
        OList *elem;

        if (msgsize < sizeof(rtcp_app_t))
        {
          ortp_warning("Receiving too short RTCP rr packet... discarded");
          return NULL;
        }

        for (elem=session->bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
          RtpMedia *rtpMedia=(RtpMedia*) elem->data;
          if (ntohl(app->ssrc) == rtpMedia->session.rcv.ssrc) {
            return &rtpMedia->session;
          }
        }
        return NULL;
      }
    default:
      ortp_warning("Receiving unknown RTCP packet type... discarded (demux)");
      return NULL;
    }

    msgsize -= rtcp_pk_size;              /* size of unparsed portion of UDP packet, in octets */
    rtcp = (rtcp_common_header_t *) (rtcp_pk_size + (char *) rtcp);  /* pointer to next RTCP packet in current UDP packet */
  }
  return NULL;
}

RtpSession *rtp_session_demux_media(RtpBundle *bundle, RtpSession *session, mblk_t *mp, struct sockaddr *addr, socklen_t addrlen)
{
  OList *elem;
  rtp_header_t *rtp;
  int msgsize;
  RtpStream *rtpstream=&session->rtp;
#if 0
  rtp_stats_t *stats=&rtpstream->stats;
#endif

  msgsize=(int)(mp->b_wptr-mp->b_rptr);

#if 0
  if (sockisvalid(session->rtp.socket) && session->symmetric_rtp && stats->packet_recv==0){
    /* store the sender rtp address to do symmetric RTP */
    if (addrlen!=session->rtp.rem_addrlen
      || memcmp(&session->rtp.rem_addr,addr, addrlen)!=0)
    {
      if (addrlen>0) {
        ortp_message("Symmetric RTP with initial RTP/data packet! %i", session->rtp.loc_port);
        memcpy(&session->rtp.rem_addr,addr,addrlen);
        session->rtp.rem_addrlen=addrlen;
        rtp_session_set_dscp(session, -1);
      }
    }
  }
#endif

  if (msgsize<RTP_FIXED_HEADER_SIZE){
    ortp_warning("Packet too small to be a rtp packet (%i)!",msgsize);
    rtpstream->stats.bad++;
    ortp_global_stats.bad++;
    freemsg(mp);
    return NULL;
  }

  rtp=(rtp_header_t*)mp->b_rptr;
  if (rtp->version!=2)
  {
    return session; 
  }

  if (session->rtcp_mux && rtp->paytype>=64 && rtp->paytype<=95)
  {
    /* rtcp packet: decide which session here... */
    RtpSession * final_session = rtp_session_rtcp_demux(session, mp);
    if (final_session==NULL)
    {
      freemsg(mp);
      return NULL;
    }
    return final_session;
  }

  if (rtp->paytype==session->rcv.telephone_events_pt)
    return session;

  if (session->hw_recv_pt==rtp->paytype)
    return session;

  if (bundle==NULL)
    return session;

  if (!session->use_bundle)
    return session;

  for (elem=bundle->rtpMedias;elem!=NULL;elem=o_list_next(elem)){
    PayloadType *pt_new;
    RtpMedia *rtpMedia=(RtpMedia*) elem->data;
    if (rtpMedia->session.hw_recv_pt==rtp->paytype)
      return &rtpMedia->session;
    pt_new=rtp_profile_get_payload(rtpMedia->session.rcv.profile,rtp->paytype);
    if (pt_new!=0) {
      ortp_message("rtpsession found for RTP stream with payload (%i)!", rtp->paytype);
      return &rtpMedia->session;
    }
  }

  ortp_warning("no rtp stream for packet (%i)!",msgsize);
  rtpstream->stats.bad++;
  ortp_global_stats.bad++;
  freemsg(mp);
  return NULL;
}

int rtp_session_rtp_parse(RtpSession *session, mblk_t *mp, struct sockaddr *addr, socklen_t addrlen, int initial_packet_size, int *rtp_overhead)
{
  int i;
  rtp_header_t *rtp;
  int msgsize;
  RtpStream *rtpstream=&session->rtp;
  rtp_stats_t *stats=&rtpstream->stats;

  *rtp_overhead=0;

  msgsize=(int)(mp->b_wptr-mp->b_rptr);

  if (msgsize<RTP_FIXED_HEADER_SIZE){
    ortp_warning("Packet too small to be a rtp packet (%i)!",msgsize);
    rtpstream->stats.bad++;
    ortp_global_stats.bad++;
    freemsg(mp);
    return RFC5764BIS_BAD;
  }

#if 0
  if (sockisvalid(session->rtp.socket) && session->symmetric_rtp && stats->packet_recv==0){
    /* store the sender rtp address to do symmetric RTP */
    if (addrlen!=session->rtp.rem_addrlen
      || memcmp(&session->rtp.rem_addr,addr, addrlen)!=0)
    {
      if (addrlen>0) {
        ortp_message("Symmetric RTP with initial RTP/data packet! %i", session->rtp.loc_port);
        memcpy(&session->rtp.rem_addr,addr,addrlen);
        session->rtp.rem_addrlen=addrlen;
        rtp_session_set_dscp(session, -1);
      }
    }
  }
#endif

  rtp=(rtp_header_t*)mp->b_rptr;
  if (rtp->version!=2)
  {
    /* discard in two case: the packet is not stun OR nobody is interested by STUN (no eventqs) */
    ortp_debug("Receiving rtp packet with version number !=2...discarded");
    stats->bad++;
    ortp_global_stats.bad++;
    freemsg(mp);
    return RFC5764BIS_BAD;
  }

  if (session->rtcp_mux && rtp->paytype>=64 && rtp->paytype<=95)
  {
    mblk_t *mp2 = dupb(mp);
    rtp_session_notify_inc_rtcp(session,mp);
    if (mp2!=NULL)
    {
      rtp_session_rtcp_parse(session,mp2); //(this modify the content of RTCP data so we work on duplicate)
      freemsg(mp2);
    }
    return RFC5764BIS_RTCP;
  }

  /* only count non-stun packets. */
  ortp_global_stats.packet_recv++;
  stats->packet_recv++;
  ortp_global_stats.hw_recv+=msgsize;
  stats->hw_recv+=msgsize;

  /* convert all header data from network order to host order */
  rtp->seq_number=ntohs(rtp->seq_number);
  rtp->timestamp=ntohl(rtp->timestamp);
  rtp->ssrc=ntohl(rtp->ssrc);
  /* convert csrc if necessary */
  if (rtp->cc*sizeof(uint32_t) > (uint32_t) (msgsize-RTP_FIXED_HEADER_SIZE)){
    ortp_debug("Receiving too short rtp packet.");
    stats->bad++;
    ortp_global_stats.bad++;
    freemsg(mp);
    return RFC5764BIS_BAD;
  }

#ifndef PERF
  /* Write down the last RTP/RTCP packet reception time. */
  gettimeofday(&session->last_recv_time, NULL);
#endif

  for (i=0;i<rtp->cc;i++)
    rtp->csrc[i]=ntohl(rtp->csrc[i]);
  /*the goal of the following code is to lock on an incoming SSRC to avoid
  receiving "mixed streams"*/
  if (session->ssrc_set){
    /*the ssrc is set, so we must check it */
    if (session->rcv.ssrc!=rtp->ssrc){
      if (session->inc_ssrc_candidate==rtp->ssrc){
        session->inc_same_ssrc_count++;
      }else{
        session->inc_same_ssrc_count=0;
        session->inc_ssrc_candidate=rtp->ssrc;
      }
      if (session->inc_same_ssrc_count>=session->rtp.ssrc_changed_thres){

#if 0
        /* store the sender rtp address to do symmetric RTP */
        if (sockisvalid(session->rtp.socket) && session->symmetric_rtp){
          /* store the sender rtp address to do symmetric RTP */
          if (addrlen!=session->rtp.rem_addrlen
            || memcmp(&session->rtp.rem_addr,addr, addrlen)!=0)
          {
            if (addrlen>0) {
              ortp_message("Symmetric RTP with new adopted SSRC! %i", session->rtp.loc_port);
              memcpy(&session->rtp.rem_addr,addr,addrlen);
              session->rtp.rem_addrlen=addrlen;
              rtp_session_set_dscp(session, -1);
            }
          }
        }
#endif
        session->rtp.rcv_last_ts = rtp->timestamp;
        session->rcv.ssrc=rtp->ssrc;
        rtp_signal_table_emit(&session->on_ssrc_changed);
        rtpstream->hwrcv_extseq=0;
        rtpstream->hwrcv_seq_at_last_SR=rtp->seq_number;
        rtpstream->hwrcv_since_last_SR=0;
        rtpstream->last_rcv_SR_time.tv_usec = 0;
        rtpstream->last_rcv_SR_time.tv_sec = 0;
        rtpstream->last_rcv_SR_ts=0;
        flushq(&session->rtp.rq, FLUSHALL);
      }else{
        /*discard the packet*/
        if (session->rtp.jittctl.clock_rate==90000)
          ortp_warning("%u Receiving packet with unknown ssrc.", rtp->ssrc);
        stats->bad++;
        ortp_global_stats.bad++;
        freemsg(mp);
        return RFC5764BIS_BAD;
      }
    }
    else{
      /* The SSRC change must not happen if we still receive
      ssrc from the initial source. */
      session->inc_same_ssrc_count=0;
#if 0
      if (sockisvalid(session->rtp.socket) && session->symmetric_rtp){
        /* store the sender rtp address to do symmetric RTP */
        if (addrlen!=session->rtp.rem_addrlen
          || memcmp(&session->rtp.rem_addr,addr, addrlen)!=0)
        {
          session->inc_same_ssrc_new_origin_count++;
          if ((!(session->flags & RTP_SESSION_FIRST_PACKET_DELIVERED)) || session->inc_same_ssrc_new_origin_count>10){
            session->inc_same_ssrc_new_origin_count=0;
            if (addrlen>0) {
              ortp_message("Symmetric RTP with initial RTP/delivered packet! %i", session->rtp.loc_port);
              memcpy(&session->rtp.rem_addr,addr,addrlen);
              session->rtp.rem_addrlen=addrlen;
              rtp_session_set_dscp(session, -1);
            }
          }
        } else {
          session->inc_same_ssrc_new_origin_count=0;
        }
      }
#endif
    }

  }else{
    if (session->rcv.ssrc!=0 && session->rcv.ssrc!=rtp->ssrc) {
      ortp_warning("rtp_parse: drop RTP packet, ssrc was negotiated, but it's wrong here (%u:%u)", session->rcv.ssrc, rtp->ssrc);
      //freemsg(mp);
      //stats->bad++;
      //ortp_global_stats.bad++;
      //return RFC5764BIS_BAD;
    }
    session->ssrc_set=TRUE;
    session->rcv.ssrc=rtp->ssrc;

    /* the first sequence number received should be initialized at the beginning, so that the first receiver reports contains valid loss rate*/
    session->rtp.rcv_last_ts = rtp->timestamp;
    rtpstream->hwrcv_seq_at_last_SR=rtp->seq_number;
    rtpstream->hwrcv_since_last_SR=0;
    flushq(&session->rtp.rq, FLUSHALL);

#if 0
    if (sockisvalid(session->rtp.socket) && session->symmetric_rtp){
      /* store the sender rtp address to do symmetric RTP */
      if (addrlen!=session->rtp.rem_addrlen
        || memcmp(&session->rtp.rem_addr,addr, addrlen)!=0)
      {
        if (addrlen>0) {
          ortp_message("Symmetric RTP with initial SSRC delivered packet! %i", session->rtp.loc_port);
          memcpy(&session->rtp.rem_addr,addr,addrlen);
          session->rtp.rem_addrlen=addrlen;
          rtp_session_set_dscp(session, -1);
        }
      }
    }
#endif
  }

  if (session->flags & RTP_SESSION_FIRST_PACKET_DELIVERED) {
    if (rtp->timestamp==0 && rtp->seq_number==0)
    {
      /* interop issue with SPIRIT sending ARS packets */
      /* drop */
      ortp_warning("rtp_parse: drop RTP packet (timestamp=0 seqnumber=0)");
      freemsg(mp);
      stats->bad++;
      ortp_global_stats.bad++;
      return RFC5764BIS_BAD;
    }
  }

  {
    int header_size;
    header_size=RTP_FIXED_HEADER_SIZE+ (4*rtp->cc)+(4*rtp->extbit);

    if (rtp->extbit)
    {
      uint16_t *ext_buf = (uint16_t *)(mp->b_rptr+header_size-2);
      header_size += (ntohs(ext_buf[0])*4);
    }
    if (initial_packet_size>header_size) {
      struct timeval tv;
      gettimeofday(&tv, NULL);

      *rtp_overhead=header_size;
    }
  }

  /* check for possible telephone events */
  if (rtp->paytype==session->rcv.telephone_events_pt){

    session->rtp.hwrcv_since_last_SR++;
    /* update hwrcv_extseq */
    {
      poly32_t *extseq=(poly32_t*)&rtpstream->hwrcv_extseq;
      if (rtp->seq_number>extseq->split.lo){
        extseq->split.lo=rtp->seq_number;
      }else if (rtp->seq_number<200 && extseq->split.lo>((1<<16) - 200)){
        /* this is a check for sequence number looping */
        extseq->split.lo=rtp->seq_number;
        extseq->split.hi++;
      }
    }

    queue_packet(&session->rtp.tev_rq,session->rtp.max_rq_size,mp,rtp,&i);
    stats->discarded+=i;
    ortp_global_stats.discarded+=i;
    return RFC5764BIS_RTP;
  }

  /* check for possible payload type change, in order to update accordingly our clock-rate dependant
  parameters */
  if (session->hw_recv_pt!=rtp->paytype){
    PayloadType *pt_old=rtp_profile_get_payload(session->rcv.profile,session->hw_recv_pt);
    PayloadType *pt_new=rtp_profile_get_payload(session->rcv.profile,rtp->paytype);
    if (pt_old!=0 && pt_new!=0){
      int pt_new_rate = pt_new->clock_rate;
      int pt_old_rate = pt_old->clock_rate;
      if (strcasecmp("g722", pt_new->mime_type)==0)
        pt_new_rate=8000;
      if (strcasecmp("g722", pt_old->mime_type)==0)
        pt_old_rate=8000;
      session->rtp.rcv_last_app_ts = (session->rtp.rcv_last_app_ts*(pt_new_rate/1000)/(pt_old_rate/1000));
    }
    if (rtp_session_update_payload_type(session,rtp->paytype)<0) {
      ortp_warning("rtp_parse: drop RTP packet - unsupported payload %i", rtp->paytype);
      freemsg(mp);
      stats->bad++;
      ortp_global_stats.bad++;
      return RFC5764BIS_BAD;
    }
    rtp_session_resync(session);
    jitter_control_init(&session->rtp.jittctl,-1,pt_new);
  }

  session->rtp.hwrcv_since_last_SR++;
  /* update hwrcv_extseq */
  {
    poly32_t *extseq=(poly32_t*)&rtpstream->hwrcv_extseq;
    if (rtp->seq_number>extseq->split.lo){
      extseq->split.lo=rtp->seq_number;
    }else if (rtp->seq_number<200 && extseq->split.lo>((1<<16) - 200)){
      /* this is a check for sequence number looping */
      extseq->split.lo=rtp->seq_number;
      extseq->split.hi++;
    }
  }


  jitter_control_new_packet(&session->rtp.jittctl,rtp->timestamp,session->rtp.rcv_last_app_ts);
  if (session->rtp.jittctl.adaptive){
    uint16_t tmp_mark = (uint16_t) MAX(session->rtp.jittctl.jitt_comp_ts,2*session->rtp.jittctl.jitter);
    if (session->rtp.extended_stats.djb_nominal_delay_hwmark<tmp_mark)
      session->rtp.extended_stats.djb_nominal_delay_hwmark = tmp_mark;
    if (session->rtp.extended_stats.djb_nominal_delay_hwmark>tmp_mark || session->rtp.extended_stats.djb_nominal_delay_hwmark==0)
      session->rtp.extended_stats.djb_nominal_delay_hwmark = tmp_mark;
    session->rtp.extended_stats.djb_nominal_delay = session->rtp.jittctl.adapt_jitt_comp_ts;
    session->rtp.extended_stats.djb_max_delay = session->rtp.jittctl.adapt_jitt_comp_ts + 20;
  }


  if (session->flags & RTP_SESSION_FIRST_PACKET_DELIVERED) {
    /* detect timestamp important jumps in the future, to workaround stupid rtp senders */
    if (RTP_TIMESTAMP_IS_NEWER_THAN(rtp->timestamp,session->rtp.rcv_last_ts+session->rtp.ts_jump)){
      ortp_debug("rtp_parse: timestamp jump ?");
      rtp_signal_table_emit2(&session->on_timestamp_jump,(uintptr_t)&rtp->timestamp);
    }
    else if (RTP_TIMESTAMP_IS_STRICTLY_NEWER_THAN(session->rtp.rcv_last_ts,rtp->timestamp)){
      /* don't queue packets older than the last returned packet to the application*/
      /* Call timstamp jumb in case of
      * large negative Ts jump or if ts is set to 0
      */

      if ( RTP_TIMESTAMP_IS_STRICTLY_NEWER_THAN(session->rtp.rcv_last_ts, rtp->timestamp + session->rtp.ts_jump) ){
        ortp_warning("rtp_parse: negative timestamp jump");
        rtp_signal_table_emit2(&session->on_timestamp_jump, (uintptr_t)&rtp->timestamp);
      }
      ortp_warning("rtp_parse: discarding too old packet (ts=%i)",rtp->timestamp);
      freemsg(mp);
      stats->outoftime++;
      ortp_global_stats.outoftime++;
      rtp_session_update_voip_metrics(session,0,0,1);
      return RFC5764BIS_RTP;
    }
  }

  if (queue_packet(&session->rtp.rq,session->rtp.max_rq_size,mp,rtp,&i))
    jitter_control_update_average_buffer_size(&session->rtp.jittctl, qlast(&session->rtp.rq), qbegin(&session->rtp.rq));

  stats->discarded+=i;
  ortp_global_stats.discarded+=i;
  return RFC5764BIS_RTP;
}

