/*
The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#if defined(WIN32) || defined(_WIN32_WCE)
#include "ortp-config-win32.h"
#elif HAVE_CONFIG_H
#include "ortp-config.h"
#endif

#if defined(__linux) || defined(__APPLE__)
#include <inttypes.h>
#endif

#include "ortp/ortp.h"
#include "ortp/telephonyevents.h"
#include "ortp/rtcp.h"
#include "jitterctl.h"
#include "utils.h"
#include "rtpsession_priv.h"

#define ENABLE_RTP_SCHEDULER
#if (_WIN32_WINNT >= 0x0600)
#define ENABLE_RTP_QOS
#if (_MSC_VER >= 1700) && !defined(_USING_V110_SDK71_)
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#undef ENABLE_RTP_SCHEDULER
#undef ENABLE_RTP_QOS
#endif
#endif
#endif

#ifdef ENABLE_RTP_QOS
#include <delayimp.h>
#undef ExternC /* avoid redefinition... */
#include <QOS2.h>
#endif

#if defined(WIN32) || defined(_WIN32_WCE)
#ifndef _WIN32_WCE
#include <errno.h>
#endif
#ifdef EWOULDBLOCK
#define is_would_block_error(errnum)	(errnum==WSAEWOULDBLOCK || errnum==EWOULDBLOCK)
#else
#define is_would_block_error(errnum)	(errnum==WSAEWOULDBLOCK)
#endif
#define sockisvalid(s) (s != ~0)
#else
#define is_would_block_error(errnum)	(errnum==EWOULDBLOCK || errnum==EAGAIN)
#define sockisvalid(s) ((s) >= 0)
#endif

extern mblk_t *rtcp_create_simple_bye_packet(uint32_t ssrc, const char *reason);
extern int rtcp_sr_init(RtpSession *session, char *buf, int size);
extern int rtcp_rr_init(RtpSession *session, char *buf, int size);

extern void tc_qos_delete_flow(RtpSession *session);

/* this function initialize all session parameter's that depend on the payload type */
static void payload_type_changed(RtpSession *session, PayloadType *pt){
  jitter_control_set_payload(&session->rtp.jittctl,pt);
  session->rtp.rtcp_report_snt_interval=RTCP_DEFAULT_REPORT_INTERVAL*pt->clock_rate;
  rtp_session_set_time_jump_limit(session,session->rtp.time_jump);
  if (pt->type==PAYLOAD_VIDEO){
    session->rtp.rtcp_report_snt_interval=RTCP_DEFAULT_REPORT_INTERVAL_VIDEO*pt->clock_rate;
    session->permissive=TRUE;
    ortp_message("Using permissive algorithm");
  }
  else session->permissive=FALSE;
}

extern void rtp_parse(RtpSession *session, mblk_t *mp, uint32_t local_str_ts,
struct sockaddr *addr, socklen_t addrlen);

#define RTP_SEQ_IS_GREATER(seq1,seq2)\
  ((uint16_t)((uint16_t)(seq1) - (uint16_t)(seq2))< (uint16_t)(1<<15))

/* put an rtp packet in queue. It is called by rtp_parse()*/
void rtp_putq(queue_t *q, mblk_t *mp)
{
  mblk_t *tmp;
  rtp_header_t *rtp=(rtp_header_t*)mp->b_rptr,*tmprtp;
  /* insert message block by increasing time stamp order : the last (at the bottom)
  message of the queue is the newest*/
  ortp_debug("rtp_putq(): Enqueuing packet with ts=%i and seq=%i",rtp->timestamp,rtp->seq_number);

  if (qempty(q)) {
    putq(q,mp);
    return;
  }
  tmp=qlast(q);
  /* we look at the queue from bottom to top, because enqueued packets have a better chance
  to be enqueued at the bottom, since there are surely newer */
  while (!qend(q,tmp))
  {
    tmprtp=(rtp_header_t*)tmp->b_rptr;
    ortp_debug("rtp_putq(): Seeing packet with seq=%i",tmprtp->seq_number);

    if (rtp->seq_number == tmprtp->seq_number)
    {
      /* this is a duplicated packet. Don't queue it */
      ortp_debug("rtp_putq: duplicated message.");
      freemsg(mp);
      return;
    }else if (RTP_SEQ_IS_GREATER(rtp->seq_number,tmprtp->seq_number)){

      insq(q,tmp->b_next,mp);
      return;
    }
    tmp=tmp->b_prev;
  }
  /* this packet is the oldest, it has to be 
  placed on top of the queue */
  insq(q,qfirst(q),mp);

}



mblk_t *rtp_getq(RtpSession *session, queue_t *q,uint32_t timestamp, int *rejected, mblk_t **last_dropped)
{
  mblk_t *tmp,*ret=NULL,*old=NULL;
  rtp_header_t *tmprtp;
  uint32_t ts_found=0;

  *last_dropped=NULL;
  *rejected=0;
  /* ortp_debug("rtp_getq(): Timestamp %i wanted.",timestamp); */
  if (qempty(q))
  {
    /*ortp_debug("rtp_getq: q is empty.");*/
    return NULL;
  }

  /* return the packet with ts just equal or older than the asked timestamp */
  /* packets with older timestamps are discarded */
  while ((tmp=qfirst(q))!=NULL)
  {
    tmprtp=(rtp_header_t*)tmp->b_rptr;
    /* ortp_debug("rtp_getq: Seeing packet with ts=%i",tmprtp->timestamp); */
    if ( RTP_TIMESTAMP_IS_NEWER_THAN(timestamp,tmprtp->timestamp) )
    {
      if (ret!=NULL && tmprtp->timestamp==ts_found) {
        /* we've found two packets with same timestamp. return the first one */
        break;
      }
      if (ret!=NULL && ((tmprtp->timestamp-ts_found)<=80)) { /* hack for ptime lower than interval */
        /* we've found two packets with same timestamp. return the first one */
        break;
      }
      if (old!=NULL) {
        if (session->rtp.jittctl.adaptive){
          /* only update correction offset between packets of different timestamps*/
          if (ts_found!=session->rtp.rcv_last_ts)
            jitter_control_update_corrective_slide(&session->rtp.jittctl);
        }
        session->rtp.rcv_last_ts = ts_found;
        if (*last_dropped!=NULL) {
          ortp_message("rtp_getq: discarding too old packet with cseq=%u ts=%u", rtp_get_seqnumber(*last_dropped), rtp_get_timestamp(*last_dropped));
          freemsg(*last_dropped);
          (*rejected)++;
        }
        *last_dropped=old;
      }
      ret=getq(q); /* dequeue the packet, since it has an interesting timestamp*/
      ts_found=tmprtp->timestamp;
      ortp_debug("rtp_getq: Found packet with ts=%u",tmprtp->timestamp);
      old=ret;
    }
    else
    {
      break;
    }
  }
  return ret;
}

static mblk_t *rtp_getq_permissive(queue_t *q,uint32_t timestamp, int *rejected)
{
  mblk_t *tmp,*ret=NULL;
  rtp_header_t *tmprtp;

  *rejected=0;
  ortp_debug("rtp_getq_permissive(): Timestamp %i wanted.",timestamp);

  if (qempty(q))
  {
    /*ortp_debug("rtp_getq: q is empty.");*/
    return NULL;
  }
  /* return the packet with the older timestamp (provided that it is older than
  the asked timestamp) */
  tmp=qfirst(q);
  tmprtp=(rtp_header_t*)tmp->b_rptr;
  ortp_debug("rtp_getq_permissive: Seeing packet with ts=%i",tmprtp->timestamp);
  if ( RTP_TIMESTAMP_IS_NEWER_THAN(timestamp,tmprtp->timestamp) )
  {
    ret=getq(q); /* dequeue the packet, since it has an interesting timestamp*/
    ortp_debug("rtp_getq_permissive: Found packet with ts=%i",tmprtp->timestamp);
  }
  return ret;
}


void
  rtp_session_init (RtpSession * session, int mode)
{
  JBParameters jbp;
  if (session == NULL) 
  {
    ortp_debug("rtp_session_init: Invalid paramter (session=NULL)");
    return;
  }
  memset (session, 0, sizeof (RtpSession));
  session->mode = (RtpSessionMode) mode;
  if ((mode == RTP_SESSION_RECVONLY) || (mode == RTP_SESSION_SENDRECV))
  {
    rtp_session_set_flag (session, RTP_SESSION_RECV_SYNC);
    rtp_session_set_flag (session, RTP_SESSION_RECV_NOT_STARTED);

  }
  if ((mode == RTP_SESSION_SENDONLY) || (mode == RTP_SESSION_SENDRECV))
  {
    rtp_session_set_flag (session, RTP_SESSION_SEND_NOT_STARTED);
    session->snd.ssrc=ortp_random();
    /* set default source description */
    rtp_session_set_source_description(session,"unknown@unknown",NULL,NULL,
      NULL,NULL,"oRTP-" ORTP_VERSION,"This is free sofware (LGPL) !");
  }
  session->rtp.snd_seq = 1; /* start with 1 */
  session->snd.telephone_events_pt=-1;	/* not defined a priori */
  session->rcv.telephone_events_pt=-1;	/* not defined a priori */
  rtp_session_set_profile (session, &av_profile); /*the default profile to work with */
  memset(session->rtp.socketv4, -1, sizeof(ortp_socket_t)*10);
  memset(session->rtcp.socketv4, -1, sizeof(ortp_socket_t)*10);
#ifndef WIN32
  session->rtp.snd_socket_size=0;	/*use OS default value unless on windows where they are definitely too short*/
  session->rtp.rcv_socket_size=0;
#else
  session->rtp.snd_socket_size=session->rtp.rcv_socket_size=65536;
#endif
  session->rtp.ssrc_changed_thres=50;
  session->rtp.rtcp_report_snt_interval = RTCP_DEFAULT_REPORT_INTERVAL*90000; //needs to be initialized - if we don't receive anyting.
  session->dscp=RTP_DEFAULT_DSCP;
  session->multicast_ttl=RTP_DEFAULT_MULTICAST_TTL;
  session->multicast_loopback=RTP_DEFAULT_MULTICAST_LOOPBACK;
  qinit(&session->rtp.rq);
  qinit(&session->rtp.tev_rq);
  qinit(&session->contributing_sources);
  session->eventqs=NULL;
  /* init signal tables */
  rtp_signal_table_init (&session->on_ssrc_changed, session,"ssrc_changed");
  rtp_signal_table_init (&session->on_payload_type_changed, session,"payload_type_changed");
  rtp_signal_table_init (&session->on_telephone_event, session,"telephone-event");
  rtp_signal_table_init (&session->on_telephone_event_packet, session,"telephone-event_packet");
  rtp_signal_table_init (&session->on_timestamp_jump,session,"timestamp_jump");
  rtp_signal_table_init (&session->on_network_error,session,"network_error");
  rtp_signal_table_init (&session->on_rtcp_bye,session,"rtcp_bye");
  /*defaults send payload type to 0 (pcmu)*/
  rtp_session_set_send_payload_type(session,0);
  /*sets supposed recv payload type to undefined */
  rtp_session_set_recv_payload_type(session,-1);
  /* configure jitter buffer with working default parameters */
  jbp.min_size=RTP_DEFAULT_JITTER_TIME;
  jbp.nom_size=RTP_DEFAULT_JITTER_TIME;
  jbp.max_size=-1;
  jbp.max_packets= 100;/* maximum number of packet allowed to be queued */
  jbp.adaptive=TRUE;
  rtp_session_enable_jitter_buffer(session,TRUE);
  rtp_session_set_jitter_buffer_params(session,&jbp);
  rtp_session_set_time_jump_limit(session,5000);
  rtp_session_enable_rtcp(session,TRUE);
  session->recv_buf_size = UDP_MAX_SIZE;
  session->symmetric_rtp = FALSE;
  session->permissive=FALSE;
  session->rtcp_mux=FALSE;
  msgb_allocator_init(&session->allocator);
  session->FEC_N_interval=0;
}


/**
* Creates a new rtp session.
* If the session is able to send data (RTP_SESSION_SENDONLY or
* RTP_SESSION_SENDRECV), then a random SSRC number is choosed for 
* the outgoing stream.
* @param mode One of the RtpSessionMode flags.	
*
* @return the newly created rtp session.
**/
RtpSession *
  rtp_session_new (int mode)
{
  RtpSession *session;
  session = (RtpSession *) ortp_malloc (sizeof (RtpSession));
  if (session == NULL)
  {
    ortp_error("rtp_session_new: Memory allocation failed");
    return NULL;
  }
  rtp_session_init (session, mode);
  return session;
}

/**
* Enable bundle RTP read on socket.
*
*@param session a rtp session.
*
**/
void rtp_session_enable_sendrecv(RtpSession *session) {
  rtp_session_set_flag (session, RTP_SESSION_ENABLE_SENDRECV);
}

/**
* Activate bundle RTP read on socket.
*
*@param session a rtp session.
*
**/
void rtp_session_disable_sendrecv(RtpSession *session) {
  rtp_session_unset_flag (session, RTP_SESSION_ENABLE_SENDRECV);
}

/**
* Set bundle for this rtp stream.
* If the bundle is set, the first session of the bundle is the one to use.
* If the bundle is set, a thread bundle will read the RTP data.
*
*@param session a rtp session.
*@param bundle  pointer to a bundle for several media.
*
**/void
rtp_session_set_bundle (RtpSession * session, struct _RtpBundle *bundle)
{
  session->bundle = bundle;
}

/**
* Deprecated.
*
*@param session a rtp session.
*@param yesno 	a boolean to indicate the scheduling mode.
*
*
**/
void
  rtp_session_set_scheduling_mode (RtpSession * session, int yesno)
{
  /* removed */
}

/**
*	Deprecated.
*
*  @param session a rtp session
*  @param yesno a boolean
**/
void
  rtp_session_set_blocking_mode (RtpSession * session, int yesno)
{
  /* removed */
}

/**
*	Set the RTP profile to be used for the session. By default, all session are created by
*	rtp_session_new() are initialized with the AV profile, as defined in RFC 3551. The application
*	can set any other profile instead using that function.
*
* @param session a rtp session
* @param profile a rtp profile
**/

void
  rtp_session_set_profile (RtpSession * session, RtpProfile * profile)
{
  session->snd.profile = profile;
  session->rcv.profile = profile;
  rtp_session_telephone_events_supported(session);
}

/**
*	By default oRTP automatically sends RTCP SR or RR packets. If
*	yesno is set to FALSE, the RTCP sending of packet is disabled.
*	This functionnality might be needed for some equipments that do not
*	support RTCP, leading to a traffic of ICMP errors on the network.
*	It can also be used to save bandwidth despite the RTCP bandwidth is 
*	actually and usually very very low.
**/
void rtp_session_enable_rtcp(RtpSession *session, bool_t yesno){
  session->rtcp.enabled=yesno;
}

/**
*	Set the RTP profile to be used for the sending by this session. By default, all session are created by
*	rtp_session_new() are initialized with the AV profile, as defined in RFC 3551. The application
*	can set any other profile instead using that function.
* @param session a rtp session
* @param profile a rtp profile
*
**/

void
  rtp_session_set_send_profile (RtpSession * session, RtpProfile * profile)
{
  session->snd.profile = profile;
  rtp_session_send_telephone_events_supported(session);
}



/**
*	Set the RTP profile to be used for the receiveing by this session. By default, all session are created by
*	rtp_session_new() are initialized with the AV profile, as defined in RFC 3551. The application
*	can set any other profile instead using that function.
*
* @param session a rtp session
* @param profile a rtp profile
**/

void
  rtp_session_set_recv_profile (RtpSession * session, RtpProfile * profile)
{
  session->rcv.profile = profile;
  rtp_session_recv_telephone_events_supported(session);
}

/**
*@param session a rtp session
*
*	DEPRECATED! Returns current send profile.
*	Use rtp_session_get_send_profile() or rtp_session_get_recv_profile()
*
**/
RtpProfile *rtp_session_get_profile(RtpSession *session){
  return session->snd.profile;
}


/**
*@param session a rtp session
*
*	Returns current send profile.
*
**/
RtpProfile *rtp_session_get_send_profile(RtpSession *session){
  return session->snd.profile;
}

/**
*@param session a rtp session
*
*	Returns current receive profile.
*
**/
RtpProfile *rtp_session_get_recv_profile(RtpSession *session){
  return session->rcv.profile;
}

/**
*	The default value is UDP_MAX_SIZE bytes, a value which is working for mostly everyone.
*	However if your application can make assumption on the sizes of received packet,
*	it can be interesting to set it to a lower value in order to save memory.
*
* @param session a rtp session
* @param bufsize max size in bytes for receiving packets
**/
void rtp_session_set_recv_buf_size(RtpSession *session, int bufsize){
  session->recv_buf_size=bufsize;
}

/**
*	Set kernel send maximum buffer size for the rtp socket.
*	A value of zero defaults to the operating system default.
**/
void rtp_session_set_rtp_socket_send_buffer_size(RtpSession * session, unsigned int size){
  session->rtp.snd_socket_size=size;
}

/**
*	Set kernel recv maximum buffer size for the rtp socket.
*	A value of zero defaults to the operating system default.
**/
void rtp_session_set_rtp_socket_recv_buffer_size(RtpSession * session, unsigned int size){
  session->rtp.rcv_socket_size=size;
}

/**
*	This function provides the way for an application to be informed of various events that
*	may occur during a rtp session. @signal is a string identifying the event, and @cb is 
*	a user supplied function in charge of processing it. The application can register
*	several callbacks for the same signal, in the limit of #RTP_CALLBACK_TABLE_MAX_ENTRIES.
*	Here are name and meaning of supported signals types:
*
*	"ssrc_changed" : the SSRC of the incoming stream has changed.
*
*	"payload_type_changed" : the payload type of the incoming stream has changed.
*
*	"telephone-event_packet" : a telephone-event rtp packet (RFC2833) is received.
*
*	"telephone-event" : a telephone event has occured. This is a high-level shortcut for "telephone-event_packet".
*
*	"network_error" : a network error happened on a socket. Arguments of the callback functions are
*						a const char * explaining the error, an int errno error code and the user_data as usual.
*
*	"timestamp_jump" : we have received a packet with timestamp in far future compared to last timestamp received.
*						The farness of far future is set by rtp_sesssion_set_time_jump_limit()
*  "rtcp_bye": we have received a RTCP bye packet. Arguments of the callback
*              functions are a const char * containing the leaving reason and
*              the user_data.
* 
*	Returns: 0 on success, -EOPNOTSUPP if the signal does not exists, -1 if no more callbacks
*	can be assigned to the signal type.
*
* @param session 	a rtp session
* @param signal_name	the name of a signal
* @param cb		a RtpCallback
* @param user_data	a pointer to any data to be passed when invoking the callback.
*
**/
int
  rtp_session_signal_connect (RtpSession * session, const char *signal_name,
  RtpCallback cb, uintptr_t user_data)
{
  OList *elem;
  for (elem=session->signal_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpSignalTable *s=(RtpSignalTable*) elem->data;
    if (strcmp(signal_name,s->signal_name)==0){
      return rtp_signal_table_add(s,cb,user_data);
    }
  }
  ortp_warning ("rtp_session_signal_connect: inexistant signal %s",signal_name);
  return -1;
}


/**
*	Removes callback function @cb to the list of callbacks for signal @signal.
*
* @param session a rtp session
* @param signal_name	a signal name
* @param cb	a callback function.
* @return: 0 on success, a negative value if the callback was not found.
**/
int
  rtp_session_signal_disconnect_by_callback (RtpSession * session, const char *signal_name,
  RtpCallback cb)
{
  OList *elem;
  for (elem=session->signal_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpSignalTable *s=(RtpSignalTable*) elem->data;
    if (strcmp(signal_name,s->signal_name)==0){
      return rtp_signal_table_remove_by_callback(s,cb);
    }
  }
  ortp_warning ("rtp_session_signal_connect: inexistant signal %s",signal_name);
  return -1;
}


/**
* sets the initial sequence number of a sending session.
* @param session		a rtp session freshly created.
* @param addr			a 16 bit unsigned number.
*
**/
void rtp_session_set_seq_number(RtpSession *session, uint16_t seq){
  session->rtp.snd_seq=seq;
}


uint16_t rtp_session_get_seq_number(RtpSession *session){
  return session->rtp.snd_seq;
}


/**
*	Sets the SSRC for the outgoing stream.
*  If not done, a random ssrc is used.
*
* @param session a rtp session.
* @param ssrc an unsigned 32bit integer representing the synchronisation source identifier (SSRC).
**/
void
  rtp_session_set_ssrc (RtpSession * session, uint32_t ssrc)
{
  session->snd.ssrc = ssrc;
}

/**
*	Sets the SSRC for the incoming stream.
*  If not done, first SSRC received will be accepted.
*
* @param session a rtp session.
* @param ssrc an unsigned 32bit integer representing the synchronisation source identifier (SSRC).
**/
void
  rtp_session_set_recv_ssrc (RtpSession * session, uint32_t ssrc)
{
  if (session->rcv.ssrc!=ssrc) {
    RtpStream *rtpstream=&session->rtp;
    rtpstream->hwrcv_extseq=0;
    rtpstream->hwrcv_seq_at_last_SR=0;
    rtpstream->hwrcv_since_last_SR=0;
    rtpstream->last_rcv_SR_time.tv_usec = 0;
    rtpstream->last_rcv_SR_time.tv_sec = 0;
    rtpstream->last_rcv_SR_ts=0;
    session->ssrc_set=FALSE;
  }
  session->rcv.ssrc = ssrc;
}


int rtp_session_update_payload_type(RtpSession *session, int paytype){
  /* check if we support this payload type */
  PayloadType *pt=rtp_profile_get_payload(session->rcv.profile,paytype);
  if (pt!=0){
    session->hw_recv_pt=paytype;
    ortp_message ("payload type changed to %i(%s) !",
      paytype,pt->mime_type);
    payload_type_changed(session,pt);
  }else{
    ortp_warning("Receiving packet with unknown payload type %i.",paytype);
    return -1;
  }
  return 0;
}
/**
*	Sets the payload type of the rtp session. It decides of the payload types written in the
*	of the rtp header for the outgoing stream, if the session is SENDRECV or SENDONLY.
*	For payload type in incoming packets, the application can be informed by registering
*	for the "payload_type_changed" signal, so that it can make the necessary changes
*	on the downstream decoder that deals with the payload of the packets.
*
* @param session a rtp session
* @param paytype the payload type number
* @return 0 on success, -1 if the payload is not defined.
**/

int
  rtp_session_set_send_payload_type (RtpSession * session, int paytype)
{
  session->snd.pt=paytype;
  return 0;
}

/**
*@param session a rtp session
*
*@return the payload type currently used in outgoing rtp packets
**/
int rtp_session_get_send_payload_type(const RtpSession *session){
  return session->snd.pt;
}

/**
*
*	Sets the expected payload type for incoming packets.
*	If the actual payload type in incoming packets is different that this expected payload type, thus
*	the "payload_type_changed" signal is emitted.
*
*@param session a rtp session
*@param paytype the payload type number
*@return 0 on success, -1 if the payload is not defined.
**/

int
  rtp_session_set_recv_payload_type (RtpSession * session, int paytype)
{
  PayloadType *pt;
  session->rcv.pt=paytype;
  session->hw_recv_pt=paytype;
  pt=rtp_profile_get_payload(session->rcv.profile,paytype);
  if (pt!=NULL){
    payload_type_changed(session,pt);
  }
  return 0;
}

/**
*@param session a rtp session
*
* @return the payload type currently used in incoming rtp packets
**/
int rtp_session_get_recv_payload_type(const RtpSession *session){
  return session->rcv.pt;
}

/**
*	Sets the expected payload type for incoming packets and payload type to be used for outgoing packets.
*	If the actual payload type in incoming packets is different that this expected payload type, thus
*	the "payload_type_changed" signal is emitted.
*
* @param session a rtp session
* @param paytype the payload type number
* @return 0 on success, -1 if the payload is not defined.
**/
int rtp_session_set_payload_type(RtpSession *session, int pt){
  if (rtp_session_set_send_payload_type(session,pt)<0) return -1;
  if (rtp_session_set_recv_payload_type(session,pt)<0) return -1;
  return 0;
}


static void rtp_header_init_from_session(rtp_header_t *rtp, RtpSession *session){
  rtp->version = 2;
  rtp->padbit = 0;
  rtp->extbit = 0;
  rtp->markbit= 0;
  rtp->cc = 0;
  rtp->paytype = session->snd.pt;
  rtp->ssrc = session->snd.ssrc;
  rtp->timestamp = 0;	/* set later, when packet is sended */
  /* set a seq number */
  rtp->seq_number=session->rtp.snd_seq;
}

/**
*	Allocates a new rtp packet. In the header, ssrc and payload_type according to the session's
*	context. Timestamp is not set, it will be set when the packet is going to be
*	sent with rtp_session_sendm_with_ts(). Sequence number is initalized to previous sequence number sent + 1
*	If payload_size is zero, thus an empty packet (just a RTP header) is returned.
*
*@param session a rtp session.
*@param header_size the rtp header size. For standart size (without extensions), it is RTP_FIXED_HEADER_SIZE
*@param payload data to be copied into the rtp packet.
*@param payload_size size of data carried by the rtp packet.
*@return a rtp packet in a mblk_t (message block) structure.
**/
mblk_t * rtp_session_create_packet(RtpSession *session,int header_size, const uint8_t *payload, int payload_size)
{
  mblk_t *mp;
  int msglen=header_size+payload_size;
  rtp_header_t *rtp;

  mp=allocb(msglen,0);
  rtp=(rtp_header_t*)mp->b_rptr;
  rtp_header_init_from_session(rtp,session);
  /*copy the payload, if any */
  mp->b_wptr+=header_size;
  if (payload_size){
    memcpy(mp->b_wptr,payload,payload_size);
    mp->b_wptr+=payload_size;
  }
  return mp;
}

/**
*	Creates a new rtp packet using the given payload buffer (no copy). The header will be allocated separetely.
*  In the header, ssrc and payload_type according to the session's
*	context. Timestamp and seq number are not set, there will be set when the packet is going to be
*	sent with rtp_session_sendm_with_ts().
*	oRTP will send this packet using libc's sendmsg() (if this function is availlable!) so that there will be no
*	packet concatenation involving copies to be done in user-space.
*  @freefn can be NULL, in that case payload will be kept untouched.
*
* @param session a rtp session.
* @param payload the data to be sent with this packet
* @param payload_size size of data
* @param freefn a function that will be called when the payload buffer is no more needed.
* @return: a rtp packet in a mblk_t (message block) structure.
**/

mblk_t * rtp_session_create_packet_with_data(RtpSession *session, uint8_t *payload, int payload_size, void (*freefn)(void*))
{
  mblk_t *mp,*mpayload;
  int header_size=RTP_FIXED_HEADER_SIZE; /* revisit when support for csrc is done */
  rtp_header_t *rtp;

  mp=allocb(header_size,0);
  rtp=(rtp_header_t*)mp->b_rptr;
  rtp_header_init_from_session(rtp,session);
  mp->b_wptr+=header_size;
  /* create a mblk_t around the user supplied payload buffer */
  mpayload=esballoc(payload,payload_size,0,freefn);
  mpayload->b_wptr+=payload_size;
  /* link it with the header */
  mp->b_cont=mpayload;
  return mp;
}


/**
* Creates a new rtp packet using the buffer given in arguments (no copy). 
* In the header, ssrc and payload_type according to the session's
*context. Timestamp and seq number are not set, there will be set when the packet is going to be
*	sent with rtp_session_sendm_with_ts().
*  @freefn can be NULL, in that case payload will be kept untouched.
*
* @param session a rtp session.
* @param buffer a buffer that contains first just enough place to write a RTP header, then the data to send.
* @param size the size of the buffer
* @param freefn a function that will be called once the buffer is no more needed (the data has been sent).
* @return a rtp packet in a mblk_t (message block) structure.
**/
mblk_t * rtp_session_create_packet_in_place(RtpSession *session,uint8_t *buffer, int size, void (*freefn)(void*) )
{
  mblk_t *mp;
  rtp_header_t *rtp;

  mp=esballoc(buffer,size,0,freefn);

  rtp=(rtp_header_t*)mp->b_rptr;
  rtp_header_init_from_session(rtp,session);
  return mp;
}


int
  __rtp_session_sendm_with_ts (RtpSession * session, mblk_t *mp, uint32_t packet_ts, uint32_t send_ts)
{
  rtp_header_t *rtp;
  int error = 0;
  int packsize;
  RtpStream *stream=&session->rtp;

  if (session->flags & RTP_SESSION_SEND_NOT_STARTED)
  {
    //session->rtp.snd_ts_offset = send_ts;
    /* Set initial last_rcv_time to first send time. */
    if ((session->flags & RTP_SESSION_RECV_NOT_STARTED)
      || session->mode == RTP_SESSION_SENDONLY)
    {
      gettimeofday(&session->last_recv_time, NULL);
    }
    rtp_session_unset_flag (session,RTP_SESSION_SEND_NOT_STARTED);
  }

  if(mp==NULL) {/*for people who just want to be blocked but
                do not want to send anything.*/
    session->rtp.snd_last_ts = packet_ts;
    return 0;
  }

  rtp=(rtp_header_t*)mp->b_rptr;

  packsize = msgdsize(mp) ;

  rtp->timestamp=packet_ts;
  if (session->snd.telephone_events_pt==rtp->paytype)
  {
    rtp->seq_number = session->rtp.snd_seq;
    session->rtp.snd_seq++;
  }
  else
    session->rtp.snd_seq=rtp->seq_number+1;
  session->rtp.snd_last_ts = packet_ts;


  error = rtp_session_rtp_send (session, mp);
  if (error>0) {
    ortp_global_stats.sent += packsize;
    stream->sent_payload_bytes+=packsize-RTP_FIXED_HEADER_SIZE;
    stream->stats.sent += packsize;
    ortp_global_stats.packet_sent++;
    stream->stats.packet_sent++;
  }
  /*send RTCP packet if needed */
  rtp_session_rtcp_process_send(session);
  /* receives rtcp packet if session is send-only*/
  /*otherwise it is done in rtp_session_recvm_with_ts */
  if (session->mode==RTP_SESSION_SENDONLY) {
    int i;
    for (i=0;i<10;i++) {
      rtp_session_rtcp_recv(session, i);
    }
  }
  return error;
}

/**
*	Send the rtp datagram @mp to the destination set by rtp_session_set_remote_addr() 
*	with timestamp @timestamp. For audio data, the timestamp is the number
*	of the first sample resulting of the data transmitted. See rfc1889 for details.
*  The packet (@mp) is freed once it is sended.
*
*@param session a rtp session.
*@param mp a rtp packet presented as a mblk_t.
*@param timestamp the timestamp of the data to be sent.
* @return the number of bytes sent over the network.
**/

int rtp_session_sendm_with_ts(RtpSession *session, mblk_t *packet, uint32_t timestamp){
  return __rtp_session_sendm_with_ts(session,packet,timestamp,timestamp);
}




/**
*	Send a rtp datagram to the destination set by rtp_session_set_remote_addr() containing
*	the data from @buffer with timestamp @userts. This is a high level function that uses
*	rtp_session_create_packet() and rtp_session_sendm_with_ts() to send the data.
*
*@param session a rtp session.
*@param buffer a buffer containing the data to be sent in a rtp packet.
*@param len the length of the data buffer, in bytes.
*@param userts	the timestamp of the data to be sent. Refer to the rfc to know what it is.
*
*@param return the number of bytes sent over the network.
**/
int
  rtp_session_send_with_ts (RtpSession * session, const uint8_t * buffer, int len,
  uint32_t userts)
{
  mblk_t *m;
  int err;
#ifdef USE_SENDMSG
  m=rtp_session_create_packet_with_data(session,(uint8_t*)buffer,len,NULL);
#else
  m = rtp_session_create_packet(session,RTP_FIXED_HEADER_SIZE,(uint8_t*)buffer,len);
#endif
  err=rtp_session_sendm_with_ts(session,m,userts);
  return err;
}



extern void rtcp_parse(RtpSession *session, mblk_t *mp);



static void payload_type_changed_notify(RtpSession *session, int paytype){
  PayloadType *pt = rtp_profile_get_payload(session->rcv.profile,paytype);
  if (pt) {
    session->rcv.pt = paytype;
    rtp_signal_table_emit (&session->on_payload_type_changed);
  }
}


/**
*	Try to get a rtp packet presented as a mblk_t structure from the rtp session.
*	The @user_ts parameter is relative to the first timestamp of the incoming stream. In other
*	words, the application does not have to know the first timestamp of the stream, it can
*	simply call for the first time this function with @user_ts=0, and then incrementing it
*	as it want. The RtpSession takes care of synchronisation between the stream timestamp
*	and the user timestamp given here.
*
*	This function returns the entire packet (with header).
*
*	The behaviour of this function has changed since version 0.15.0. Previously the payload data could be 
*	accessed using  mblk_t::b_cont::b_rptr field of the returned mblk_t.
*	This is no more the case.
*	The convenient way of accessing the payload data is to use rtp_get_payload() :
*	@code
*	unsigned char *payload;
*	int payload_size;
*	payload_size=rtp_get_payload(mp,&payload);
*	@endcode
*	OR simply skip the header this way, the data is then comprised between mp->b_rptr and mp->b_wptr:
*	@code
*	rtp_get_payload(mp,&mp->b_rptr);
*	@endcode
*
*
* @param session a rtp session.
* @param user_ts a timestamp.
*
* @return a rtp packet presented as a mblk_t.
**/
mblk_t *
  rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts)
{
  mblk_t *last_dropped=NULL;
  mblk_t *m;
  m = rtp_session_recvm_with_ts_extended (session, user_ts, &last_dropped);
  freemsg(last_dropped);
  return m;
}

/**
*	Try to get a rtp packet presented as a mblk_t structure from the rtp session.
*	The @user_ts parameter is relative to the first timestamp of the incoming stream. In other
*	words, the application does not have to know the first timestamp of the stream, it can
*	simply call for the first time this function with @user_ts=0, and then incrementing it
*	as it want. The RtpSession takes care of synchronisation between the stream timestamp
*	and the user timestamp given here.
*
*	This function returns the entire packet (with header).
*
*	The behaviour of this function has changed since version 0.15.0. Previously the payload data could be 
*	accessed using  mblk_t::b_cont::b_rptr field of the returned mblk_t.
*	This is no more the case.
*	The convenient way of accessing the payload data is to use rtp_get_payload() :
*	@code
*	unsigned char *payload;
*	int payload_size;
*	payload_size=rtp_get_payload(mp,&payload);
*	@endcode
*	OR simply skip the header this way, the data is then comprised between mp->b_rptr and mp->b_wptr:
*	@code
*	rtp_get_payload(mp,&mp->b_rptr);
*	@endcode
*
*
* @param session a rtp session.
* @param user_ts a timestamp.
* @param last_dropped last packet dropped by jitter.
*
* @return a rtp packet presented as a mblk_t.
**/
mblk_t *
  rtp_session_recvm_with_ts_extended (RtpSession * session, uint32_t user_ts, mblk_t **last_dropped)
{
  mblk_t *mp = NULL;
  rtp_header_t *rtp;
  uint32_t ts;
  RtpStream *stream=&session->rtp;
  int rejected=0;
  bool_t read_socket=TRUE;

  *last_dropped=NULL;
  /* if we are scheduled, remember the scheduler time at which the application has
  * asked for its first timestamp */

  if (session->flags & RTP_SESSION_RECV_NOT_STARTED)
  {
    //session->rtp.rcv_query_ts_offset = user_ts;
    /* Set initial last_rcv_time to first recv time. */
    if ((session->flags & RTP_SESSION_SEND_NOT_STARTED)
      || session->mode == RTP_SESSION_RECVONLY){
        gettimeofday(&session->last_recv_time, NULL);
    }
    rtp_session_unset_flag (session,RTP_SESSION_RECV_NOT_STARTED);
  }else{
    /*prevent reading from the sockets when two 
    consecutives calls for a same timestamp*/
    if (user_ts==session->rtp.rcv_last_app_ts)
      read_socket=FALSE;
  }
  session->rtp.rcv_last_app_ts = user_ts;
  if (read_socket){
    if (session->bundle==NULL) {
      int i;
      for (i=0;i<10;i++) {
        rtp_session_rtp_recv (session, i);
        rtp_session_rtcp_recv(session, i);
      }
    }
  }
  /* check for telephone event first */
  /* LOCK HERE? */
  mp=getq(&session->rtp.tev_rq);
  if (mp!=NULL){
    int msgsize=msgdsize(mp);
    ortp_global_stats.recv += msgsize;
    stream->stats.recv += msgsize;
    rtp_signal_table_emit2(&session->on_telephone_event_packet,(uintptr_t)mp);
    rtp_session_check_telephone_events(session,mp);
    freemsg(mp);
    mp=NULL;
  }

  /* then now try to return a media packet, if possible */
  /* first condition: if the session is starting, don't return anything
  * until the queue size reaches jitt_comp */

  if (session->flags & RTP_SESSION_RECV_SYNC)
  {
    queue_t *q = &session->rtp.rq;
    if (qempty(q))
    {
      ortp_debug ("Queue is empty.");
      goto end;
    }
    rtp = (rtp_header_t *) qfirst(q)->b_rptr;

    if (session->rcv.pt != rtp->paytype)
    {
      payload_type_changed_notify(session, rtp->paytype);
    }

    //session->rtp.rcv_ts_offset = rtp->timestamp;
    //session->rtp.rcv_last_ret_ts = session->rtp.rcv_last_app_ts;	/* just to have an init value */
    session->rcv.ssrc = rtp->ssrc;
    /* delete the recv synchronisation flag */
    rtp_session_unset_flag (session, RTP_SESSION_RECV_SYNC);
  }

  /*calculate the stream timestamp from the user timestamp */
  ts = jitter_control_get_compensated_timestamp(&session->rtp.jittctl,session->rtp.rcv_last_app_ts);
  if (session->rtp.jittctl.enabled==TRUE){
    if (session->permissive)
      mp = rtp_getq_permissive(&session->rtp.rq, ts,&rejected);
    else{
      mp = rtp_getq(session, &session->rtp.rq, ts,&rejected, last_dropped);
    }
  }else mp=getq(&session->rtp.rq);/*no jitter buffer at all*/

  /* update gap and burst voip-metrics */
  rtp_session_update_voip_metrics(session, (mp!=NULL)?1:0, 0, rejected);
  
  stream->stats.outoftime+=rejected;
  ortp_global_stats.outoftime+=rejected;

  goto end;

end:
  if (mp != NULL)
  {
    int msgsize = msgdsize (mp);	/* evaluate how much bytes (including header) is received by app */
    uint32_t packet_ts;
    ortp_global_stats.recv += msgsize;
    stream->stats.recv += msgsize;
    rtp = (rtp_header_t *) mp->b_rptr;
    packet_ts=rtp->timestamp;
    ortp_debug("Returning mp with ts=%i", packet_ts);
    /* check for payload type changes */
    if (session->rcv.pt != rtp->paytype)
    {
      payload_type_changed_notify(session, rtp->paytype);
    }
    /* update the packet's timestamp so that it corrected by the 
    adaptive jitter buffer mechanism */
    if (session->rtp.jittctl.adaptive){
      /* only update correction offset between packets of different timestamps*/
      if (packet_ts!=session->rtp.rcv_last_ts)
        jitter_control_update_corrective_slide(&session->rtp.jittctl);
      /* AMD: DO NOT UPDATE WITH COMPENSATED TIMESTANP changed_ts=packet_ts+session->rtp.jittctl.corrective_slide; */
      /* AMD: DO NOT UPDATE WITH COMPENSATED TIMESTANP rtp->timestamp=changed_ts; */
      /*ortp_debug("Returned packet has timestamp %u, with clock slide compensated it is %u (queusize=%i)",packet_ts,rtp->timestamp, session->rtp.rq.q_mcount); */
    }
    session->rtp.rcv_last_ts = packet_ts;
    if (!(session->flags & RTP_SESSION_FIRST_PACKET_DELIVERED)){
      rtp_session_set_flag(session,RTP_SESSION_FIRST_PACKET_DELIVERED);
    }
  }
  else
  {
    ortp_debug ("No mp for timestamp queried");
    stream->stats.unavaillable++;
    ortp_global_stats.unavaillable++;
  }
  /* LOCK HERE? */
  rtp_session_rtcp_process_recv(session);

  return mp;
}


/**
*	NOTE: use of this function is discouraged when sending payloads other than
*	pcm/pcmu/pcma/adpcm types.
*	rtp_session_recvm_with_ts() does better job.
*
*	Tries to read the bytes of the incoming rtp stream related to timestamp ts. In case 
*	where the user supplied buffer @buffer is not large enough to get all the data 
*	related to timestamp ts, then *( have_more) is set to 1 to indicate that the application
*	should recall the function with the same timestamp to get more data.
*	
*  When the rtp session is scheduled (see rtp_session_set_scheduling_mode() ), and the 
*	blocking mode is on (see rtp_session_set_blocking_mode() ), then the calling thread
*	is suspended until the timestamp given as argument expires, whatever a received packet 
*	fits the query or not.
*
*	Important note: it is clear that the application cannot know the timestamp of the first
*	packet of the incoming stream, because it can be random. The @ts timestamp given to the
*	function is used relatively to first timestamp of the stream. In simple words, 0 is a good
*	value to start calling this function.
*
*	This function internally calls rtp_session_recvm_with_ts() to get a rtp packet. The content
*	of this packet is then copied into the user supplied buffer in an intelligent manner:
*	the function takes care of the size of the supplied buffer and the timestamp given in  
*	argument. Using this function it is possible to read continous audio data (e.g. pcma,pcmu...)
*	with for example a standart buffer of size of 160 with timestamp incrementing by 160 while the incoming
*	stream has a different packet size.
*
*Returns: if a packet was availlable with the corresponding timestamp supplied in argument 
*	then the number of bytes written in the user supplied buffer is returned. If no packets
*	are availlable, either because the sender has not started to send the stream, or either
*	because silence packet are not transmitted, or either because the packet was lost during
*	network transport, then the function returns zero.
*@param session a rtp session.
*@param buffer a user supplied buffer to write the data.
*@param len the length in bytes of the user supplied buffer.
*@param ts the timestamp wanted.
*@param have_more the address of an integer to indicate if more data is availlable for the given timestamp.
*
**/
int rtp_session_recv_with_ts (RtpSession * session, uint8_t * buffer,
                              int len, uint32_t ts, int * have_more){
                                mblk_t *mp=NULL;
                                int plen,blen=0;
                                *have_more=0;
                                while(1){
                                  if (session->pending){
                                    mp=session->pending;
                                    session->pending=NULL;
                                  }else {
                                    mp=rtp_session_recvm_with_ts(session,ts);
                                    if (mp!=NULL) rtp_get_payload(mp,&mp->b_rptr);
                                  }
                                  if (mp){
                                    plen=(int)(mp->b_wptr-mp->b_rptr);
                                    if (plen<=len){
                                      memcpy(buffer,mp->b_rptr,plen);
                                      buffer+=plen;
                                      blen+=plen;
                                      len-=plen;
                                      freemsg(mp);
                                      mp=NULL;
                                    }else{
                                      memcpy(buffer,mp->b_rptr,len);
                                      mp->b_rptr+=len;
                                      buffer+=len;
                                      blen+=len;
                                      len=0;
                                      session->pending=mp;
                                      *have_more=1;
                                      break;
                                    }
                                  }else break;
                                }
                                return blen;
}

/**
* oRTP has the possibility to inform the application through a callback registered 
* with rtp_session_signal_connect about crazy incoming RTP stream that jumps from 
* a timestamp N to N+some_crazy_value. This lets the opportunity for the application
* to reset the session in order to resynchronize, or any other action like stopping the call
* and reporting an error.
* @param session the rtp session
* @param ts_step a time interval in miliseconds
*
**/
void rtp_session_set_time_jump_limit(RtpSession *session, int milisecs){
  uint32_t ts;
  session->rtp.time_jump=milisecs;
  ts=rtp_session_time_to_ts(session,milisecs);
  if (ts==0) session->rtp.ts_jump=1<<31;	/* do not detect ts jump */
  else session->rtp.ts_jump=ts;
}

/**
* Closes the rtp and rtcp sockets.
**/
void rtp_session_release_sockets(RtpSession *session){
  int i;
  for (i=0;i<10;i++) {
    if (sockisvalid(session->rtp.socketv4[i])) close_socket (session, session->rtp.socketv4[i]);
    if (sockisvalid(session->rtcp.socketv4[i])) close_socket (session, session->rtcp.socketv4[i]);
  }
  memset(session->rtp.socketv4, -1, sizeof(ortp_socket_t)*10);
  memset(session->rtcp.socketv4, -1, sizeof(ortp_socket_t)*10);

#if 0
  /* we may want only to reset the stream and keep the transport */
  ortp_transport_free(session->rtp.tr);
  ortp_transport_free(session->rtcp.tr);
  session->rtp.tr = 0;
  session->rtcp.tr = 0;
#endif

  /* don't discard remote addresses, then can be preserved for next use.
  session->rtp.rem_addrlen=0;
  session->rtcp.rem_addrlen=0;
  */
  tc_qos_delete_flow(session);
}

/**
* Register an event queue.
* An application can use an event queue to get informed about various RTP events.
**/
void rtp_session_register_event_queue(RtpSession *session, OrtpEvQueue *q){
  session->eventqs=o_list_append(session->eventqs,q);
}

void rtp_session_unregister_event_queue(RtpSession *session, OrtpEvQueue *q){
  session->eventqs=o_list_remove(session->eventqs,q);
}

void rtp_session_dispatch_event(RtpSession *session, OrtpEvent *ev){
  OList *it;
  int i;
  for(i=0,it=session->eventqs;it!=NULL;it=it->next,++i){
    ortp_ev_queue_put((OrtpEvQueue*)it->data,ortp_event_dup(ev));
  }	
  ortp_event_destroy(ev);
}


void rtp_session_uninit (RtpSession * session)
{
  /*flush all queues */
  flushq(&session->rtp.rq, FLUSHALL);
  flushq(&session->rtp.tev_rq, FLUSHALL);

  if (session->eventqs!=NULL) o_list_free(session->eventqs);
  /* close sockets */
  rtp_session_release_sockets(session);

  if (session->current_tev!=NULL) freemsg(session->current_tev);
  if (session->rtp.cached_mp!=NULL) freemsg(session->rtp.cached_mp);
  if (session->rtcp.cached_mp!=NULL) freemsg(session->rtcp.cached_mp);
  if (session->sd!=NULL) freemsg(session->sd);

  session->signal_tables = o_list_free(session->signal_tables);
  msgb_allocator_uninit(&session->allocator);

#ifdef ENABLE_RTP_QOS
  if (session->rtp.QoSFlowID != 0)
  {
    OSVERSIONINFOEX ovi;
    memset(&ovi, 0, sizeof(ovi));
    ovi.dwOSVersionInfoSize = sizeof(ovi);
    GetVersionEx((LPOSVERSIONINFO) & ovi);

    ortp_message("check OS support for qwave.lib: %i %i %i",
      ovi.dwMajorVersion, ovi.dwMinorVersion, ovi.dwBuildNumber);
    if (ovi.dwMajorVersion > 5) {
      HRESULT hr = E_FAIL;
      __try {
        hr = __HrLoadAllImportsForDll("qwave.dll");
      }
      __except(EXCEPTION_EXECUTE_HANDLER) {
        hr = E_FAIL;
      }
      if (! SUCCEEDED(hr)) {
        ortp_warning("Failed to load qwave.dll: no QoS available" );
      }
      else
      {
        BOOL QoSResult;
        QoSResult = QOSRemoveSocketFromFlow(session->rtp.QoSHandle, 
          0, 
          session->rtp.QoSFlowID, 
          0);

        if (QoSResult != TRUE){
          ortp_error("QOSRemoveSocketFromFlow failed to end a flow with error %d", 
            GetLastError());
        }
        session->rtp.QoSFlowID=0;
      }
    }
  }

  if (session->rtp.QoSHandle != NULL)
  {
    QOSCloseHandle(session->rtp.QoSHandle);
    session->rtp.QoSHandle=NULL;
  }
#endif
  if (session->modules_tables) {
    rtp_session_module_remove_all(session);
  }

  if (session->rtp_icechecklist!=NULL && session->rtp_icechecklist->cand_pairs!=NULL) {
    OList *elem;
    for (elem=session->rtp_icechecklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
      struct CandidatePair *cd=(struct CandidatePair*) elem->data;
      ortp_free(cd);
    }
    session->rtp_icechecklist->cand_pairs = o_list_free(session->rtp_icechecklist->cand_pairs);
  }
  if (session->rtp_icechecklist!=NULL)
    ortp_free(session->rtp_icechecklist);

  if (session->rtcp_icechecklist!=NULL && session->rtcp_icechecklist->cand_pairs!=NULL) {
    OList *elem;
    for (elem=session->rtcp_icechecklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
      struct CandidatePair *cd=(struct CandidatePair*) elem->data;
      ortp_free(cd);
    }
    session->rtcp_icechecklist->cand_pairs = o_list_free(session->rtcp_icechecklist->cand_pairs);
  }
  if (session->rtcp_icechecklist!=NULL)
    ortp_free(session->rtcp_icechecklist);

  BandwidthStatistics_destroy(&session->rtp.recv_total_rtppayload_bitrate);

  BandwidthStatistics_destroy(&session->rtp.recv_total_rtp_bitrate);
  BandwidthStatistics_destroy(&session->rtp.recv_total_rtcp_bitrate);
  BandwidthStatistics_destroy(&session->rtp.recv_total_stun_bitrate);
  BandwidthStatistics_destroy(&session->rtp.recv_total_other_bitrate);
  BandwidthStatistics_destroy(&session->rtp.recv_total_bad_bitrate);

  BandwidthStatistics_destroy(&session->rtp.sent_total_rtp_bitrate);
  BandwidthStatistics_destroy(&session->rtp.sent_total_rtcp_bitrate);
  BandwidthStatistics_destroy(&session->rtp.sent_total_stun_bitrate);
  BandwidthStatistics_destroy(&session->rtp.sent_total_other_bitrate);

}

/**
* Sets the number of packets containg a new SSRC that will trigger the
* "ssrc_changed" callback.
**/
void rtp_session_set_ssrc_changed_threshold(RtpSession *session, int numpackets){
  session->rtp.ssrc_changed_thres=numpackets;
}

/**
* Resynchronize to the incoming RTP streams.
* This can be useful to handle discoutinuous timestamps.
* For example, call this function from the timestamp_jump signal handler.
* @param session the rtp session
**/
void rtp_session_resync(RtpSession *session){
  flushq (&session->rtp.rq, FLUSHALL);
  rtp_session_set_flag(session, RTP_SESSION_RECV_SYNC);
  rtp_session_unset_flag(session,RTP_SESSION_FIRST_PACKET_DELIVERED);
  jitter_control_init(&session->rtp.jittctl,-1,NULL);
}

/**
* Reset the session: local and remote addresses are kept. It resets timestamp, sequence 
* number, and calls rtp_session_resync().
*
* @param session a rtp session.
**/
void rtp_session_reset (RtpSession * session)
{
  rtp_session_set_flag (session, RTP_SESSION_RECV_NOT_STARTED);
  rtp_session_set_flag (session, RTP_SESSION_SEND_NOT_STARTED);
  //session->ssrc=0;
  //session->rtp.snd_time_offset = 0;
  //session->rtp.snd_ts_offset = 0;
  //session->rtp.snd_rand_offset = 0;
  session->rtp.snd_last_ts = 0;
  //session->rtp.rcv_time_offset = 0;
  //session->rtp.rcv_ts_offset = 0;
  //session->rtp.rcv_query_ts_offset = 0;
  session->rtp.rcv_last_ts = 0;
  session->rtp.rcv_last_app_ts = 0;
  session->rtp.hwrcv_extseq = 0;
  session->rtp.hwrcv_since_last_SR=0;
  session->rtp.snd_seq = 1; /* start with 1 */
  session->rtp.sent_payload_bytes=0;
  rtp_session_clear_send_error_code(session);
  rtp_session_clear_recv_error_code(session);
  rtp_stats_reset(&session->rtp.stats);
  rtp_session_resync(session);
  session->ssrc_set=FALSE;
}

/**
* Retrieve the session's statistics.
**/
const rtp_stats_t * rtp_session_get_stats(const RtpSession *session){
  return &session->rtp.stats;
}

const rtp_extended_stats_t * rtp_session_get_extented_stats(const RtpSession *session){

  return &session->rtp.extended_stats;
}

void rtp_session_reset_stats(RtpSession *session){
  memset(&session->rtp.stats,0,sizeof(rtp_stats_t));
}

/**
* Stores some application specific data into the session, so that it is easy to retrieve it from the signal callbacks using rtp_session_get_data().
* @param session a rtp session
* @param data an opaque pointer to be stored in the session
**/

void rtp_session_set_data(RtpSession *session, void *data){
  session->user_data=data;
}

/**
* @param session a rtp session
* @return the void pointer previously set using rtp_session_set_data()
**/
void *rtp_session_get_data(const RtpSession *session){
  return session->user_data;
}

/**
* Enable or disable the "rtp symmetric" hack which consists of the following:
* after the first packet is received, the source address of the packet 
* is set to be the destination address for all next packets.
* This is useful to pass-through firewalls.
* @param session a rtp session
* @param yesno a boolean to enable or disable the feature
*
**/
void
  rtp_session_set_symmetric_rtp (RtpSession * session, bool_t yesno)
{
  session->symmetric_rtp =yesno;
}

/**
* Enable or disable the "rtcp mux" (rfc5761)
* Condition: no RTP payload sould be offered within [64-95] range.
* @param session a rtp session
* @param yesno a boolean to enable or disable the feature
*
**/
void
  rtp_session_set_rtcp_mux (RtpSession * session, bool_t yesno)
{
  session->rtcp_mux =yesno;
}

/**
* Enable or disable the "rtcp-fb: nack" (rfc4585)
* @param session a rtp session
* @param yesno a boolean to enable or disable the feature
*
**/
void
  rtp_session_set_rtcp_fb_nack (RtpSession * session, bool_t yesno)
{
  session->rtcpfeedback_nack =yesno;
}

/**
* Enable or disable the "rtcp-fb: nack pli" (rfc4585)
* @param session a rtp session
* @param yesno a boolean to enable or disable the feature
*
**/
void
  rtp_session_set_rtcp_fb_pli (RtpSession * session, bool_t yesno)
{
  session->rtcpfeedback_pli =yesno;
}

/**
* Enable or disable the "rtcp-fb: ccm tmmbr" (rfc5104)
* @param session a rtp session
* @param yesno a boolean to enable or disable the feature
*
**/
void
  rtp_session_set_rtcp_fb_ccm_tmmbr (RtpSession * session, bool_t yesno)
{
  session->rtcpfeedback_ccm_tmmbr =yesno;
}

/**
* Enable or disable the "rtcp-fb: goog-remb" (v)
* @param session a rtp session
* @param yesno a boolean to enable or disable the feature
*
**/
void
  rtp_session_set_rtcp_fb_goog_remb (RtpSession * session, bool_t yesno)
{
  session->rtcpfeedback_goog_remb =yesno;
}

/**
 * Enable or disable the "rtcp-xr: voip-metrics" (rfc)
 * @param session a rtp session
 * @param yesno a boolean to enable or disable the feature
 *
 **/
void
rtp_session_set_rtcp_xr_voip_metrics (RtpSession * session, bool_t yesno)
{
  session->rtcpxr_voip_metrics =yesno;
}

static void
rtp_session_calculate_voip_metrics(RtpSession * session)
{
  float p23;
  float p32;
  uint64_t ctotal;
  uint64_t c13;
  float m;
  
  if (session->rtp.stats.frameduration_ms==0)
    return;
  if (session->rtp.stats.frameduration_ms>160) {
    ortp_warning("frameduration seems very high? %ims", session->rtp.stats.frameduration_ms);
    return;
  }
  if (session->rtp.stats.c11 + session->rtp.stats.c14==0)
    return;
  
  // Calculate additional transition counts.
  session->rtp.stats.c31 = session->rtp.stats.c13;
  session->rtp.stats.c32 = session->rtp.stats.c23;
  ctotal = session->rtp.stats.c11 + session->rtp.stats.c14 + session->rtp.stats.c13 + session->rtp.stats.c22 + session->rtp.stats.c23 + session->rtp.stats.c31 + session->rtp.stats.c32 + session->rtp.stats.c33;
  
  // Calculate burst and densities.
  if (session->rtp.stats.c31 + session->rtp.stats.c32 + session->rtp.stats.c33 == 0)
    p32 = 0;
  else
    p32 = 1.0f * session->rtp.stats.c32 / (session->rtp.stats.c31 + session->rtp.stats.c32 + session->rtp.stats.c33);
  if((session->rtp.stats.c22 + session->rtp.stats.c23) < 1) {
    p23 = 1;
  }
  else {
    p23 = 1.0f - 1.0f * session->rtp.stats.c22/(session->rtp.stats.c22 + session->rtp.stats.c23);
  }
  session->rtp.stats.burst_density = 1.0f * p23 / (p23 + p32);
  if (session->rtp.stats.c11 == 0)
    session->rtp.stats.gap_density = 0;
  else
    session->rtp.stats.gap_density = 1.0f * session->rtp.stats.c14 / (session->rtp.stats.c11 + session->rtp.stats.c14);
  //burst_density = 256.0 * p23 / (p23 + p32);
  //gap_density = 256.0 * session->rtp.stats.c14 / (session->rtp.stats.c11 + session->rtp.stats.c14);
  ortp_message("burst_density: %f // gap_density: %f", session->rtp.stats.burst_density, session->rtp.stats.gap_density);
  
  // Calculate burst and gap durations in ms
  //m = frameDuration_in_ms * framesPerRTPPkt;
  m=1.0f * session->rtp.stats.frameduration_ms; //20ms usually;
  c13 = session->rtp.stats.c13;
  if (c13 == 0) {
    c13 = 1;
    ctotal += 1;
  }
  session->rtp.stats.gap_length = (session->rtp.stats.c11 + session->rtp.stats.c14 + session->rtp.stats.c13) * m / c13;
  session->rtp.stats.burst_length = ctotal * m / c13 - session->rtp.stats.gap_length; //(assume lgap is gap_length)
#if defined(__linux)
  ortp_message("burst_length: %"PRIu64"ms // gap_length: %"PRIu64"ms", session->rtp.stats.burst_length, session->rtp.stats.gap_length);
#elif defined(__APPLE__)
  ortp_message("burst_length: %"PRIu64"ms // gap_length: %"PRIu64"ms", session->rtp.stats.burst_length, session->rtp.stats.gap_length);
#elif !defined(WIN32)
  ortp_message("burst_length: %%lldms // gap_length: %%lldms", (long long)session->rtp.stats.burst_length, (long long)session->rtp.stats.gap_length);
#else
  ortp_message("burst_length: %I64ums // gap_length: %I64ums", session->rtp.stats.burst_length, session->rtp.stats.gap_length);
#endif
  /* calculate loss and discard rates */
  //loss_rate = 256.0 * session->rtp.stats.voip_metrics_loss_count / ctotal;
  //discard_rate = 256.0 * session->rtp.stats.voip_metrics_discard_count / ctotal;
}

void
rtp_session_update_voip_metrics(RtpSession * session, int packet_received, int packet_lost, int packet_discarded)
{
  if (!session->rtcpxr_voip_metrics)
    return;
  if (packet_received==0 && packet_lost==0 && packet_discarded==0)
    return;
  session->rtp.stats.gmin=16;
  if(packet_lost) {
    session->rtp.stats.voip_metrics_loss_count+=packet_lost;
  }
  if(packet_discarded) {
    session->rtp.stats.voip_metrics_discard_count+=packet_discarded;
  }
  if(packet_lost || packet_discarded) {
    int count=(packet_discarded>0)?packet_discarded:packet_lost;
    for (;count>0;count--) {
      if(session->rtp.stats.voip_metrics_pkt_received_count >= session->rtp.stats.gmin) {
        if(session->rtp.stats.voip_metrics_lost_within_current_burst == 1) {
          session->rtp.stats.c14++;
        }
        else {
          session->rtp.stats.c13++;
        }
        session->rtp.stats.voip_metrics_lost_within_current_burst = 1;
        session->rtp.stats.c11 += session->rtp.stats.voip_metrics_pkt_received_count;
      }
      else {
        session->rtp.stats.voip_metrics_lost_within_current_burst++;
        if(session->rtp.stats.voip_metrics_pkt_received_count == 0) {
          session->rtp.stats.c33++;
        }
        else {
          session->rtp.stats.c23++;
          session->rtp.stats.c22 += (session->rtp.stats.voip_metrics_pkt_received_count - 1);
        }
      }
      session->rtp.stats.voip_metrics_pkt_received_count = 0;
    }

    rtp_session_calculate_voip_metrics(session);
  }
  
  if(packet_received) {
    session->rtp.stats.voip_metrics_pkt_received_count++;
    if(session->rtp.stats.voip_metrics_pkt_received_count >= session->rtp.stats.gmin*7) {
      /* if no new loss/discard is happening for a while, force to update the stats to show a larger gap duration */
      session->rtp.stats.c11 += session->rtp.stats.gmin*6;
      session->rtp.stats.voip_metrics_pkt_received_count -= session->rtp.stats.gmin*6;
      
      rtp_session_calculate_voip_metrics(session);
    }
  }
}

static float compute_bw(struct timeval *orig, unsigned int bytes){
  struct timeval current;
  float time;
  if (bytes==0) return 0;
  gettimeofday(&current,NULL);
  time=(float)(current.tv_sec - orig->tv_sec) +
    ((float)(current.tv_usec - orig->tv_usec)*1e-6f);
  return ((float)bytes)*8/(time+0.001f); 
  /*+0.0001 avoids a division by zero without changing the results significatively*/
}

float rtp_session_get_recv_rtp_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.recv_total_rtp_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.recv_total_rtp_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.recv_total_rtp_bitrate.current_total_bitrate;
}

float rtp_session_get_sent_rtp_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.sent_total_rtp_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.sent_total_rtp_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.sent_total_rtp_bitrate.current_total_bitrate;
}

float rtp_session_get_recv_stun_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.recv_total_stun_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.recv_total_stun_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.recv_total_stun_bitrate.current_total_bitrate;
}

float rtp_session_get_sent_stun_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.sent_total_stun_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.sent_total_stun_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.sent_total_stun_bitrate.current_total_bitrate;
}

float rtp_session_get_recv_other_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.recv_total_other_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.recv_total_other_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.recv_total_other_bitrate.current_total_bitrate;
}

float rtp_session_get_sent_other_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.sent_total_other_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.sent_total_other_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.sent_total_other_bitrate.current_total_bitrate;
}

float rtp_session_get_recv_bad_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.recv_total_bad_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.recv_total_bad_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.recv_total_bad_bitrate.current_total_bitrate;
}

float rtp_session_get_recv_rtcp_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.recv_total_rtcp_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.recv_total_rtcp_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.recv_total_rtcp_bitrate.current_total_bitrate;
}

float rtp_session_get_sent_rtcp_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.sent_total_rtcp_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.sent_total_rtcp_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.sent_total_rtcp_bitrate.current_total_bitrate;
}

float rtp_session_get_recv_rtppayload_bandwidth(RtpSession *session, BandwidthStatistics *val){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  BandwidthStatistics_update(&session->rtp.recv_total_rtppayload_bitrate, 0, 0,&tv);
  if (val!=NULL) memcpy(val, &session->rtp.recv_total_rtppayload_bitrate, sizeof(BandwidthStatistics));
  return (float)session->rtp.recv_total_rtppayload_bitrate.current_total_bitrate;
}

float rtp_session_compute_recv_bandwidth(RtpSession *session){
  float bw;
  bw=compute_bw(&session->rtp.recv_bw_start,session->rtp.recv_bytes);
  session->rtp.recv_bytes=0;
  return bw;
}

float rtp_session_compute_send_bandwidth(RtpSession *session){
  float bw;
  bw=compute_bw(&session->rtp.send_bw_start,session->rtp.sent_bytes);
  session->rtp.sent_bytes=0;
  return bw;
}

float rtp_session_compute_rtcp_recv_bandwidth(RtpSession *session){
  float bw;
  bw=compute_bw(&session->rtcp.recv_bw_start,session->rtcp.recv_bytes);
  session->rtcp.recv_bytes=0;
  return bw;
}

float rtp_session_compute_rtcp_send_bandwidth(RtpSession *session){
  float bw;
  bw=compute_bw(&session->rtcp.send_bw_start,session->rtcp.sent_bytes);
  session->rtcp.sent_bytes=0;
  return bw;
}

/**
* initialize a bandwidth calculator
* @param session a rtp session
* @param yesno a boolean to enable or disable the feature
*
**/
void BandwidthStatistics_init(BandwidthStatistics *bw_stats, int period) {
  bw_stats->period=period;
  if (bw_stats->period<=0)
    bw_stats->period=5000; /* ms */

  bw_stats->num_entries = (bw_stats->period/20);
  bw_stats->bytes = (uint16_t*)ortp_malloc(sizeof(uint16_t)*bw_stats->num_entries);
  memset(bw_stats->bytes, 0, sizeof(uint16_t)*bw_stats->num_entries);
  bw_stats->overhead_bytes = (uint16_t*)ortp_malloc(sizeof(uint16_t)*bw_stats->num_entries);
  memset(bw_stats->overhead_bytes, 0, sizeof(uint16_t)*bw_stats->num_entries);
  bw_stats->times = (struct timeval*)ortp_malloc(sizeof(struct timeval)*bw_stats->num_entries);
  memset(bw_stats->times, 0, sizeof(struct timeval)*bw_stats->num_entries);
}

void BandwidthStatistics_destroy(BandwidthStatistics *bw_stats) {
  ortp_free(bw_stats->bytes);
  ortp_free(bw_stats->overhead_bytes);
  ortp_free(bw_stats->times);
  bw_stats->final_bytes=0;
  bw_stats->final_overhead_bytes=0;
  bw_stats->bytes=NULL;
  bw_stats->overhead_bytes=NULL;
  bw_stats->times=NULL;
  bw_stats->total_bytes=0;
  bw_stats->total_overhead_bytes=0;
  bw_stats->index_last=0;
}

void BandwidthStatistics_update(BandwidthStatistics *bw_stats, int nbytes, int overhead_nbytes, struct timeval *now) {
  uint64_t interval;
  uint64_t last_interval;
  if (bw_stats->bytes==NULL)
    BandwidthStatistics_init(bw_stats, 0);

  interval = (now->tv_sec-bw_stats->times[0].tv_sec)*1000;
  interval+=(now->tv_usec-bw_stats->times[0].tv_usec)/1000;

  bw_stats->final_bytes+=nbytes;
  bw_stats->final_overhead_bytes+=nbytes;

  if (interval<=40) {
    bw_stats->bytes[0]+=nbytes;
    bw_stats->total_bytes+=nbytes;
    bw_stats->overhead_bytes[0]+=overhead_nbytes;
    bw_stats->total_overhead_bytes+=overhead_nbytes;
    return;
  }

  last_interval=0;
  while (bw_stats->index_last>0) {
    last_interval = (now->tv_sec-bw_stats->times[bw_stats->index_last-1].tv_sec)*1000;
    last_interval+= (now->tv_usec-bw_stats->times[bw_stats->index_last-1].tv_usec)/1000;
    if (last_interval<bw_stats->period && bw_stats->index_last<bw_stats->num_entries)
      break;
    bw_stats->total_bytes-=bw_stats->bytes[bw_stats->index_last-1];
    bw_stats->total_overhead_bytes-=bw_stats->overhead_bytes[bw_stats->index_last-1];
    bw_stats->index_last--;
  }

  if (nbytes>0 || overhead_nbytes>0) {
    bw_stats->total_bytes+=nbytes;
    bw_stats->total_overhead_bytes+=overhead_nbytes;
    memmove(((uint8_t*)bw_stats->bytes)+sizeof(uint16_t), bw_stats->bytes, sizeof(uint16_t)*(bw_stats->index_last));
    memmove(((uint8_t*)bw_stats->overhead_bytes)+sizeof(uint16_t), bw_stats->overhead_bytes, sizeof(uint16_t)*(bw_stats->index_last));
    memmove(((uint8_t*)bw_stats->times)+sizeof(struct timeval), bw_stats->times, sizeof(struct timeval)*(bw_stats->index_last));
    bw_stats->bytes[0]=nbytes;
    bw_stats->overhead_bytes[0]=overhead_nbytes;
    bw_stats->times[0].tv_sec = now->tv_sec;
    bw_stats->times[0].tv_usec = now->tv_usec;
    bw_stats->index_last++;
  }

  if (bw_stats->index_last>1) {
    bw_stats->current_total_bitrate = (uint32_t)compute_bw(&bw_stats->times[bw_stats->index_last-1], bw_stats->total_bytes);
    bw_stats->current_overhead_bitrate = (uint32_t)compute_bw(&bw_stats->times[bw_stats->index_last-1], bw_stats->total_overhead_bytes);
  } else if (bw_stats->index_last==1 && bw_stats->total_bytes>0) {
    bw_stats->current_total_bitrate = (uint32_t) (((float)bw_stats->total_bytes)*8/(bw_stats->period+0.001f));
    bw_stats->current_overhead_bitrate = (uint32_t) (((float)bw_stats->total_overhead_bytes)*8/(bw_stats->period+0.001f));
    if (bw_stats->current_total_bitrate==0)
      bw_stats->current_total_bitrate = 1; /* workaround: minimum value so the stats show "something" */
    if (bw_stats->current_overhead_bitrate==0)
      bw_stats->current_overhead_bitrate = 1; /* workaround: minimum value so the stats show "something" */
  } else {
    bw_stats->current_total_bitrate=0;
    bw_stats->current_overhead_bitrate=0;
  }
}

float rtp_session_compute_average_jitter(RtpSession *session) {
  double diff;
  if (session->rtp.jittctl.average_buffer_size_count==0)
    return 0;

  diff=((double)session->rtp.jittctl.average_buffer_size)/(double)session->rtp.jittctl.average_buffer_size_count;
  session->rtp.jittctl.average_buffer_size=0;
  session->rtp.jittctl.average_buffer_size_count=0;
  session->rtp.stats.average_jitter = (float)(1000.0*(float)diff/(float)session->rtp.jittctl.clock_rate);
  return session->rtp.stats.average_jitter;
}

int rtp_session_get_last_send_error_code(RtpSession *session){
  return session->rtp.send_errno;
}

void rtp_session_clear_send_error_code(RtpSession *session){
  session->rtp.send_errno=0;
}

int rtp_session_get_last_recv_error_code(RtpSession *session){
  return session->rtp.recv_errno;
}

void rtp_session_clear_recv_error_code(RtpSession *session){
  session->rtp.recv_errno=0;
}

/**
* Destroys a rtp session.
* All memory allocated for the RtpSession is freed.
*
* @param session a rtp session.
**/
void rtp_session_destroy (RtpSession * session)
{
  rtp_session_uninit (session);
  ortp_free (session);
}

void rtp_session_make_time_distorsion(RtpSession *session, int milisec)
{
  //session->rtp.snd_time_offset+=milisec;
}


/* packet api */

void rtp_add_csrc(mblk_t *mp, uint32_t csrc)
{
  rtp_header_t *hdr=(rtp_header_t*)mp->b_rptr;
  hdr->csrc[hdr->cc]=csrc;
  hdr->cc++;
}

/**
* Get a pointer to the beginning of the payload data of the RTP packet.
* @param packet a RTP packet represented as a mblk_t
* @param start a pointer to the beginning of the payload data, pointing inside the packet.
* @return the length of the payload data.
**/
int rtp_get_payload(mblk_t *packet, unsigned char **start){
  unsigned char *tmp;
  int header_len=RTP_FIXED_HEADER_SIZE+(rtp_get_cc(packet)*4)+(rtp_get_extbit(packet)*4);
  if (rtp_get_extbit(packet))
  {
    uint16_t *ext_buf = (uint16_t *)(packet->b_rptr+header_len-2);
    header_len += (ntohs(ext_buf[0])*4);
  }

  if (rtp_get_padbit(packet))
  {
    if (packet->b_cont==NULL) {
      int padding_size = *(packet->b_wptr-1);
      if (packet->b_wptr-padding_size>=packet->b_rptr)
        packet->b_wptr=packet->b_wptr-padding_size;
    }
  }

  tmp=packet->b_rptr+header_len;
  if (tmp>packet->b_wptr){
    if (packet->b_cont!=NULL){
      tmp=packet->b_cont->b_rptr+(header_len- (packet->b_wptr-packet->b_rptr));
      if (tmp<=packet->b_cont->b_wptr){
        *start=tmp;
        return (int)(packet->b_cont->b_wptr-tmp);
      }
    }
    ortp_warning("Invalid RTP packet");
    return -1;
  }
  *start=tmp;
  return (int)(packet->b_wptr-tmp);
}


/**
*  Gets last time a valid RTP or RTCP packet was received.
* @param session RtpSession to get last receive time from.
* @param tv Pointer to struct timeval to fill.
*
**/
void
  rtp_session_get_last_recv_time(RtpSession *session, struct timeval *tv)
{
#ifdef PERF
  ortp_error("rtp_session_get_last_recv_time() feature disabled.");
#else
  *tv = session->last_recv_time;
#endif
}



uint32_t rtp_session_time_to_ts(RtpSession *session, int millisecs){
  PayloadType *payload;
  payload =
    rtp_profile_get_payload (session->snd.profile,
    session->snd.pt);
  if (payload == NULL)
  {
    ortp_warning
      ("rtp_session_ts_to_t: use of unsupported payload type %d.", session->snd.pt);
    return 0;
  }
  /* the return value is in milisecond */
  return (uint32_t) (payload->clock_rate*(double) (millisecs/1000.0f));
}

/* function used by the scheduler only:*/
uint32_t rtp_session_ts_to_time (RtpSession * session, uint32_t timestamp)
{
  PayloadType *payload;
  payload =
    rtp_profile_get_payload (session->snd.profile,
    session->snd.pt);
  if (payload == NULL)
  {
    ortp_warning
      ("rtp_session_ts_to_t: use of unsupported payload type %d.", session->snd.pt);
    return 0;
  }
  /* the return value is in milisecond */
  return (uint32_t) (1000.0 *
    ((double) timestamp /
    (double) payload->clock_rate));
}

static int _rtp_session_ice_restart(struct IceCheckList *checklist)
{
  OList *candidate_pair_table;
  OList *elem;

  int count_waiting=0;

  if (checklist==NULL)
    return 0;
  candidate_pair_table = checklist->cand_pairs;
  if (candidate_pair_table==NULL)
    return 0;

  for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem)){
    struct CandidatePair *cd=(struct CandidatePair*) elem->data;
    if (strcasecmp(cd->local_candidate.cand_type, "srflx")==0)
    {
      /* search for a highest priority "equivalent" pair */
      for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem)){
        struct CandidatePair *cand_pair=(struct CandidatePair*) elem->data;
        if (cand_pair==cd)
          break; /* stop */

        /* same "base" address (origin of STUN connectivity check to the remote candidate */
        if (strcasecmp(cd->local_candidate.rel_addr, /* base address for "reflexive" address */
          cand_pair->local_candidate.conn_addr)==0) /* base address for "host" address */
        {
          /* if same target remote candidate: -> remove the one with lowest priority */
          if (strcasecmp(cd->remote_candidate.conn_addr,
            cand_pair->remote_candidate.conn_addr)==0)
          {    
            /* useless cpair */                 
            ortp_message("ice.c(?): Removing useless pair! (%s:%i/%s:%i/%s :: %s:%i/%s:%i/%s)",
              cd->local_candidate.conn_addr,
              cd->local_candidate.conn_port,
              (cd->local_candidate.rel_addr[0]=='\0')?"-":cd->local_candidate.rel_addr,
              cd->local_candidate.rel_port,
              cd->local_candidate.cand_type,
              cd->remote_candidate.conn_addr,
              cd->remote_candidate.conn_port,
              (cd->remote_candidate.rel_addr[0]=='\0')?"-":cd->remote_candidate.rel_addr,
              cd->remote_candidate.rel_port,
              cd->remote_candidate.cand_type);
            cd->connectivitycheck.state = ICE_PRUNED;
          }
        }
      }
    }
  }

  /* no currently nominated pair */
  checklist->nominated_pair_index = -1;

  for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem)){
    struct CandidatePair *cd=(struct CandidatePair*) elem->data;
    if (cd->connectivitycheck.state == ICE_PRUNED)
      continue;
    if (cd->connectivitycheck.state == ICE_FROZEN)
      cd->connectivitycheck.state = ICE_WAITING;
  }

  checklist->Ta = 40;

  for (elem=candidate_pair_table;elem!=NULL;elem=o_list_next(elem)){
    struct CandidatePair *cd=(struct CandidatePair*) elem->data;
    if (cd->connectivitycheck.state == ICE_PRUNED)
      continue;

    if (cd->connectivitycheck.state == ICE_WAITING)
      count_waiting++;
  }

  checklist->RTO = MAX(200, checklist->Ta*count_waiting);
  return 0;
}

int rtp_session_set_icechecklist (RtpSession * session, struct IceCheckList *rtp_checklist, struct IceCheckList *rtcp_checklist)
{
  if (session->rtp_icechecklist!=NULL && session->rtp_icechecklist->cand_pairs!=NULL) {
    OList *elem;
    for (elem=session->rtp_icechecklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
      struct CandidatePair *cd=(struct CandidatePair*) elem->data;
      ortp_free(cd);
    }
    session->rtp_icechecklist->cand_pairs = o_list_free(session->rtp_icechecklist->cand_pairs);
  }
  if (session->rtp_icechecklist!=NULL)
    ortp_free(session->rtp_icechecklist);

  session->rtp_icechecklist=ortp_new0(struct IceCheckList,1);

  if (rtp_checklist!=NULL) {
    OList *elem;

    memcpy(session->rtp_icechecklist, rtp_checklist, sizeof(struct IceCheckList));
    session->rtp_icechecklist->cand_pairs=NULL; /* build a duplicated list */

    for (elem=rtp_checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
      struct CandidatePair *cd=(struct CandidatePair*) elem->data;
      struct CandidatePair *cd_dup=ortp_new(struct CandidatePair,1);
      memcpy(cd_dup, cd, sizeof(struct CandidatePair));
      session->rtp_icechecklist->cand_pairs = o_list_append(session->rtp_icechecklist->cand_pairs, cd_dup);
    }
    _rtp_session_ice_restart(session->rtp_icechecklist);
  }


  if (session->rtcp_icechecklist!=NULL && session->rtcp_icechecklist->cand_pairs!=NULL) {
    OList *elem;
    for (elem=session->rtcp_icechecklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
      struct CandidatePair *cd=(struct CandidatePair*) elem->data;
      ortp_free(cd);
    }
    session->rtcp_icechecklist->cand_pairs = o_list_free(session->rtcp_icechecklist->cand_pairs);
  }
  if (session->rtcp_icechecklist!=NULL)
    ortp_free(session->rtcp_icechecklist);

  session->rtcp_icechecklist=ortp_new0(struct IceCheckList,1);

  if (rtcp_checklist!=NULL) {
    OList *elem;

    memcpy(session->rtcp_icechecklist, rtcp_checklist, sizeof(struct IceCheckList));
    session->rtcp_icechecklist->cand_pairs=NULL; /* build a duplicated list */

    for (elem=rtcp_checklist->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
      struct CandidatePair *cd=(struct CandidatePair*) elem->data;
      struct CandidatePair *cd_dup=ortp_new(struct CandidatePair,1);
      memcpy(cd_dup, cd, sizeof(struct CandidatePair));
      session->rtcp_icechecklist->cand_pairs = o_list_append(session->rtcp_icechecklist->cand_pairs, cd_dup);
    }
    _rtp_session_ice_restart(session->rtcp_icechecklist);
  }

  return 0;
}

int rtp_session_set_max_recv_bitrate(RtpSession *session, uint32_t max_ccm_bitrate) {
  RtpStream *st=&session->rtp;
  st->cur_bitrate_table.max_bitrate=max_ccm_bitrate;
  ortp_message("zigzag: set max bitrate limit=%i", st->cur_bitrate_table.max_bitrate);
  return 0;
}
