/*
The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ortp/ortp.h"
#include "utils.h"
#include "ortp/rtpsession.h"
#include "rtpsession_priv.h"

#if defined(WIN32) || defined(_WIN32_WCE)
#include "ortp-config-win32.h"
#elif HAVE_CONFIG_H
#include "ortp-config.h" /*needed for HAVE_SYS_UIO_H */
#endif

#if (_WIN32_WINNT >= 0x0600)
#define ENABLE_RTP_QOS
//#define ENABLE_RTP_TRAFFIC
#if (_MSC_VER >= 1700) && !defined(_USING_V110_SDK71_)
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#undef ENABLE_RTP_QOS
#undef ENABLE_RTP_TRAFFIC
#endif
#endif
#endif


#ifdef ENABLE_RTP_QOS
#include <delayimp.h>
#undef ExternC
#include <QOS2.h>
#endif

#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#define USE_SENDMSG 1
#endif

#if defined(WIN32) || defined(_WIN32_WCE)
#ifndef _WIN32_WCE
#include <errno.h>
#endif
#ifdef EWOULDBLOCK
#define is_would_block_error(errnum)	(errnum==WSAEWOULDBLOCK || errnum==EWOULDBLOCK)
#else
#define is_would_block_error(errnum)	(errnum==WSAEWOULDBLOCK)
#endif
#define sockisvalid(s) (s != ~0)
#else
#define is_would_block_error(errnum)	(errnum==EWOULDBLOCK || errnum==EAGAIN)
#define sockisvalid(s) ((s) >= 0)
#endif

#ifdef TSC_SUPPORT
//#define TSC_REDUNDANCY

#include "tsc_socket_api.h"
#include "tsc_control_api.h"
#endif

void tc_qos_delete_flow(RtpSession *session);

/**
*close_socket
*@session: rtpsession
*@sock: the socket to close
*
* this method is an utility method that calls close() on UNIX or
* closesocket on Win32.
*
* Returns: the result of the system method.
*/
int close_socket(RtpSession * session, ortp_socket_t sock){
#ifdef TSC_SUPPORT
  if (session->tsc_handle) {
    return tsc_close(sock);
  }
#endif
#if	!defined(_WIN32) && !defined(_WIN32_WCE)
  return close (sock);
#else
  return closesocket(sock);
#endif
}

static ortp_socket_t create_and_bind(RtpSession * session, const char *addr, int port, int *sock_family, bool_t reuse_addr){
  int err;
  int optval = 1;
  ortp_socket_t sock=-1;

  char num[8];
  struct addrinfo hints, *res0, *res;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_DGRAM;
  snprintf(num, sizeof(num), "%d",port);
  err = getaddrinfo(addr,num, &hints, &res0);
  if (err!=0) {
    ortp_warning ("Error in getaddrinfo on (addr=%s port=%i): %s", addr, port, gai_strerror(err));
    return -1;
  }

#ifdef TSC_SUPPORT
  if (session->tsc_handle) {
    // one interface for tunnel
    res0->ai_next = NULL;
  }
#endif

  for (res = res0; res; res = res->ai_next) {

#ifdef TSC_SUPPORT
    if (session->tsc_handle) {
      unsigned long flags = 1;
      int value;

      sock = tsc_socket(session->tsc_handle, res->ai_family, res->ai_socktype, 0);

      if (sock==-1) {
        ortp_warning ("Error tsc_socket errno:%i", tsc_get_errno());
        continue;
      }

      // unblocking
      flags = 1;
      tsc_ioctl(sock, FIONBIO, &flags);
      value = 1;
      tsc_setsockopt(sock, SOL_SOCKET, SO_TSM_REALTIME, (char *)&value, sizeof(int));
#ifdef TSC_REDUNDANCY
      if (session->tsc_redundancy_factor > 0 && !(port & 1))
      {
        tsc_handle tunnel;
        tsc_tunnel_socket_info tunnel_info;
        tsc_so_redundancy redundancy;

        memset(&redundancy, 0, sizeof(tsc_so_redundancy));
        redundancy.redundancy_factor = session->tsc_redundancy_factor;

        tunnel = tsc_get_tunnel(sock); /* ? same as session->tsc_handle? */
        tsc_get_tunnel_socket_info(tunnel, &tunnel_info);

        if (tunnel_info.transport == tsc_transport_udp || tunnel_info.transport == tsc_transport_dtls) {
          redundancy.redundancy_method = tsc_so_redundancy_method_udp_dtls;
        } else {
          if (session->tsc_redundancy_lb) {
            redundancy.redundancy_method = tsc_so_redundancy_method_tcp_tls_load_balance;
          } else {
            redundancy.redundancy_method = tsc_so_redundancy_method_tcp_tls_fan_out;
          }
        }

        tsc_setsockopt(sock, SOL_SOCKET, SO_TSC_REDUNDANCY, (char *)&redundancy, sizeof(tsc_so_redundancy));
      }
#endif
    }
    else {
      sock = socket(res->ai_family, res->ai_socktype, 0);
    }
#else
    sock = socket(res->ai_family, res->ai_socktype, 0);
#endif
    
    if (sock==-1)
      continue;

#if SO_NOSIGPIPE
    optval = 1;
    setsockopt (sock, SOL_SOCKET, SO_NOSIGPIPE, (SOCKET_OPTION_VALUE) &optval, sizeof (int));
#endif

#ifdef IPV6_V6ONLY
    if (res->ai_family==AF_INET6) {
      optval = 1;
      setsockopt (sock, IPPROTO_IPV6, IPV6_V6ONLY, (SOCKET_OPTION_VALUE) &optval, sizeof (int));
    }
#endif
    
    if (reuse_addr){
#ifdef TSC_SUPPORT
      if (session->tsc_handle) {
        err = tsc_setsockopt (sock, SOL_SOCKET, SO_REUSEADDR,
          (SOCKET_OPTION_VALUE)&optval, sizeof (optval));
        if (err < 0)
        {
          ortp_warning ("Fail to set rtp address reusable: %s.", getSocketError());
        }
      } else {
        err = setsockopt (sock, SOL_SOCKET, SO_REUSEADDR,
          (SOCKET_OPTION_VALUE)&optval, sizeof (optval));
        if (err < 0)
        {
          ortp_warning ("Fail to set rtp address reusable: %s.", getSocketError());
        }
      }
#else
      err = setsockopt (sock, SOL_SOCKET, SO_REUSEADDR,
        (SOCKET_OPTION_VALUE)&optval, sizeof (optval));
      if (err < 0)
      {
        ortp_warning ("Fail to set rtp address reusable: %s.", getSocketError());
      }
#endif
    }

    *sock_family=res->ai_family;
#ifdef TSC_SUPPORT
    if (session->tsc_handle) {
      struct sockaddr_in *addr;
      tsc_config config;
      tsc_get_config(session->tsc_handle, &config);

      addr = (struct sockaddr_in *)(res->ai_addr);
      addr->sin_addr.s_addr = htonl(config.internal_address.address);

      err = tsc_bind (sock, res->ai_addr, res->ai_addrlen);
    }
    else {
      err = bind (sock, res->ai_addr, res->ai_addrlen);
    }
    if (err != 0){
      ortp_warning ("Fail to bind rtp socket to (addr=%s port=%i) : %i.", addr,port, tsc_get_errno());
      close_socket (session, sock);
      sock=-1;
      continue;
    }
#else
    err = bind (sock, res->ai_addr, res->ai_addrlen);
    if (err != 0){
      ortp_warning ("Fail to bind rtp socket to (addr=%s port=%i) : %s.", addr,port, getSocketError());
      close_socket (session, sock);
      sock=-1;
      continue;
    }
#endif
#ifndef __hpux
    switch (res->ai_family)
    {
    case AF_INET:
      if (IN_MULTICAST(ntohl(((struct sockaddr_in *) res->ai_addr)->sin_addr.s_addr)))
      {
        struct ip_mreq mreq;
        mreq.imr_multiaddr.s_addr = ((struct sockaddr_in *) res->ai_addr)->sin_addr.s_addr;
        mreq.imr_interface.s_addr = INADDR_ANY;
        err = setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (SOCKET_OPTION_VALUE) &mreq, sizeof(mreq));
        if (err < 0){
          ortp_warning ("Fail to join address group: %s.", getSocketError());
          close_socket (session, sock);
          sock=-1;
          continue;
        }
      }
      break;
    case AF_INET6:
      if (IN6_IS_ADDR_MULTICAST(&(((struct sockaddr_in6 *) res->ai_addr)->sin6_addr)))
      {
        struct ipv6_mreq mreq;
        mreq.ipv6mr_multiaddr = ((struct sockaddr_in6 *) res->ai_addr)->sin6_addr;
        mreq.ipv6mr_interface = 0;
        err = setsockopt(sock, IPPROTO_IPV6, IPV6_JOIN_GROUP, (SOCKET_OPTION_VALUE)&mreq, sizeof(mreq));
        if (err < 0)
        {
          ortp_warning ("Fail to join address group: %s.", getSocketError());
          close_socket (session, sock);
          sock=-1;
          continue;
        }
      }
      break;
    }
#endif /*hpux*/
    break;
  }
  freeaddrinfo(res0);
  if (sock!=-1){
    set_non_blocking_socket (sock);
  }
  return sock;
}

static void set_socket_sizes(ortp_socket_t sock, unsigned int sndbufsz, unsigned int rcvbufsz){
  int err;
  bool_t done=FALSE;
  if (sndbufsz>0){
#ifdef SO_SNDBUFFORCE
    err = setsockopt(sock, SOL_SOCKET, SO_SNDBUFFORCE, (void *)&sndbufsz, sizeof(sndbufsz)); 
    if (err == -1) {
      ortp_warning("Fail to increase socket's send buffer size with SO_SNDBUFFORCE: %s.", getSocketError());
    }else done=TRUE;
#endif
    if (!done){
      err = setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (void *)&sndbufsz, sizeof(sndbufsz)); 
      if (err == -1) {
        ortp_warning("Fail to increase socket's send buffer size with SO_SNDBUF: %s.", getSocketError());
      }
    }
  }
  done=FALSE;
  if (rcvbufsz>0){
#ifdef SO_RCVBUFFORCE
    err = setsockopt(sock, SOL_SOCKET, SO_RCVBUFFORCE, (void *)&rcvbufsz, sizeof(rcvbufsz)); 
    if (err == -1) {
      ortp_warning("Fail to increase socket's recv buffer size with SO_RCVBUFFORCE: %s.", getSocketError());
    }
#endif
    if (!done){
      err = setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (void *)&rcvbufsz, sizeof(rcvbufsz)); 
      if (err == -1) {
        ortp_warning("Fail to increase socket's recv buffer size with SO_RCVBUF: %s.", getSocketError());
      }
    }

  }
}

static ortp_socket_t create_and_bind_random(RtpSession * session, const char *localip, int *sock_family, int *port){
  int retry;
  ortp_socket_t sock = -1;
  for (retry=0;retry<100;retry++)
  {
    int localport;
    do
    {
      localport = (rand () + 5000) & 0xfffe;
    }
    while ((localport < 5000) || (localport > 0xffff));
    /*do not set REUSEADDR in case of random allocation */
    sock = create_and_bind(session, localip, localport, sock_family,FALSE);
    if (sock!=-1) {
      *port=localport;
      return sock;
    }
  }
  ortp_warning("create_and_bind_random: Could not find a random port for %s !",localip);
  return -1;
}

/**
*rtp_session_set_local_addr:
*@session:		a rtp session freshly created.
*@addr:		a local IP address in the xxx.xxx.xxx.xxx form.
*@port:		a local port or -1 to let oRTP choose the port randomly
*
*	Specify the local addr to be use to listen for rtp packets or to send rtp packet from.
*	In case where the rtp session is send-only, then it is not required to call this function:
*	when calling rtp_session_set_remote_addr(), if no local address has been set, then the 
*	default INADRR_ANY (0.0.0.0) IP address with a random port will be used. Calling 
*	rtp_sesession_set_local_addr() is mandatory when the session is recv-only or duplex.
*
*	Returns: 0 on success.
**/

int
  rtp_session_set_local_addr (RtpSession * session, int idx, const char * addr, int port)
{
  ortp_socket_t sock;
  int sockfamily;
  bool_t reuse_addr;

  if (sockisvalid(session->rtp.socketv4[idx])){
    /* don't rebind, but close before*/
    if (sockisvalid(session->rtp.socketv4[idx])) close_socket (session, session->rtp.socketv4[idx]);
    if (sockisvalid(session->rtcp.socketv4[idx])) close_socket (session, session->rtcp.socketv4[idx]);
  }
#ifdef SO_REUSE_ADDR
  reuse_addr=TRUE;
#else
  reuse_addr=FALSE;
#endif
  /* try to bind the rtp port */
  if (port>0)
    sock=create_and_bind(session, addr,port,&sockfamily,reuse_addr);
  else
    sock=create_and_bind_random(session, addr,&sockfamily,&port);
  if (sock!=-1){
    set_socket_sizes(sock,session->rtp.snd_socket_size,session->rtp.rcv_socket_size);
    session->rtp.sockfamily[idx]=sockfamily;
    session->rtp.socketv4[idx]=sock;
    session->rtp.loc_port=port;
    /*try to bind rtcp port */
    sock=create_and_bind(session, addr,port+1,&sockfamily,reuse_addr);
    if (sock!=-1){
      session->rtcp.sockfamily[idx]=sockfamily;
      session->rtcp.socketv4[idx]=sock;
    }else{
      ortp_warning("Could not create and bind rtcp socket.");
    }

    /* set socket options (but don't change chosen states) */
    rtp_session_set_dscp( session, idx, -1 );
    rtp_session_set_multicast_ttl( session, idx, -1 );
    rtp_session_set_multicast_loopback( session, idx, -1 );

    return 0;
  }
  return -1;
}


/**
*rtp_session_set_multicast_ttl:
*@session: a rtp session
*@ttl: desired Multicast Time-To-Live
*
* Sets the TTL (Time-To-Live) for outgoing multicast packets.
*
* Returns: 0 on success.
*
**/
int rtp_session_set_multicast_ttl(RtpSession *session, int idx, int ttl)
{
  int retval;

  // Store new TTL if one is specified
  if (ttl>0) session->multicast_ttl = ttl;

  // Don't do anything if socket hasn't been created yet
  if (!sockisvalid(session->rtp.socketv4[idx])) return 0;

  switch (session->rtp.sockfamily[idx]) {
  case AF_INET: {

    retval= setsockopt(session->rtp.socketv4[idx], IPPROTO_IP, IP_MULTICAST_TTL,
      (SOCKET_OPTION_VALUE)  &session->multicast_ttl, sizeof(session->multicast_ttl));

    if (retval<0) break;

    retval= setsockopt(session->rtcp.socketv4[idx], IPPROTO_IP, IP_MULTICAST_TTL,
      (SOCKET_OPTION_VALUE)	   &session->multicast_ttl, sizeof(session->multicast_ttl));

                } break;
  case AF_INET6: {

    retval= setsockopt(session->rtp.socketv4[idx], IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
      (SOCKET_OPTION_VALUE)&session->multicast_ttl, sizeof(session->multicast_ttl));

    if (retval<0) break;

    retval= setsockopt(session->rtcp.socketv4[idx], IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
      (SOCKET_OPTION_VALUE) &session->multicast_ttl, sizeof(session->multicast_ttl));

                 } break;
  default:
    retval=-1;
  }

  if (retval<0)
    ortp_warning("Failed to set multicast TTL on socket.");


  return retval;
}


/**
*rtp_session_get_multicast_ttl:
*@session: a rtp session
*
* Returns the TTL (Time-To-Live) for outgoing multicast packets.
*
**/
int rtp_session_get_multicast_ttl(RtpSession *session)
{
  return session->multicast_ttl;
}


/**
*rtp_session_set_multicast_loopback:
*@session: a rtp session
*@ttl: desired Multicast Time-To-Live
*
* Sets the TTL (Time-To-Live) for outgoing multicast packets.
*
* Returns: 0 on success.
*
**/
int rtp_session_set_multicast_loopback(RtpSession *session, int idx, int yesno)
{
  int retval;

  // Store new loopback state if one is specified
  if (yesno==0) {
    // Don't loop back
    session->multicast_loopback = 0;
  } else if (yesno>0) {
    // Do loop back
    session->multicast_loopback = 1;
  }

  // Don't do anything if socket hasn't been created yet
  if (!sockisvalid(session->rtp.socketv4[idx])) return 0;

  switch (session->rtp.sockfamily[idx]) {
  case AF_INET: {

    retval= setsockopt(session->rtp.socketv4[idx], IPPROTO_IP, IP_MULTICAST_LOOP,
      (SOCKET_OPTION_VALUE)   &session->multicast_loopback, sizeof(session->multicast_loopback));

    if (retval<0) break;

    retval= setsockopt(session->rtcp.socketv4[idx], IPPROTO_IP, IP_MULTICAST_LOOP,
      (SOCKET_OPTION_VALUE)   &session->multicast_loopback, sizeof(session->multicast_loopback));

                } break;
  case AF_INET6: {

    retval= setsockopt(session->rtp.socketv4[idx], IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
      (SOCKET_OPTION_VALUE)	&session->multicast_loopback, sizeof(session->multicast_loopback));

    if (retval<0) break;

    retval= setsockopt(session->rtcp.socketv4[idx], IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
      (SOCKET_OPTION_VALUE)	&session->multicast_loopback, sizeof(session->multicast_loopback));

                 } break;
  default:
    retval=-1;
  }

  if (retval<0)
    ortp_warning("Failed to set multicast loopback on socket.");


  return retval;
}


/**
*rtp_session_get_multicast_loopback:
*@session: a rtp session
*
* Returns the multicast loopback state of rtp session (true or false).
*
**/
int rtp_session_get_multicast_loopback(RtpSession *session)
{
  return session->multicast_loopback;
}

#ifdef ENABLE_RTP_TRAFFIC

#ifndef UNICODE
#define UNICODE
#endif

#include <ntddndis.h> 
#include <Traffic.h>

VOID MyClNotifyHandler(
  IN  HANDLE      ClRegCtx,
  IN  HANDLE      ClIfcCtx,
  IN  ULONG       Event,      // See list below
  IN  HANDLE      SubCode,
  IN  ULONG       BufSize,
  IN  PVOID       Buffer
  )
{
  ortp_message("MyClNotifyHandler handler");
}

struct _tc_qos_ifc {
  char interfaceName[256];
  HANDLE flowHandle;
  HANDLE FilterHandle;
  HANDLE ifcHandle;
};

static HANDLE QosClientHandle = NULL;
static BOOL tc_initialized = FALSE;

static void tc_qos_init()
{
  TCI_CLIENT_FUNC_LIST QoSFunctions;
  long result;

  if (tc_initialized==TRUE)
    return;
  tc_initialized = TRUE;

  QoSFunctions.ClAddFlowCompleteHandler = NULL;
  QoSFunctions.ClDeleteFlowCompleteHandler = NULL;
  QoSFunctions.ClModifyFlowCompleteHandler = NULL;
  QoSFunctions.ClNotifyHandler = (TCI_NOTIFY_HANDLER)MyClNotifyHandler;

  //Register call back functions with Traffic control interface.
  result = TcRegisterClient(CURRENT_TCI_VERSION, NULL, &QoSFunctions , &QosClientHandle); 

  if (result!= NO_ERROR)
  {
    ortp_warning("TC: no admin rights for TcRegisterClient usage");
    return;
  }
  else if (result!= NO_ERROR)
  {
    ortp_error("TC: %d reported by TcRegisterClient",result);
    return;
  }
}

void tc_qos_delete_flow(RtpSession *session)
{
  if (session->tc_qos.num_ifc<=0)
  {
    session->tc_qos.num_ifc=0;
    if (session->tc_qos.ifc!=NULL)
      ortp_free(session->tc_qos.ifc);
    session->tc_qos.ifc=NULL;
  }
  while (session->tc_qos.num_ifc>0)
  {
    ULONG retCode;
    struct _tc_qos_ifc *ifc;
    session->tc_qos.num_ifc--;
    ifc=&session->tc_qos.ifc[session->tc_qos.num_ifc];

    // Add filter to the flow. This point onwards all traffic matching the filter, will pass through the flow.
    if (ifc->FilterHandle!=NULL)
    {
      retCode = TcDeleteFilter(ifc->FilterHandle);

      if (retCode!= NO_ERROR)
      {
        ortp_error("TC: %s: %d reported by in TcDeleteFilter", ifc->interfaceName, retCode);
      }
    }
    if (ifc->flowHandle!=NULL)
    {
      retCode = TcDeleteFlow(ifc->flowHandle);
      if(retCode != NO_ERROR)
      {
        ortp_error("TC: %s: %d reported by TcDeleteFlow",ifc->interfaceName, retCode);
      }
    }
    if (ifc->ifcHandle!=NULL)
    {
      retCode = TcCloseInterface(ifc->ifcHandle);
      if (retCode !=  NO_ERROR)
      {
        ortp_error("TC: %s: %d reported by TcCloseInterface",ifc->interfaceName, retCode);
      }
    }
    ifc->FilterHandle=NULL;
    ifc->flowHandle=NULL;
    ifc->ifcHandle=NULL;

    ortp_message("TC: %s: flow and filter removed", ifc->interfaceName);
  }
  if (session->tc_qos.ifc!=NULL)
    ortp_free(session->tc_qos.ifc);
  session->tc_qos.ifc=NULL;
}

int rtp_session_set_dscp_with_tcapi(RtpSession *session, int dscp)
{
  long result;
  PTC_IFC_DESCRIPTOR pInterfaceBuffer = NULL;
  ULONG BufferSize;
  ULONG RemainingBufferSize;
  int idx;
  PTC_IFC_DESCRIPTOR pCurrentIfc;

  PTC_GEN_FLOW newFlow;

  // Create the new temp flow.
  LPQOS_OBJECT_HDR objHdr = NULL; 

  LPQOS_DS_CLASS pQOSClass;

  ULONG retCode;
  TC_GEN_FILTER GenericFilter;
  IP_PATTERN Pattern, Mask;

  char src6host[NI_MAXHOST];
  int recvport = 0;

  tc_qos_init();
  if (QosClientHandle==NULL)
    return -1; /* -> no admin rights? */

  if (session->rtp.rem_addrlen<=0)
    return -1;

  memset(src6host, 0, sizeof(src6host));

  if (((struct sockaddr_in *) &session->rtp.rem_addr)->sin_family == AF_INET)
    recvport = ntohs(((struct sockaddr_in *) &session->rtp.rem_addr)->sin_port);
  else
    recvport = ntohs(((struct sockaddr_in6 *) &session->rtp.rem_addr)->sin6_port);

  result = getnameinfo((struct sockaddr *) &session->rtp.rem_addr, session->rtp.rem_addrlen,
    src6host, NI_MAXHOST, NULL, 0, NI_NUMERICHOST);

  if (result != 0) {
    ortp_error("TC: getnameinfo failure: %s", gai_strerror(result));
    return -1;
  }

  // Create a filter. All traffic matching the filter will pass through the flow
  memset(&Pattern,0,sizeof(IP_PATTERN));
  memset(&Mask,0,sizeof(IP_PATTERN));
  GenericFilter.AddressType = NDIS_PROTOCOL_ID_TCP_IP; // Protocol stack is TCP/IP
  GenericFilter.PatternSize = sizeof(IP_PATTERN);
  GenericFilter.Pattern = &Pattern; // pattern to match, defined below
  GenericFilter.Mask = &Mask;
  // Filter pattern. This is where you can tweak the flow filter parameters
  // configured to match on all outgoing UDP packets.
  Pattern.Reserved1 = 0;
  Pattern.Reserved2 = 0;
  Pattern.SrcAddr = inet_addr("0.0.0.0"); //IP_ADDRESS;
  Pattern.DstAddr = inet_addr(src6host); //IP_ADDRESS;
  Pattern.tcSrcPort = 0;
  Pattern.tcDstPort = htons(recvport); 
  Pattern.ProtocolId = IPPROTO_UDP;
  Pattern.Reserved3[0] = 0;
  Pattern.Reserved3[1] = 0;
  Pattern.Reserved3[2] = 0;
  // Patterns mask
  Mask.Reserved1 = 0;
  Mask.Reserved2 = 0;
  Mask.SrcAddr = htonl( 0 );
  Mask.DstAddr = 0xFFFFFFFF;
  Mask.tcSrcPort = 0;
  Mask.tcDstPort = 0xFFFF;
  Mask.ProtocolId = 0xFF;
  Mask.Reserved3[0] = 0;
  Mask.Reserved3[1] = 0;
  Mask.Reserved3[2] = 0;

  //Find traffic control enabled interfaces on the cmachine
  //Enumerate the available protocols
  BufferSize = 4096;
  pInterfaceBuffer = (PTC_IFC_DESCRIPTOR)ortp_malloc(BufferSize);
  result = TcEnumerateInterfaces (QosClientHandle, &BufferSize,pInterfaceBuffer);
  while (result==ERROR_INSUFFICIENT_BUFFER)
  {
    BufferSize *= 2;
    pInterfaceBuffer = (PTC_IFC_DESCRIPTOR)ortp_realloc(pInterfaceBuffer, BufferSize);
    result = TcEnumerateInterfaces (QosClientHandle, &BufferSize,pInterfaceBuffer);
  }
  if (result != NO_ERROR) 
  {
    ortp_error("TC: %d reported by TcEnumerateInterfaces",result);
    ortp_free(pInterfaceBuffer);
    return E_FAIL;
  }

  if (BufferSize == 0)
  {
    ortp_error("TC: no traffic control interfaces with QOS Packet Scheduler installed");
    ortp_free(pInterfaceBuffer);
    return E_FAIL;
  }

  tc_qos_delete_flow(session);
  session->tc_qos.num_ifc=0;

  RemainingBufferSize = BufferSize;
  pCurrentIfc = pInterfaceBuffer;
  while (RemainingBufferSize > 0)
  {
    RemainingBufferSize -= pInterfaceBuffer->Length;
    pInterfaceBuffer = (PTC_IFC_DESCRIPTOR) ((PBYTE) pInterfaceBuffer + pInterfaceBuffer->Length);
    session->tc_qos.num_ifc++;
  }

  if (session->tc_qos.num_ifc<=0)
  {
    ortp_error("TC: no interfaces available");
    ortp_free(pCurrentIfc);
    return -1;
  }

  newFlow = ( PTC_GEN_FLOW )ortp_malloc(sizeof (TC_GEN_FLOW )+ sizeof (QOS_DS_CLASS) + sizeof(QOS_OBJECT_HDR));

  session->tc_qos.ifc = (struct _tc_qos_ifc *)ortp_malloc(sizeof(struct _tc_qos_ifc)*session->tc_qos.num_ifc);
  memset(session->tc_qos.ifc, 0, sizeof(struct _tc_qos_ifc)*session->tc_qos.num_ifc);

  RemainingBufferSize = BufferSize;
  pInterfaceBuffer = pCurrentIfc;
  idx=0;
  while (RemainingBufferSize > 0)
  {
    struct _tc_qos_ifc *ifc=&session->tc_qos.ifc[idx];
    idx++;
    WideCharToMultiByte(CP_UTF8,0,pInterfaceBuffer->pInterfaceName, -1,ifc->interfaceName,sizeof(ifc->interfaceName), 0, 0 );

    result = TcOpenInterface(pInterfaceBuffer->pInterfaceName, QosClientHandle, NULL, &ifc->ifcHandle );

    if (result!= NO_ERROR)
    {
      ortp_error("TC: %s: %d reported by TcOpenInterface",ifc->interfaceName, result);
      RemainingBufferSize -= pInterfaceBuffer->Length;
      pInterfaceBuffer = (PTC_IFC_DESCRIPTOR) ((PBYTE) pInterfaceBuffer + pInterfaceBuffer->Length);
      continue;
    }

    //Set the Flow Spec
    newFlow->ReceivingFlowspec.DelayVariation = QOS_NOT_SPECIFIED;
    newFlow->ReceivingFlowspec.Latency = QOS_NOT_SPECIFIED;
    newFlow->ReceivingFlowspec.MaxSduSize = QOS_NOT_SPECIFIED;
    newFlow->ReceivingFlowspec.MinimumPolicedSize = QOS_NOT_SPECIFIED;
    newFlow->ReceivingFlowspec.PeakBandwidth = QOS_NOT_SPECIFIED;
    newFlow->ReceivingFlowspec.ServiceType = SERVICETYPE_BESTEFFORT;
    newFlow->ReceivingFlowspec.TokenBucketSize = QOS_NOT_SPECIFIED;
    newFlow->ReceivingFlowspec.TokenRate =QOS_NOT_SPECIFIED;

    newFlow->SendingFlowspec.DelayVariation = QOS_NOT_SPECIFIED;
    newFlow->SendingFlowspec.Latency = QOS_NOT_SPECIFIED;
    newFlow->SendingFlowspec.MaxSduSize = QOS_NOT_SPECIFIED;
    newFlow->SendingFlowspec.MinimumPolicedSize = QOS_NOT_SPECIFIED;
    newFlow->SendingFlowspec.PeakBandwidth = QOS_NOT_SPECIFIED;
    newFlow->SendingFlowspec.ServiceType=SERVICETYPE_BESTEFFORT;
    newFlow->SendingFlowspec.TokenBucketSize = QOS_NOT_SPECIFIED;
    newFlow->SendingFlowspec.TokenRate = QOS_NOT_SPECIFIED;

    newFlow->TcObjectsLength =sizeof(QOS_DS_CLASS) + sizeof(QOS_OBJECT_HDR); 

    pQOSClass = (LPQOS_DS_CLASS)(&(newFlow->TcObjects[0]));

    pQOSClass->ObjectHdr.ObjectType = QOS_OBJECT_DS_CLASS;
    pQOSClass->ObjectHdr.ObjectLength = sizeof(QOS_DS_CLASS);
    pQOSClass->DSField = 0x2E;
    objHdr = (LPQOS_OBJECT_HDR)((char *)&(newFlow->TcObjects[0]) + sizeof(QOS_DS_CLASS)); 

    objHdr->ObjectType = QOS_OBJECT_END_OF_LIST;
    objHdr->ObjectLength = sizeof(QOS_OBJECT_HDR); 

    // adds a new flow on the specified interface. The flow is like a channel,
    // that determines how to shape the traffic if it tunnelled through it.
    retCode = TcAddFlow(ifc->ifcHandle, NULL, 0, newFlow, &ifc->flowHandle );
    if(retCode != NO_ERROR)
    {
      ortp_error("TC: %s: %d reported by TcAddFlow",ifc->interfaceName, retCode);
      RemainingBufferSize -= pInterfaceBuffer->Length;
      pInterfaceBuffer = (PTC_IFC_DESCRIPTOR) ((PBYTE) pInterfaceBuffer + pInterfaceBuffer->Length);
      continue;
    }

    // Add filter to the flow. This point onwards all traffic matching the filter, will pass through the flow.
    retCode = TcAddFilter(ifc->flowHandle, &GenericFilter, &ifc->FilterHandle); 

    if (retCode!= NO_ERROR)
    {
      ortp_error("TC: %s: %d reported by in TcAddFilter", ifc->interfaceName, retCode);
      RemainingBufferSize -= pInterfaceBuffer->Length;
      pInterfaceBuffer = (PTC_IFC_DESCRIPTOR) ((PBYTE) pInterfaceBuffer + pInterfaceBuffer->Length);
      continue;
    }

    ortp_message("TC: %s: DSCP %2X for UDP->%s:%i", ifc->interfaceName, dscp, src6host, recvport);

    RemainingBufferSize -= pInterfaceBuffer->Length;
    pInterfaceBuffer = (PTC_IFC_DESCRIPTOR) ((PBYTE) pInterfaceBuffer + pInterfaceBuffer->Length);
  }

  ortp_free(newFlow);
  ortp_free(pCurrentIfc);
  return 0;
}
#else
void tc_qos_delete_flow(RtpSession *session)
{
}
#endif

#if 0

#include <qos.h>
#include <qossp.h>

struct all_objects {
  QOS_DESTADDR qos_dest;
  QOS_DS_CLASS qos_ds;
  QOS_OBJECT_HDR qos_hdr;
};

int rtp_session_set_dscp_BROKEN(RtpSession *session, int dscp)
{
  QOS qos;
  DWORD dwBytes;
  struct all_objects obj;
  int err;

  if (session->rtp.rem_addrlen<=0)
    return -1;

  memset(&obj, 0, sizeof(struct all_objects));

  qos.ReceivingFlowspec.ServiceType = SERVICE_BESTEFFORT | SERVICE_NO_QOS_SIGNALING;
  qos.ReceivingFlowspec.TokenRate = QOS_NOT_SPECIFIED;
  qos.ReceivingFlowspec.TokenBucketSize = QOS_NOT_SPECIFIED;
  qos.ReceivingFlowspec.PeakBandwidth = QOS_NOT_SPECIFIED;
  qos.ReceivingFlowspec.Latency	 = QOS_NOT_SPECIFIED;
  qos.ReceivingFlowspec.DelayVariation = QOS_NOT_SPECIFIED;
  qos.ReceivingFlowspec.MinimumPolicedSize = QOS_NOT_SPECIFIED;
  qos.ReceivingFlowspec.MaxSduSize = QOS_NOT_SPECIFIED;

  qos.SendingFlowspec.ServiceType = SERVICETYPE_BESTEFFORT | SERVICE_NO_QOS_SIGNALING;
  qos.SendingFlowspec.TokenRate = QOS_NOT_SPECIFIED;
  qos.SendingFlowspec.TokenBucketSize = QOS_NOT_SPECIFIED;
  qos.SendingFlowspec.PeakBandwidth = QOS_NOT_SPECIFIED;
  qos.SendingFlowspec.Latency	 = QOS_NOT_SPECIFIED;
  qos.SendingFlowspec.DelayVariation = QOS_NOT_SPECIFIED;
  qos.SendingFlowspec.MinimumPolicedSize = QOS_NOT_SPECIFIED;
  qos.SendingFlowspec.MaxSduSize = QOS_NOT_SPECIFIED;
  qos.ProviderSpecific.len = 0;
  qos.ProviderSpecific.buf = NULL;

  obj.qos_dest.ObjectHdr.ObjectType = QOS_OBJECT_DESTADDR;
  obj.qos_dest.ObjectHdr.ObjectLength = sizeof(QOS_DESTADDR);
  obj.qos_dest.SocketAddress = (SOCKADDR *)&session->rtp.rem_addr;
  obj.qos_dest.SocketAddressLength = session->rtp.rem_addrlen;

  obj.qos_ds.ObjectHdr.ObjectType = QOS_OBJECT_DS_CLASS;
  obj.qos_ds.ObjectHdr.ObjectLength = sizeof(QOS_DS_CLASS);
  obj.qos_ds.DSField = 0x2e;

  // Set the end of the list
  obj.qos_hdr.ObjectType = QOS_OBJECT_END_OF_LIST;
  obj.qos_hdr.ObjectLength = sizeof(QOS_OBJECT_HDR); 


  //qos.ProviderSpecific.len = sizeof(QOS_DS_CLASS)+sizeof(QOS_DESTADDR)+sizeof(QOS_OBJECT_HDR);
  //qos.ProviderSpecific.buf = (char*)&obj;
  qos.ProviderSpecific.len = sizeof(QOS_DS_CLASS);
  qos.ProviderSpecific.buf = (char*)&obj.qos_ds;

  err = WSAIoctl(session->rtp.socket, SIO_SET_QOS, &qos, sizeof(QOS)+qos.ProviderSpecific.len, NULL, 0, &dwBytes, NULL, NULL);
  if (err != 0)
  {
    ortp_error("WSAIoctl FAILED: ");
    return -1;
  } else {
    ortp_message("LIPSocket::SetTOSHigh() WSAIoctl(SIO_SET_QOS) OK");
    return 0;
  }
}
#endif

/**
*rtp_session_set_dscp:
*@session: a rtp session
*@dscp: desired DSCP PHB value
*
* Sets the DSCP (Differentiated Services Code Point) for outgoing RTP packets.
*
* Returns: 0 on success.
*
**/
int rtp_session_set_dscp(RtpSession *session, int idx, int dscp){
  int retval=0;
  int tos;
#ifdef ENABLE_RTP_QOS
  OSVERSIONINFOEX ovi;
#endif

  // Store new DSCP value if one is specified
  if (dscp>=0) session->dscp = dscp;

  if (session->dscp<=0)
    return 0;

  // Don't do anything if socket hasn't been created yet
  if (!sockisvalid(session->rtp.socketv4[idx])) return 0;

#ifdef ENABLE_RTP_QOS

  memset(&ovi, 0, sizeof(ovi));
  ovi.dwOSVersionInfoSize = sizeof(ovi);
  GetVersionEx((LPOSVERSIONINFO) & ovi);

  //This works on vista with admin rights
  if (session->dscp==0 && ovi.dwMajorVersion > 5)
    tos=QOSTrafficTypeBestEffort;
  else if (session->dscp==0x8 && ovi.dwMajorVersion > 5)
    tos=QOSTrafficTypeBackground;
  else if (session->dscp==0x28 && ovi.dwMajorVersion > 5)
    tos=QOSTrafficTypeAudioVideo;
  else if (session->dscp==0x38 && ovi.dwMajorVersion > 5)
    tos=QOSTrafficTypeVoice;
  else {
    //if (rtp_session_set_dscp_with_tcapi(session, session->dscp)==0)
    //  return 0;
    //else
    //{
    //  if (QosClientHandle!=NULL)
    //  {
    //    /* we have admin rights but remote addr is most probably not yet defined */
    //    /* postpone dscp activation for later */
    //    return -1;
    //  }
    //}
  }

  ortp_message("check OS support for qwave.lib: %i %i %i",
    ovi.dwMajorVersion, ovi.dwMinorVersion, ovi.dwBuildNumber);
  if (ovi.dwMajorVersion > 5) {
    HRESULT hr = E_FAIL;
    __try {
      hr = __HrLoadAllImportsForDll("qwave.dll");
    }
    __except(EXCEPTION_EXECUTE_HANDLER) {
      hr = E_FAIL;
    }
    if (! SUCCEEDED(hr)) {
      ortp_warning("Failed to load qwave.dll: no QoS available" );
    }
    else
    {
      if (session->dscp==0)
        tos=QOSTrafficTypeBestEffort;
      else if (session->dscp<=0x8)
        tos=QOSTrafficTypeBackground;
      else if (session->dscp<=0x28)
        tos=QOSTrafficTypeAudioVideo;
      else if (session->dscp<=0x38)
        tos=QOSTrafficTypeVoice;
      else
        tos=QOSTrafficTypeExcellentEffort; /* 0x28 */

      if (session->rtp.QoSHandle==NULL) {
        QOS_VERSION version;
        BOOL QoSResult;

        version.MajorVersion = 1;
        version.MinorVersion = 0;

        QoSResult = QOSCreateHandle(&version, &session->rtp.QoSHandle);

        if (QoSResult != TRUE){
          ortp_error("QOSCreateHandle failed to create handle with error %d",
            GetLastError());
          retval=-1;
        }
      }
      if (session->rtp.QoSHandle!=NULL && session->rtp.rem_addrlen>0) {
        BOOL QoSResult;
        QoSResult = QOSAddSocketToFlow(
          session->rtp.QoSHandle, 
          session->rtp.socketv4[idx],
          (struct sockaddr*)&session->rtp.rem_addr,
          tos, 
          QOS_NON_ADAPTIVE_FLOW, 
          &session->rtp.QoSFlowID);

        if (QoSResult != TRUE){
          ortp_error("QOSAddSocketToFlow failed to add a flow with error %d", 
            GetLastError());
          retval=-1;
        } else{
#if 0
          DWORD dscp_val = session->dscp;
          QoSResult = QOSSetFlow(session->rtp.QoSHandle,
            session->rtp.QoSFlowID,
            /* -> available on windows 7/ no vista support */ 
            2, /* QOSSetOutgoingDSCPValue, */
            sizeof(dscp_val),
            &dscp_val,
            0,
            NULL);
          if (QoSResult != TRUE)
          {
            ortp_error("QOSSetFlow failed to set dscp value on flow with error %d", 
              GetLastError());
          }
          else
          {
            ortp_message("QOSSetFlow succeeded to set dscp value on flow");
          }
#endif
        }
      }
    }
  } else {
#endif
    // DSCP value is in the upper six bits of the TOS field
    tos = (session->dscp << 2) & 0xFC;
    switch (session->rtp.sockfamily[idx]) {
    case AF_INET:
      retval = setsockopt(session->rtp.socketv4[idx], IPPROTO_IP, IP_TOS, (SOCKET_OPTION_VALUE)&tos, sizeof(tos));
      break;
    case AF_INET6:
#	ifdef IPV6_TCLASS /*seems not defined by my libc*/
      retval = setsockopt(session->rtp.socketv4[idx], IPPROTO_IPV6, IPV6_TCLASS,
        (SOCKET_OPTION_VALUE)&tos, sizeof(tos));
#	else
      /*in case that works:*/
      retval = setsockopt(session->rtp.socketv4[idx], IPPROTO_IPV6, IP_TOS,
        (SOCKET_OPTION_VALUE)&tos, sizeof(tos));
#	endif
      break;
    default:
      retval=-1;
    }
#ifdef ENABLE_RTP_QOS
  }
#endif

  if (retval<0)
    ortp_warning("Failed to set DSCP value on socket.");

  return retval;
}


/**
*rtp_session_get_dscp:
*@session: a rtp session
*
* Returns the DSCP (Differentiated Services Code Point) for outgoing RTP packets.
*
**/
int rtp_session_get_dscp(const RtpSession *session)
{
  return session->dscp;
}


/**
*rtp_session_get_local_port:
*@session:	a rtp session for which rtp_session_set_local_addr() or rtp_session_set_remote_addr() has been called
*
*	This function can be useful to retrieve the local port that was randomly choosen by 
*	rtp_session_set_remote_addr() when rtp_session_set_local_addr() was not called.
*
*	Returns: the local port used to listen for rtp packets, -1 if not set.
**/

int rtp_session_get_local_port(const RtpSession *session){
  return (session->rtp.loc_port>0) ? session->rtp.loc_port : -1;
}

/**
*rtp_session_set_tsc_server:
*@session: a rtp session
*@tsc_handle: a tsc tunnel handle
*
* Returns: 0 on success.
*
**/
int rtp_session_set_tsc_server(RtpSession *session, void *tsc_handle, unsigned int tsc_redundancy_factor, int tsc_redundancy_lb) {
#ifdef TSC_SUPPORT
  session->tsc_handle = tsc_handle;
  session->tsc_redundancy_factor = tsc_redundancy_factor;
  session->tsc_redundancy_lb = tsc_redundancy_lb;
  return 0;
#else
  return -1;
#endif
}

static char * ortp_inet_ntoa(struct sockaddr *addr, int addrlen, char *dest, int destlen){
  int err;
  dest[0]=0;
  err=getnameinfo(addr,addrlen,dest,destlen,NULL,0,NI_NUMERICHOST);
  if (err!=0){
    ortp_warning("getnameinfo error: %s",gai_strerror(err));
  }
  return dest;
}

/**
*rtp_session_set_remote_addr:
*@session:		a rtp session freshly created.
*@addr:		a local IP address in the xxx.xxx.xxx.xxx form.
*@port:		a local port.
*
*	Sets the remote address of the rtp session, ie the destination address where rtp packet
*	are sent. If the session is recv-only or duplex, it also sets the origin of incoming RTP 
*	packets. Rtp packets that don't come from addr:port are discarded.
*
*	Returns: 0 on success.
**/
int
  rtp_session_set_remote_addr (RtpSession * session, int idx, const char * addr, int port){
    return rtp_session_set_remote_addr_full (session, idx, addr, port, port+1);
}

/**
*rtp_session_set_remote_addr_full:
*@session:		a rtp session freshly created.
*@addr:		a local IP address in the xxx.xxx.xxx.xxx form.
*@rtp_port:		a local rtp port.
*@rtcp_port:		a local rtcp port.
*
*	Sets the remote address of the rtp session, ie the destination address where rtp packet
*	are sent. If the session is recv-only or duplex, it also sets the origin of incoming RTP 
*	packets. Rtp packets that don't come from addr:port are discarded.
*
*	Returns: 0 on success.
**/

int
  rtp_session_set_remote_addr_full (RtpSession * session, int idx, const char * addr, int rtp_port, int rtcp_port)
{
  int err;
  struct addrinfo hints, *res0, *res;
  char num[8];

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_DGRAM;
  snprintf(num, sizeof(num), "%d", rtp_port);
  err = getaddrinfo(addr, num, &hints, &res0);
  if (err) {
    ortp_warning ("Error in socket address: %s", gai_strerror(err));
    return -1;
  }
  if (session->rtp.socketv4[idx] == -1){
    /* the session has not its socket bound, do it */
    ortp_message ("Setting random local addresses.");
    /* bind to an address type that matches the destination address */
    if (res0->ai_addr->sa_family==AF_INET6)
      err = rtp_session_set_local_addr (session, idx, "::", -1);
    else err=rtp_session_set_local_addr (session, idx, "0.0.0.0", -1);
    if (err<0) return -1;
  }

  err=1;
  for (res = res0; res; res = res->ai_next) {
    /* set a destination address that has the same type as the local address */
    if (res->ai_family==session->rtp.sockfamily[idx] ) {
      memcpy( &session->rtp.rem_addr, res->ai_addr, res->ai_addrlen);
      session->rtp.rem_addrlen=res->ai_addrlen;
      err=0;
      break;
    }
  }
  freeaddrinfo(res0);
  if (err) {
    ortp_warning("Could not set destination for RTP socket to %s:%i.",addr,rtp_port);
    return -1;
  }

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_DGRAM;
  snprintf(num, sizeof(num), "%d", rtcp_port);
  err = getaddrinfo(addr, num, &hints, &res0);
  if (err) {
    ortp_warning ("Error: %s", gai_strerror(err));
    return err;
  }
  err=1;
  for (res = res0; res; res = res->ai_next) {
    /* set a destination address that has the same type as the local address */
    if (res->ai_family==session->rtp.sockfamily[idx] ) {
      err=0;
      memcpy( &session->rtcp.rem_addr, res->ai_addr, res->ai_addrlen);
      session->rtcp.rem_addrlen=res->ai_addrlen;
      break;
    }
  }
  freeaddrinfo(res0);
  if (err) {
    ortp_warning("Could not set destination for RCTP socket to %s:%i.",addr,rtcp_port);
    return -1;
  }
  return 0;
}

int
  rtp_session_set_remote_addr_and_port(RtpSession * session, int idx, const char * addr, int rtp_port, int rtcp_port){
    return rtp_session_set_remote_addr_full(session,idx,addr,rtp_port,rtcp_port);
}

void rtp_session_set_transports(RtpSession *session, struct _RtpTransport *rtptr)
{
  if (session->rtp.tr!=NULL)
    ortp_transport_free(session->rtp.tr);
  session->rtp.tr = 0;

  session->rtp.tr = rtptr;
#if 0
  if (rtptr)
    rtptr->session=session;
#endif

  if (rtptr)
    session->flags|=(RTP_SESSION_USING_TRANSPORT);
  else session->flags&=~(RTP_SESSION_USING_TRANSPORT);
}

void rtp_session_add_module(RtpSession *session, struct _RtpModule *module)
{
  session->modules_tables = o_list_append(session->modules_tables, module);
}

RtpModule *rtp_session_get_module(RtpSession *session, const char *name){
  OList *elem;
  for (elem=session->modules_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpModule *s=(RtpModule*) elem->data;
    if (strcmp(name,s->name)==0){
      return s;
    }
  }
  return NULL;
}

int rtp_session_remove_module(RtpSession *session, const char *name)
{
  OList *elem;
  for (elem=session->modules_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpModule *module=(RtpModule*) elem->data;
    if (strcmp(name,module->name)==0){
      session->modules_tables = o_list_remove(session->modules_tables, module);
      ortp_module_free(module);
      return 0;
    }
  }
  return -1;
}

void rtp_session_module_remove_all(RtpSession *session) {
  OList *elem;
  for (elem=session->modules_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpModule *module=(RtpModule*) elem->data;
    ortp_module_free(module);
  }
  session->modules_tables = o_list_free(session->modules_tables);
}

/**
*rtp_session_flush_sockets:
*@session: a rtp session
*
* Flushes the sockets for all pending incoming packets.
* This can be usefull if you did not listen to the stream for a while
* and wishes to start to receive again. During the time no receive is made
* packets get bufferised into the internal kernel socket structure.
*
**/
void rtp_session_flush_sockets(RtpSession *session){
  unsigned char trash[4096];
  struct sockaddr_storage from;
  socklen_t fromlen=sizeof(from);
  int i;

  if (rtp_session_using_transport(session, rtp))
  {
    mblk_t *trashmp=esballoc(trash,sizeof(trash),0,NULL);

    while (session->rtp.tr->t_rtp_recvfrom(session,trashmp,0,(struct sockaddr *)&from,&fromlen)>0){};

    while (session->rtp.tr->t_rtcp_recvfrom(session,trashmp,0,(struct sockaddr *)&from,&fromlen)>0){};
    freemsg(trashmp);
    return;
  }

  for (i=0;i<10;i++) {
    if (sockisvalid(session->rtp.socketv4[i])){
      while (recvfrom(session->rtp.socketv4[i],(char*)trash,sizeof(trash),0,(struct sockaddr *)&from,&fromlen)>0){};
    }
    if (sockisvalid(session->rtcp.socketv4[i])){
      while (recvfrom(session->rtcp.socketv4[i],(char*)trash,sizeof(trash),0,(struct sockaddr*)&from,&fromlen)>0){};
    }
  }
}


#ifdef USE_SENDMSG 
#define MAX_IOV 30
static int rtp_sendmsg(int sock,mblk_t *m, struct sockaddr *rem_addr, int addr_len){
  int error;
  struct msghdr msg;
  struct iovec iov[MAX_IOV];
  int iovlen;
  for(iovlen=0; iovlen<MAX_IOV && m!=NULL; m=m->b_cont,iovlen++){
    iov[iovlen].iov_base=m->b_rptr;
    iov[iovlen].iov_len=m->b_wptr-m->b_rptr;
  }
  if (iovlen==MAX_IOV){
    ortp_error("Too long msgb, didn't fit into iov, end discarded.");
  }
  msg.msg_name=(void*)rem_addr;
  msg.msg_namelen=addr_len;
  msg.msg_iov=&iov[0];
  msg.msg_iovlen=iovlen;
  msg.msg_control=NULL;
  msg.msg_controllen=0;
  msg.msg_flags=0;
  error=(int)sendmsg(sock,&msg,0);
  return error;
}
#endif	

#define IP_UDP_OVERHEAD (20+8)

static void update_sent_bytes(RtpSession*s, int nbytes){
  if (s->rtp.sent_bytes==0){
    gettimeofday(&s->rtp.send_bw_start,NULL);
  }
  s->rtp.sent_bytes+=nbytes+IP_UDP_OVERHEAD;
}

static void update_recv_bytes(RtpSession*s, int nbytes){
  if (s->rtp.recv_bytes==0){
    gettimeofday(&s->rtp.recv_bw_start,NULL);
  }
  s->rtp.recv_bytes+=nbytes+IP_UDP_OVERHEAD;
}

static void update_rtcp_sent_bytes(RtpSession*s, int nbytes){
  if (s->rtcp.sent_bytes==0){
    gettimeofday(&s->rtcp.send_bw_start,NULL);
  }
  s->rtcp.sent_bytes+=nbytes+IP_UDP_OVERHEAD;
}

static void update_rtcp_recv_bytes(RtpSession*s, int nbytes){
  if (s->rtcp.recv_bytes==0){
    gettimeofday(&s->rtcp.recv_bw_start,NULL);
  }
  s->rtcp.recv_bytes+=nbytes+IP_UDP_OVERHEAD;
}

int
  rtp_session_rtp_send (RtpSession * orig_session, mblk_t * m)
{
  int error;
  int i;
  rtp_header_t *hdr;
  RtpSession * session = orig_session;

  /* if bundle is used, use FIRST RtpMedia from bundle */
  if (orig_session->bundle && session->use_bundle) {
    RtpMedia *rtpMedia=NULL;
    OList *elem;
    elem=orig_session->bundle->rtpMedias;
    if (elem!=NULL)
      rtpMedia=(RtpMedia*) elem->data;
    if (rtpMedia!=NULL)
      session = &rtpMedia->session;
  }

  if (session==NULL) {
    freemsg (m);
    return -1;
  }

  hdr = (rtp_header_t *) m->b_rptr;
  /* perform host to network conversions */
  hdr->ssrc = htonl (hdr->ssrc);
  hdr->timestamp = htonl (hdr->timestamp);
  hdr->seq_number = htons (hdr->seq_number);
  for (i = 0; i < hdr->cc; i++)
    hdr->csrc[i] = htonl (hdr->csrc[i]);

  error = msgdsize(m);
  {
    OList *elem;
    for (elem=session->modules_tables;elem!=NULL;elem=o_list_next(elem)){
      RtpModule *module=(RtpModule*) elem->data;
      error = module->t_rtp_sendto(session, module, m);
      if (error<0)
        break;
    }
  }

  if (error<0) {
    /* failure happened in module->t_rtp_sendto */
    if (session->on_network_error.count>0){
      rtp_signal_table_emit3(&session->on_network_error,(uintptr_t)"Error converting RTP packet",INT_TO_POINTER(error));
    } /* else ortp_warning ("Error converting rtp packet; error=%i", error); */
    freemsg (m);
  } else {
    error = rtp_session_data_send(orig_session,m, 0);
  }
  return error;
}

int
  rtp_session_data_send (RtpSession * orig_session, mblk_t * m, int packettype)
{
  int initial_packet_size;
  int error;
  struct sockaddr *destaddr;
  socklen_t destlen;
  ortp_socket_t sockfd=0;
  RtpSession * session = orig_session;

  /* if bundle is used, use FIRST RtpMedia from bundle */
  if (orig_session->bundle && session->use_bundle) {
    RtpMedia *rtpMedia=NULL;
    OList *elem;
    elem=orig_session->bundle->rtpMedias;
    if (elem!=NULL)
      rtpMedia=(RtpMedia*) elem->data;
    if (rtpMedia!=NULL)
      session = &rtpMedia->session;
  }

  destaddr=(struct sockaddr*)&session->rtp.rem_addr;
  destlen=session->rtp.rem_addrlen;

  {
    int idx;
    for (idx=0;sockisvalid(session->rtp.socketv4[idx]);idx++) {
      if (session->rtp.sockfamily[idx]==destaddr->sa_family) {
        sockfd=session->rtp.socketv4[idx];
        break;
      }
    }
  }

  initial_packet_size = msgdsize(m);

  if (rtp_session_using_transport(session, rtp)){
    error = (session->rtp.tr->t_rtp_sendto) (session,m,0,destaddr,destlen);
  }else{
#ifdef USE_SENDMSG
    error=rtp_sendmsg(sockfd,m,destaddr,destlen);
#else
    if (m->b_cont!=NULL)
      msgpullup(m,-1);
#ifdef TSC_SUPPORT
    if (session->tsc_handle) {
      error = tsc_sendto (sockfd, (char*)m->b_rptr, (int) (m->b_wptr - m->b_rptr),
        0,destaddr,destlen);
    }
    else {
      error = sendto (sockfd, (char*)m->b_rptr, (int) (m->b_wptr - m->b_rptr),
        0,destaddr,destlen);
    }
#else
    error = sendto (sockfd, (char*)m->b_rptr, (int) (m->b_wptr - m->b_rptr),
      0,destaddr,destlen);
#endif
#endif
  }
  if (error < 0){
    if (session->on_network_error.count>0){
      rtp_signal_table_emit3(&session->on_network_error,(uintptr_t)"Error sending RTP packet",INT_TO_POINTER(getSocketErrorCode()));
    }else ortp_warning ("Error sending rtp packet: %s ; socket=%i", getSocketError(), sockfd);
    session->rtp.send_errno=getSocketErrorCode();
  }else{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    if (packettype==0)
      BandwidthStatistics_update(&orig_session->rtp.sent_total_rtp_bitrate, error+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD+(error-initial_packet_size), &tv);
    else if (packettype==2)
      BandwidthStatistics_update(&orig_session->rtp.sent_total_other_bitrate, error+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD+(error-initial_packet_size), &tv);
    update_sent_bytes(orig_session,error);
  }

  freemsg (m);
  return error;
}

int
  rtp_session_rtcp_send (RtpSession * orig_session, mblk_t * m)
{
  int initial_packet_size;
  int error=0;
  ortp_socket_t sockfd;
  struct sockaddr *destaddr;
  socklen_t destlen;
  RtpSession * session = orig_session;

  /* if bundle is used, use FIRST RtpMedia from bundle */
  if (session->bundle!=NULL) rtp_bundle_lock(session->bundle);
  if (orig_session->bundle && session->use_bundle) {
    RtpMedia *rtpMedia=NULL;
    OList *elem;
    elem=orig_session->bundle->rtpMedias;
    if (elem!=NULL)
      rtpMedia=(RtpMedia*) elem->data;
    if (rtpMedia!=NULL)
      session = &rtpMedia->session;
  }
  if (session==NULL) {
    freemsg (m);
    if (session->bundle!=NULL) rtp_bundle_unlock(session->bundle);
    return -1;
  }

  sockfd=0;
  destaddr=(struct sockaddr*)&session->rtcp.rem_addr;
  destlen=session->rtcp.rem_addrlen;

  {
    int idx;
    for (idx=0;sockisvalid(session->rtcp.socketv4[idx]);idx++) {
      if (session->rtcp.sockfamily[idx]==destaddr->sa_family) {
        sockfd=session->rtcp.socketv4[idx];
        break;
      }
    }
  }

  ortp_debug("RTCP sending %i", orig_session->rtp.loc_port+1);

  if (session->rtcp_mux) {
    sockfd=0;
    destaddr=(struct sockaddr*)&session->rtp.rem_addr;
    destlen=session->rtp.rem_addrlen;

    {
      int idx;
      for (idx=0;sockisvalid(session->rtp.socketv4[idx]);idx++) {
        if (session->rtp.sockfamily[idx]==destaddr->sa_family) {
          sockfd=session->rtp.socketv4[idx];
          break;
        }
      }
    }
  }

  if (!session->rtcp.enabled) {
    freemsg(m);
    if (session->bundle!=NULL) rtp_bundle_unlock(session->bundle);
    return 0;
  }

  error = msgdsize(m);
  initial_packet_size = error;
  {
    OList *elem;
    for (elem=session->modules_tables;elem!=NULL;elem=o_list_next(elem)){
      RtpModule *module=(RtpModule*) elem->data;
      error = module->t_rtcp_sendto(session, module, m);
      if (error<0)
        break;
    }
  }

  if (error<0) {
    /* failure happened in module->t_rtcp_sendto */
  } else if (rtp_session_using_transport(session, rtp)) {
    error = (session->rtp.tr->t_rtcp_sendto) (session, m, 0, destaddr, destlen);
  } else if ( (sockisvalid(sockfd) && (destlen>0) )) {
#ifdef USE_SENDMSG
    error=rtp_sendmsg(sockfd,m,destaddr, destlen);
#else
    if (m->b_cont!=NULL){
      msgpullup(m,-1);
    }
#ifdef TSC_SUPPORT
    if (session->tsc_handle) {
      error = tsc_sendto (sockfd, (char*)m->b_rptr,
        (int) (m->b_wptr - m->b_rptr), 0,
        destaddr, destlen);
    }
    else {
      error = sendto (sockfd, (char*)m->b_rptr,
        (int) (m->b_wptr - m->b_rptr), 0,
        destaddr, destlen);
    }
#else
    error = sendto (sockfd, (char*)m->b_rptr,
      (int) (m->b_wptr - m->b_rptr), 0,
      destaddr, destlen);
#endif
#endif
  } else ortp_message("Not sending RTCP report: sockfd=%i, rem_addrlen=%i",sockfd,session->rtcp.rem_addrlen);

  if (error < 0){
    char host[65];
    if (session->on_network_error.count>0){
      rtp_signal_table_emit3(&session->on_network_error,(uintptr_t)"Error sending RTCP packet",INT_TO_POINTER(getSocketErrorCode()));
    }else ortp_warning ("Error sending RTCP packet: %s ; socket=%i; addr=%s", getSocketError(), sockfd, ortp_inet_ntoa((struct sockaddr*)&session->rtcp.rem_addr,session->rtcp.rem_addrlen,host,sizeof(host)) );
  }else{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    BandwidthStatistics_update(&orig_session->rtp.sent_total_rtcp_bitrate, error+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD+(error-initial_packet_size), &tv);
    update_rtcp_sent_bytes(orig_session,error);
  }
  if (session->bundle!=NULL) rtp_bundle_unlock(session->bundle);
  freemsg (m);
  return error;
}

int
  rtp_session_rtp_recv (RtpSession * session, int idx)
{
  int initial_packet_size;
  int error;
  ortp_socket_t sockfd=session->rtp.socketv4[idx];
  struct sockaddr_storage remaddr;
  socklen_t addrlen = sizeof (remaddr);
  mblk_t *mp;
  int type = RFC5764BIS_UNKNOWN;

  if (!sockisvalid(sockfd) && !rtp_session_using_transport(session, rtp)) return -1;  /*session has no sockets for the moment*/

  while (1)
  {
    int bufsz;

    initial_packet_size = 0;
    if (session->rtp.cached_mp==NULL)
      session->rtp.cached_mp = msgb_allocator_alloc(&session->allocator,session->recv_buf_size);
    mp=session->rtp.cached_mp;
    bufsz=(int) (mp->b_datap->db_lim - mp->b_datap->db_base);

    addrlen = sizeof (remaddr); /* revert size every call so it is large enough for next IPv6 buffer */
    if (rtp_session_using_transport(session, rtp)) 
      error = (session->rtp.tr->t_rtp_recvfrom)(session, mp, 0,
      (struct sockaddr *) &remaddr,
      &addrlen);
    else {
#ifdef TSC_SUPPORT
      if (session->tsc_handle) {
        error = tsc_recvfrom(sockfd, (char*)mp->b_wptr,
          bufsz, 0,
          (struct sockaddr *) &remaddr,
          &addrlen);
        if (error < 0 && tsc_get_errno() == EFAULT) {
          return -1;
        }
      }
      else {
        error = recvfrom(sockfd, (char*)mp->b_wptr,
          bufsz, 0,
          (struct sockaddr *) &remaddr,
          &addrlen);
      }
#else
      error = (int)recvfrom(sockfd, (char*)mp->b_wptr,
        bufsz, 0,
        (struct sockaddr *) &remaddr,
        &addrlen);
#endif
    }

    if (error > 0){
#ifdef FAKE_LOST
      {
        int val = rand() % 100 + 1;
        if  (val < 5 ) {
          ortp_warning("introduce lost packet");
          continue;
        }
      }
#endif
    }

    if (error > 0){
      OList *elem;
      unsigned char *ptr=mp->b_wptr;
      mp->b_wptr += error;

      initial_packet_size = error; /* original packet size */

      for (elem=session->modules_tables;elem!=NULL;elem=o_list_next(elem)){
        RtpModule *module=(RtpModule*) elem->data;
        error = module->t_rtp_recvfrom(session, module, mp);
        if (error<0)
          break;
      }
      /* revert! */
      mp->b_wptr = ptr;


      /*
      draft-ietf-avtcore-rfc5764-mux-fixes-11
      +----------------+
      |        [0..3] -+--> forward to STUN
      |                |
      |      [16..19] -+--> forward to ZRTP
      |                |
      |      [20..63] -+--> forward to DTLS
      |                |
      |      [64..79] -+--> forward to TURN Channel
      |                |
      |    [128..191] -+--> forward to RTP/RTCP
      +----------------+
      */
      {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        if ((mp->b_rptr[0] >= 0) && (mp->b_rptr[0] <= 3)) {
          /* STUN packet detected */
          type = RFC5764BIS_STUN;
          BandwidthStatistics_update(&session->rtp.recv_total_stun_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD,&tv);
          continue;
        } else if ((mp->b_rptr[0] >= 16) && (mp->b_rptr[0] <= 19)) {
          /* ZRTP packet detected */
          type = RFC5764BIS_ZRTP;
          BandwidthStatistics_update(&session->rtp.recv_total_other_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD,&tv);
          continue;
        } else if ((mp->b_rptr[0] >= 20) && (mp->b_rptr[0] <= 63)) {
          /* DTLS packet detected */
          type = RFC5764BIS_DTLS;
          /* DTLS packet detected */
          BandwidthStatistics_update(&session->rtp.recv_total_other_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD,&tv);
          continue;	/* avoids an infinite loop ! */
        } else if ((mp->b_rptr[0] >= 64) && (mp->b_rptr[0] <= 79)) {
          /* TURN Channel packet detected */
          type = RFC5764BIS_TURN;
          if (error>0 && initial_packet_size!=error)
            BandwidthStatistics_update(&session->rtp.recv_total_stun_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD+(initial_packet_size-error),&tv);
          else
            BandwidthStatistics_update(&session->rtp.recv_total_stun_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD,&tv);
        } else if ((mp->b_rptr[0] >= 128) && (mp->b_rptr[0] <= 191)) {
          /* rtp/rtcp channel packet detected */
          RtpSession *final_session;
          if (error<0) {
            BandwidthStatistics_update(&session->rtp.recv_total_bad_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD,&tv);
            continue;
          }

          if (error > 0){
            /* then parse the message and put on queue */
            mp->b_wptr+=error;
          }
          session->rtp.cached_mp=NULL; /* mp will be released by oRTP or rtp_session_demux_media or rtp_session_rtp_parse */
          final_session = rtp_session_demux_media(session->bundle, session, mp, (struct sockaddr*)&remaddr,addrlen);
          if (final_session!=NULL) {
            int rtp_overhead=0;
            int extra_overhead=0;
            type = rtp_session_rtp_parse (final_session, mp, (struct sockaddr*)&remaddr,addrlen, initial_packet_size, &rtp_overhead);

            if (error>=0)
              extra_overhead = initial_packet_size-error;
            if (type==RFC5764BIS_RTP) {
              rtp_overhead+=IP_UDP_OVERHEAD+(extra_overhead);
              BandwidthStatistics_update(&final_session->rtp.recv_total_rtppayload_bitrate, initial_packet_size+IP_UDP_OVERHEAD, rtp_overhead,&tv);
              if (final_session->rtp.cur_bitrate_table.recv_overhead_realtime_bytes==0)
                final_session->rtp.cur_bitrate_table.recv_overhead_realtime_bytes = rtp_overhead; /* init value */
              final_session->rtp.cur_bitrate_table.recv_overhead_realtime_bytes = (15*final_session->rtp.cur_bitrate_table.recv_overhead_realtime_bytes+rtp_overhead)/16;

              BandwidthStatistics_update(&final_session->rtp.recv_total_rtp_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD+(extra_overhead),&tv);
              update_recv_bytes(final_session,initial_packet_size);
            }else if (type==RFC5764BIS_RTCP) {
              BandwidthStatistics_update(&final_session->rtp.recv_total_rtcp_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD+(extra_overhead),&tv);
              update_rtcp_recv_bytes(final_session,initial_packet_size);
            } else {
              BandwidthStatistics_update(&final_session->rtp.recv_total_bad_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD+(extra_overhead),&tv);
            }
          } else {
            BandwidthStatistics_update(&session->rtp.recv_total_bad_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD,&tv);
            continue;
          }
        } else {
          BandwidthStatistics_update(&session->rtp.recv_total_bad_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD,&tv);
          ortp_warning("data received, but not able to demux it. (%i) (error=%i)", mp->b_rptr[0], error);
          continue;
        }
      }

    }

    if (error <= 0) {
      int errnum;
#ifdef TSC_SUPPORT
      if (session->tsc_handle) {
        errnum=tsc_get_errno();
      }else{
        errnum=getSocketErrorCode();
      }
#else
      errnum=getSocketErrorCode();
#endif
      if (error == 0){
        /*0 can be returned by RtpTransport functions in case of EWOULDBLOCK*/
        /*we ignore it*/
        /*ortp_warning
        ("rtp_recv: strange... recv() returned zero.");*/
      }
      else if (!is_would_block_error(errnum))
      {
        if (session->on_network_error.count>0){
          rtp_signal_table_emit3(&session->on_network_error,(uintptr_t)"Error receiving RTP packet",INT_TO_POINTER(getSocketErrorCode()));
        }else {
#ifdef TSC_SUPPORT
          if (session->tsc_handle) {
            ortp_warning("Error receiving RTP packet: %i.",tsc_get_errno());
          } else {
            ortp_warning("Error receiving RTP packet: %s.",getSocketError());
          }
#else
          ortp_warning("Error receiving RTP packet: %s.",getSocketError());
#endif
        }
        session->rtp.recv_errno=errnum;
      }
      /* don't free the cached_mp, it will be reused next time */
      return -1;	/* avoids an infinite loop ! */
    }
  }
  return error;
}

void rtp_session_notify_inc_rtcp(RtpSession *session, mblk_t *m){
  if (session->eventqs!=NULL){
    OrtpEvent *ev=ortp_event_new(ORTP_EVENT_RTCP_PACKET_RECEIVED);
    OrtpEventData *d=ortp_event_get_data(ev);
    d->packet=m;
    rtp_session_dispatch_event(session,ev);
  }
  else freemsg(m);  /* avoid memory leak */
}

int
  rtp_session_rtcp_recv (RtpSession * session, int idx)
{
  int initial_packet_size;
  int error;
  struct sockaddr_storage remaddr;
  socklen_t addrlen=0;
  mblk_t *mp;

  if (!sockisvalid(session->rtcp.socketv4[idx]) && !rtp_session_using_transport(session, rtp)) return -1;  /*session has no rtcp sockets for the moment*/


  while (1)
  {
    if (session->rtcp.cached_mp==NULL)
      session->rtcp.cached_mp = allocb (RTCP_MAX_RECV_BUFSIZE, 0);

    mp=session->rtcp.cached_mp;

    addrlen=sizeof (remaddr);

    if (rtp_session_using_transport(session, rtp))
      error=(session->rtp.tr->t_rtcp_recvfrom)(session, mp, 0,
      (struct sockaddr *) &remaddr,
      &addrlen);
    else {
#ifdef TSC_SUPPORT
      if (session->tsc_handle) {
        error=(int)tsc_recvfrom (session->rtcp.socketv4[idx],(char*) mp->b_wptr,
          RTCP_MAX_RECV_BUFSIZE, 0,
          (struct sockaddr *) &remaddr,
          &addrlen);
        if (error < 0 && tsc_get_errno() == EFAULT) {
          return -1;
        }
      }
      else {
        error=(int)recvfrom (session->rtcp.socketv4[idx],(char*) mp->b_wptr,
          RTCP_MAX_RECV_BUFSIZE, 0,
          (struct sockaddr *) &remaddr,
          &addrlen);
      }
#else
      error=(int)recvfrom (session->rtcp.socketv4[idx],(char*) mp->b_wptr,
        RTCP_MAX_RECV_BUFSIZE, 0,
        (struct sockaddr *) &remaddr,
        &addrlen);
#endif
    }

    if (error > 0){
      OList *elem;
      unsigned char *ptr=mp->b_wptr;
      mp->b_wptr += error;

      initial_packet_size = error;

      for (elem=session->modules_tables;elem!=NULL;elem=o_list_next(elem)){
        RtpModule *module=(RtpModule*) elem->data;
        error = module->t_rtcp_recvfrom(session, module, mp);
        if (error<0) {
          break;
        }
      }
      /* revert! */
      mp->b_wptr = ptr;

	  {
		  /* the bitrates for those packets are located in RTP stat... anyway... */
		  struct timeval tv;
		  gettimeofday(&tv, NULL);
		  if ((mp->b_rptr[0] >= 0) && (mp->b_rptr[0] <= 3)) {
			  /* STUN packet detected */
			  //BandwidthStatistics_update(&session->rtp.recv_total_stun_bitrate, initial_packet_size + IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
			  continue;
		  }
		  else if ((mp->b_rptr[0] >= 16) && (mp->b_rptr[0] <= 19)) {
			  /* ZRTP packet detected */
			  //BandwidthStatistics_update(&session->rtp.recv_total_other_bitrate, initial_packet_size + IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
			  continue;
		  }
		  else if ((mp->b_rptr[0] >= 20) && (mp->b_rptr[0] <= 63)) {
			  /* DTLS packet detected */
			  //BandwidthStatistics_update(&session->rtp.recv_total_other_bitrate, initial_packet_size + IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
			  continue;	/* avoids an infinite loop ! */
		  }
		  else if ((mp->b_rptr[0] >= 64) && (mp->b_rptr[0] <= 79)) {
			  /* TURN Channel packet detected */
			  //if (error > 0 && initial_packet_size != error)
			  //  BandwidthStatistics_update(&session->rtp.recv_total_stun_bitrate, initial_packet_size + IP_UDP_OVERHEAD, IP_UDP_OVERHEAD + (initial_packet_size - error), &tv);
			  //else
			  //  BandwidthStatistics_update(&session->rtp.recv_total_stun_bitrate, initial_packet_size + IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
		  }
		  else if ((mp->b_rptr[0] >= 128) && (mp->b_rptr[0] <= 191)) {
			  /* rtp/rtcp channel packet detected */
			  if (error < 0) {
				  //BandwidthStatistics_update(&session->rtp.recv_total_bad_bitrate, initial_packet_size + IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
				  continue;
			  }
		  }
		  else {
			  //BandwidthStatistics_update(&session->rtp.recv_total_bad_bitrate, initial_packet_size + IP_UDP_OVERHEAD, IP_UDP_OVERHEAD, &tv);
			  ortp_warning("data received, but not able to demux it. (%i) (error=%i)", mp->b_rptr[0], error);
			  continue;
		  }
	  }
    }

    if (error > 0)
    {
      rtcp_common_header_t *rtcp;
      struct timeval tv;
      gettimeofday(&tv, NULL);
      BandwidthStatistics_update(&session->rtp.recv_total_rtcp_bitrate, initial_packet_size+IP_UDP_OVERHEAD, IP_UDP_OVERHEAD+(initial_packet_size-error),&tv);
      update_rtcp_recv_bytes(session,initial_packet_size);

      mp->b_wptr += error;
      /* SKIP NON VERSION 2 */
      rtcp=(rtcp_common_header_t *)mp->b_rptr;
      /* post an event to notify the application*/
      if (rtcp->version==2)
      {
        mblk_t *mp2 = dupb(mp);
        rtp_session_notify_inc_rtcp(session,mp);
        if (mp2!=NULL)
        {
          /* RTCP PARSE may access the other media stream */
          /* TODO: implement a QUEUE and process in the correct thread */
          //BUNDLE if (session->bundle!=NULL)
          //BUNDLE   rtp_bundle_lock(session->bundle);
          rtp_session_rtcp_parse(session,mp2); /* (this modify the content of RTCP data so we work on duplicate) */
          //BUNDLE if (session->bundle!=NULL)
          //BUNDLE   rtp_bundle_unlock(session->bundle);
          freemsg(mp2);
        }
      }
      session->rtcp.cached_mp=NULL;
      if (session->symmetric_rtp){
        /* store the sender rtp address to do symmetric RTP */
        if (addrlen!=session->rtcp.rem_addrlen
          || memcmp(&session->rtcp.rem_addr,&remaddr, addrlen)!=0)
        {
          ortp_message("Symmetric RTCP with initial RTCP/data packet! %i", session->rtp.loc_port+1);

          memcpy(&session->rtcp.rem_addr,&remaddr,addrlen);
          session->rtcp.rem_addrlen=addrlen;
        }
      }
    }
    else
    {
      int errnum;
#ifdef TSC_SUPPORT
      if (session->tsc_handle) {
        errnum=tsc_get_errno();
      }else{
        errnum=getSocketErrorCode();
      }
#else
      errnum=getSocketErrorCode();
#endif

      if (error == 0)
      {
#if 0
        ortp_warning
          ("rtcp_recv: strange... recv() returned zero.");
#endif
      }
      else if (!is_would_block_error(errnum))
      {
        if (session->on_network_error.count>0){
          rtp_signal_table_emit3(&session->on_network_error,(uintptr_t)"Error receiving RTCP packet",INT_TO_POINTER(errnum));
        }else {
#ifdef TSC_SUPPORT
          if (session->tsc_handle) {
            ortp_warning("Error receiving RTCP packet: %i.",tsc_get_errno());
          } else {
            ortp_warning("Error receiving RTCP packet: %s.",getSocketError());
          }
#else
          ortp_warning("Error receiving RTCP packet: %s.",getSocketError());
#endif
        }
      }
      /* don't free the cached_mp, it will be reused next time */
      return -1;	/* avoids an infinite loop ! */
    }
  }
  return error;
}

