/*
  The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	Copyright (C) 2011-2016  Aymeric Moizard <amoizard@gmail.com>
*/

#include <ortp/rtptransport.h>
#include <ortp/ortp.h>
#include <ortp/rtpsession.h>

struct _OList *modules_tables=NULL;

struct _OList *transport_tables=NULL;


RtpModule *ortp_module_new(const char *name) {
  OList *elem;
  for (elem=modules_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpModule *s=(RtpModule*) elem->data;
    if (strcmp(name,s->name)==0){
      RtpModule *tr = ortp_new0(RtpModule,1);
      if (tr==NULL)
        return NULL;
      memcpy(tr, s, sizeof(RtpModule));
      return tr;
    }
  }
  return NULL;
}

void ortp_module_add(RtpModule *module) {
  modules_tables = o_list_append(modules_tables, module);
  if (module->t_global_init!=NULL)
    module->t_global_init();
}

void ortp_module_remove_all(void) {
  OList *elem;
  for (elem=modules_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpModule *s=(RtpModule*) elem->data;
    if (s->t_global_shutdown!=NULL)
      s->t_global_shutdown();
    ortp_free(elem->data);
    elem->data = NULL;
  }
  modules_tables = o_list_free(modules_tables);
}

int ortp_module_set_option(RtpModule *module, int opt, void *value){
  if (module==NULL)
    return 0;
  if (module->t_set_option==NULL)
    return 0;
  return module->t_set_option(module, opt, value);
}

bool_t ortp_module_is_supported(const char *name) {
  OList *elem;
  for (elem=modules_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpModule *s=(RtpModule*) elem->data;
    if (strcmp(name,s->name)==0){
      return TRUE;
    }
  }
  return FALSE;
}

void ortp_module_free(RtpModule *module) {
  //module->t_release(module);
  if (module!=NULL)
  {
    module->t_destroy(module);
    ortp_free(module);
  }
}


RtpTransport *ortp_transport_new(const char *name) {
  OList *elem;
  for (elem=transport_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpTransport *s=(RtpTransport*) elem->data;
    if (strcmp(name,s->name)==0){
      RtpTransport *tr = ortp_new0(RtpTransport,1);
      if (tr==NULL)
        return NULL;
      memcpy(tr, s, sizeof(RtpTransport));
      return tr;
    }
  }
  return NULL;
}

void ortp_transport_add(RtpTransport *transport) {
  transport_tables = o_list_append(transport_tables, transport);
  if (transport->t_global_init!=NULL)
    transport->t_global_init();
}

void ortp_transport_remove_all(void) {
  OList *elem;
  for (elem=transport_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpTransport *s=(RtpTransport*) elem->data;
    if (s->t_global_shutdown!=NULL)
      s->t_global_shutdown();
    ortp_free(elem->data);
    elem->data = NULL;
  }
  transport_tables = o_list_free(transport_tables);
}

int ortp_transport_set_option(RtpTransport *transport, int opt, void *value){
  if (transport==NULL)
    return 0;
  if (transport->t_set_option==NULL)
    return 0;
  return transport->t_set_option(transport, opt, value);
}

bool_t ortp_transport_is_supported(const char *name) {
  OList *elem;
  for (elem=transport_tables;elem!=NULL;elem=o_list_next(elem)){
    RtpTransport *s=(RtpTransport*) elem->data;
    if (strcmp(name,s->name)==0){
      return TRUE;
    }
  }
  return FALSE;
}

void ortp_transport_free(RtpTransport *transport) {
  //transport->t_release(transport);
  if (transport!=NULL)
  {
    transport->t_destroy(transport);
    ortp_free(transport);
  }
}


#ifdef WIN32
typedef int (__stdcall *on_func_rtprewrite_protect) (int did, unsigned char *data, int *slen);
typedef int (__stdcall *on_func_rtcprewrite_protect) (int did, unsigned char *data, int *slen);
typedef int (__stdcall *on_func_rtprewrite_unprotect) (int did, unsigned char *data, int *slen);
typedef int (__stdcall *on_func_rtcprewrite_unprotect) (int did, unsigned char *data, int *slen);
#else
typedef int (*on_func_rtprewrite_protect) (int did, unsigned char *data, int *slen);
typedef int (*on_func_rtcprewrite_protect) (int did, unsigned char *data, int *slen);
typedef int (*on_func_rtprewrite_unprotect) (int did, unsigned char *data, int *slen);
typedef int (*on_func_rtcprewrite_unprotect) (int did, unsigned char *data, int *slen);
#endif

typedef struct {
  int val;
  int padding_size;
  on_func_rtprewrite_protect func_rtp_protect;
  on_func_rtcprewrite_protect func_rtcp_protect;
  on_func_rtprewrite_unprotect func_rtp_unprotect;
  on_func_rtcprewrite_unprotect func_rtcp_unprotect;
} rtprewrite_t;

static int rtprewrite_protect(rtprewrite_t *rtprewrite, unsigned char *data, int *len) {
  if (rtprewrite==NULL)
    return 0;
  return rtprewrite->func_rtp_protect(rtprewrite->val, data, len);
}

static int rtprewrite_unprotect(rtprewrite_t *rtprewrite, unsigned char *data, int *len) {
  if (rtprewrite==NULL)
    return 0;
  return rtprewrite->func_rtp_unprotect(rtprewrite->val, data, len);
}

static int rtprewrite_protect_rtcp(rtprewrite_t *rtprewrite, unsigned char *data, int *len) {
  if (rtprewrite==NULL)
    return 0;
  return rtprewrite->func_rtcp_protect(rtprewrite->val, data, len);
}

static int rtprewrite_unprotect_rtcp(rtprewrite_t *rtprewrite, unsigned char *data, int *len) {
  if (rtprewrite==NULL)
    return 0;
  return rtprewrite->func_rtcp_unprotect(rtprewrite->val, data, len);
}

static int _module_rtprewrite_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m, rtprewrite_t *rtprewrite){
  int slen;
  int err;
  /* enlarge the buffer for rtprewrite to write its data */
  slen=msgdsize(m);
  msgpullup(m,slen+rtprewrite->padding_size);
  err=rtprewrite_protect(rtprewrite,m->b_rptr,&slen);
  if (err==0){
    m->b_wptr=m->b_rptr+slen;
    return slen;
  }
  ortp_error("rtprewrite_protect() failed");
  return -1;
}

static int module_rtprewrite_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  rtprewrite_t *rtprewrite=(rtprewrite_t*)module->data;
  return _module_rtprewrite_sendto(session, module, m, rtprewrite);
}

static int _module_rtprewrite_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m, rtprewrite_t *rtprewrite){
  int slen;
  int ret=0;

  /* keep NON-RTP data unencrypted */
  rtp_header_t *rtp = (rtp_header_t*)m->b_rptr;
  if (msgdsize(m)>=RTP_FIXED_HEADER_SIZE)
  {
    if (rtp->version!=2)
    {
      return msgdsize(m); /* not modified */
    }
  }

  slen=msgdsize(m);
  if (session->rtcp_mux && rtp->paytype>=64 && rtp->paytype<=95) {
    ret = rtprewrite_unprotect_rtcp(rtprewrite,m->b_rptr,&slen);
  } else {
    ret = rtprewrite_unprotect(rtprewrite,m->b_rptr,&slen);
  }

  if (ret ==0)
  {
    m->b_wptr=m->b_rptr+slen;
    return slen;
  }
  ortp_error("rtprewrite_unprotect() failed");
  return -1;
}

static int module_rtprewrite_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  rtprewrite_t* rtprewrite=(rtprewrite_t*)module->data;
  return _module_rtprewrite_recvfrom(session, module, m, rtprewrite);
}

static int _module_srtcp_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m, rtprewrite_t* rtprewrite){
  int slen;
  int err;
  /* enlarge the buffer for rtprewrite to write its data */
  slen=msgdsize(m);
  msgpullup(m,slen+rtprewrite->padding_size);
  err=rtprewrite_protect_rtcp(rtprewrite,m->b_rptr,&slen);
  if (err==0){
    m->b_wptr=m->b_rptr+slen;
    return slen;
  }
  ortp_error("rtprewrite_protect_rtcp() failed");
  return -1;
}

static int module_srtcp_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  rtprewrite_t* rtprewrite=(rtprewrite_t*)module->data;
  return _module_srtcp_sendto(session, module, m, rtprewrite);
}

static int _module_srtcp_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m, rtprewrite_t* rtprewrite){
  int slen;
  slen=msgdsize(m);
  if (rtprewrite_unprotect_rtcp(rtprewrite,m->b_rptr,&slen)==0) {
    /* complete with difference */
    m->b_wptr=m->b_rptr+slen;
    return slen;
  }
  ortp_error("rtprewrite_unprotect_rtcp() failed");
  return -1;
}

static int module_srtcp_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  rtprewrite_t* rtprewrite=(rtprewrite_t*)module->data;
  return _module_srtcp_recvfrom(session, module, m, rtprewrite);
}

static void
  module_rtprewrite_destroy(struct _RtpModule *module)
{
  rtprewrite_t* rtprewrite=(rtprewrite_t*)module->data;
  if (rtprewrite!=NULL)
    ortp_free(rtprewrite);
  module->data=NULL;
}

static int rtprewrite_set_option(struct _RtpModule *module, int opt, void *value)
{
  rtprewrite_t* rtprewrite=(rtprewrite_t*)module->data;
  if (rtprewrite==NULL) {
    rtprewrite=(rtprewrite_t*)ortp_malloc0(sizeof(rtprewrite_t));
    module->data=rtprewrite;
  }

  if (opt==0) {
    rtprewrite->val = *(int*)value;
    return 0;
  }
  if (opt==1) {
    rtprewrite->padding_size = *(int*)value;
    return 0;
  }
  if (opt==2) {
    rtprewrite->func_rtp_protect = (on_func_rtprewrite_protect)value;
    return 0;
  }
  if (opt==3) {
    rtprewrite->func_rtcp_protect = (on_func_rtcprewrite_protect)value;
    return 0;
  }
  if (opt==4) {
    rtprewrite->func_rtp_unprotect = (on_func_rtprewrite_unprotect)value;
    return 0;
  }
  if (opt==5) {
    rtprewrite->func_rtcp_unprotect = (on_func_rtcprewrite_unprotect)value;
    return 0;
  }
  return -1;
}

static int module_rtprewrite_global_init(void)
{
  return 0;
}

static int module_rtprewrite_global_shutdown(void)
{
  return 0;
}

int rtprewrite_module_setup(void);

int rtprewrite_module_setup(void) {

  RtpModule *tr = ortp_new0(RtpModule,1);
  snprintf(tr->name, sizeof(tr->name), "RTPREWRITE");

  tr->t_global_init = module_rtprewrite_global_init;
  tr->t_global_shutdown = module_rtprewrite_global_shutdown;
  tr->t_set_option = rtprewrite_set_option;

  tr->t_rtp_sendto=module_rtprewrite_sendto;
  tr->t_rtp_recvfrom=module_rtprewrite_recvfrom;
  tr->t_rtcp_sendto=module_srtcp_sendto;
  tr->t_rtcp_recvfrom=module_srtcp_recvfrom;
  tr->t_destroy=module_rtprewrite_destroy;

  ortp_module_add(tr);
  return 0;
}

static int rtp_sendto(struct _RtpSession *session, mblk_t *m, int flags, const struct sockaddr *to, socklen_t tolen){
  int slen;
  slen=(int)(m->b_wptr-m->b_rptr);
  return (int)sendto(session->rtp.socketv4[0],m->b_rptr,slen,flags,to,tolen);
}

static int rtp_recvfrom(struct _RtpSession *session, mblk_t *m, int flags, struct sockaddr *from, socklen_t *fromlen){
  int err;
  err=(int)recvfrom(session->rtp.socketv4[0],m->b_wptr,m->b_datap->db_lim-m->b_datap->db_base,flags,from,fromlen);
  return err;
}


static int rtcp_sendto(struct _RtpSession *session, mblk_t *m, int flags, const struct sockaddr *to, socklen_t tolen){
  int slen;
  slen=(int)(m->b_wptr-m->b_rptr);
  return (int)sendto(session->rtcp.socketv4[0],m->b_rptr,slen,flags,to,tolen);
}

static int rtcp_recvfrom(struct _RtpSession *session, mblk_t *m, int flags, struct sockaddr *from, socklen_t *fromlen){
  int err;
  err=(int)recvfrom(session->rtcp.socketv4[0],m->b_wptr,m->b_datap->db_lim-m->b_datap->db_base,flags,from,fromlen);
  return err;
}

static ortp_socket_t rtp_getsocket(const struct _RtpSession *session)
{
  return session->rtp.socketv4[0];
}

static ortp_socket_t rtcp_getsocket(const struct _RtpSession *session)
{
  return session->rtcp.socketv4[0];
}

static int rtp_global_init(void)
{
  return 0;
}

static int rtp_global_shutdown(void)
{
  return 0;
}

int rtp_transport_setup(void);

int rtp_transport_setup(void) {

  RtpTransport *tr = ortp_new0(RtpTransport,1);
  snprintf(tr->name, sizeof(tr->name), "RTP");
  tr->t_global_init = rtp_global_init;
  tr->t_global_shutdown = rtp_global_shutdown;

  tr->t_rtp_getsocket=rtp_getsocket;
  tr->t_rtcp_getsocket=rtcp_getsocket;
  tr->t_rtp_sendto=rtp_sendto;
  tr->t_rtp_recvfrom=rtp_recvfrom;
  tr->t_rtcp_sendto=rtcp_sendto;
  tr->t_rtcp_recvfrom=rtcp_recvfrom;

  ortp_transport_add(tr);
  return 0;
}
