/*
  The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#if defined(_MSC_VER)  && (defined(WIN32) || defined(_WIN32_WCE))
#include "ortp-config-win32.h"
#elif HAVE_CONFIG_H
#include "ortp-config.h"
#endif
#include "ortp/ortp.h"

#ifdef HAVE_SRTP

#undef PACKAGE_NAME 
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME 
#undef PACKAGE_VERSION

#include "ortp/srtp.h"
#include "rtpsession_priv.h"

#define SRTP_PAD_BYTES SRTP_MAX_TRAILER_LEN + 4

static int _module_srtp_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m, srtp_t srtp){
  int slen;
  srtp_err_status_t err;
  /* enlarge the buffer for srtp to write its data */
  slen=msgdsize(m);
  msgpullup(m,slen+SRTP_PAD_BYTES);
  err=srtp_protect(srtp,m->b_rptr,&slen);
  if (err==srtp_err_status_ok){
    m->b_wptr=m->b_rptr+slen;
    return slen;
  }
  ortp_error("srtp_protect() failed");
  return -1;
}

static int module_srtp_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  srtp_t srtp=(srtp_t)module->data;
  return _module_srtp_sendto(session, module, m, srtp);
}

static int _module_srtp_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m, srtp_t srtp){
  int slen;
  srtp_err_status_t ret=srtp_err_status_ok;
  rtp_header_t *rtp = (rtp_header_t*)m->b_rptr;

  /* keep NON-RTP data unencrypted */
  if ((m->b_rptr[0] < 128) || (m->b_rptr[0] > 191)) {
    /* rtp/rtcp channel packet not detected */
    return msgdsize(m); /* not modified */
  }

  slen=msgdsize(m);
  if (session->rtcp_mux && rtp->paytype>=64 && rtp->paytype<=95) {
    ret = srtp_unprotect_rtcp(srtp,m->b_rptr,&slen);
  } else {
    ret = srtp_unprotect(srtp,m->b_rptr,&slen);
  }

  if (ret ==srtp_err_status_ok)
  {
    m->b_wptr=m->b_rptr+slen;
    return slen;
  }
  ortp_error("srtp_unprotect() failed");
  return -1;
}

static int module_srtp_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  srtp_t srtp=(srtp_t)module->data;
  return _module_srtp_recvfrom(session, module, m, srtp);
}

static int _module_srtcp_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m, srtp_t srtp){
  int slen;
  srtp_err_status_t err;
  /* enlarge the buffer for srtp to write its data */
  slen=msgdsize(m);
  msgpullup(m,slen+SRTP_PAD_BYTES);
  err=srtp_protect_rtcp(srtp,m->b_rptr,&slen);
  if (err==srtp_err_status_ok){
    m->b_wptr=m->b_rptr+slen;
    return slen;
  }
  ortp_error("srtp_protect_rtcp() failed");
  return -1;
}

static int module_srtcp_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  srtp_t srtp=(srtp_t)module->data;
  return _module_srtcp_sendto(session, module, m, srtp);
}

static int _module_srtcp_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m, srtp_t srtp){
  int slen;
  /* keep NON-RTP data unencrypted */
  if ((m->b_rptr[0] < 128) || (m->b_rptr[0] > 191)) {
	  /* rtp/rtcp channel packet not detected */
	  return msgdsize(m); /* not modified */
  }
  slen=msgdsize(m);
  if (srtp_unprotect_rtcp(srtp,m->b_rptr,&slen)==srtp_err_status_ok) {
    /* complete with difference */
    m->b_wptr=m->b_rptr+slen;
    return slen;
  }
  ortp_error("srtp_unprotect_rtcp() failed");
  return -1;
}

static int module_srtcp_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  srtp_t srtp=(srtp_t)module->data;
  return _module_srtcp_recvfrom(session, module, m, srtp);
}

static void 
  module_srtp_destroy(struct _RtpModule *module)
{
  srtp_t srtp=(srtp_t)module->data;
  if (srtp!=NULL)
    srtp_dealloc(srtp);
  module->data=NULL;
}

static int module_srtp_global_init(void)
{
  return srtp_init();
}

static int module_srtp_global_shutdown(void)
{
  return srtp_shutdown();
}

srtp_err_status_t ortp_srtp_create(srtp_t *session, const srtp_policy_t *policy)
{
  return srtp_create(session, policy);
}

srtp_err_status_t ortp_srtp_dealloc(srtp_t session)
{
  return srtp_dealloc(session);
}

srtp_err_status_t ortp_srtp_add_stream(srtp_t session, const srtp_policy_t *policy)
{
  return srtp_add_stream(session, policy);
}

#else

int ortp_srtp_create(void *session, void *policy)
{
  return -1;
}

int ortp_srtp_dealloc(void *session)
{
  return -1;
}

int ortp_srtp_add_stream(void *session, const void *policy)
{
  return -1;
}

#endif

int srtp_module_setup(void);

int srtp_module_setup(void) {

#ifdef HAVE_SRTP
  RtpModule *tr = ortp_new0(RtpModule,1);
  snprintf(tr->name, sizeof(tr->name), "SRTP");

  tr->t_global_init = module_srtp_global_init;
  tr->t_global_shutdown = module_srtp_global_shutdown;
  tr->t_set_option = NULL;

  tr->t_rtp_sendto=module_srtp_sendto;
  tr->t_rtp_recvfrom=module_srtp_recvfrom;
  tr->t_rtcp_sendto=module_srtcp_sendto;
  tr->t_rtcp_recvfrom=module_srtcp_recvfrom;
  tr->t_destroy=module_srtp_destroy;

  ortp_module_add(tr);
#endif
  return 0;
}


#if defined(HAVE_OPENSSL_SSL_H) && defined(HAVE_SRTP)

#include <openssl/ssl.h>
#include <openssl/x509v3.h>

static X509 *ortp_dtls_certificate=NULL;
static EVP_PKEY *ortp_dtls_key=NULL;

#define ORTP_SRTP_HANDSHAKE 0
#define ORTP_SRTP_FINISHED 1
#define ORTP_SRTP_FAILED 2

struct ortp_dtls_srtp {
  int verify_enable;
  int verify_failed;
  int ssl_status;
  struct timeval connect_timeout;
  int rfc4145_setup;
  uint32_t rekey_interval;
  SSL_CTX *ssl_ctx; /*!< SSL context */
  SSL *ssl;         /*!< SSL session */
  BIO *read_bio;    /*!< Memory buffer for reading */
  BIO *write_bio;   /*!< Memory buffer for writing */

  char remote_fingerprint_sha1[182]; /* 3*EVP_MAX_MD_SIZE */
  char remote_fingerprint_sha256[182]; /* 3*EVP_MAX_MD_SIZE */

  srtp_t srtp_out; /* outbound streams */
  srtp_t srtp_in; /* inbound streams */
};

typedef struct ortp_dtls_srtp ortp_dtls_srtp_t;

static void dtls_info_callback(const SSL *ssl, int where, int ret)
{
  ortp_dtls_srtp_t *dtlsctx = (ortp_dtls_srtp_t*)SSL_get_ex_data(ssl, 0);
  const char *str;
  int w;
  w=where& ~SSL_ST_MASK;
  if (w & SSL_ST_CONNECT)
    str="SSL_connect";
  else if (w & SSL_ST_ACCEPT)
    str="SSL_accept";
  else
    str="undefined";
  if (where & SSL_CB_LOOP)
  {
    ortp_message("module_dtls(%x): %s:%s\n",dtlsctx,str,SSL_state_string_long(ssl));
  }
  else if (where & SSL_CB_ALERT)
  {
    str=(where & SSL_CB_READ)?"read":"write";
    ortp_message("module_dtls(%x): SSL3 alert %s:%s:%s\n",
      dtlsctx,str,
      SSL_alert_type_string_long(ret),
      SSL_alert_desc_string_long(ret));
  }
  else if (where & SSL_CB_EXIT)
  {
    if (ret == 0)
      ortp_error("module_dtls(%x): %s:failed in %s\n", dtlsctx,
      str,SSL_state_string_long(ssl));
    else if (ret < 0)
    {
      ortp_error("module_dtls(%x): %s:error in %s\n", dtlsctx,
        str,SSL_state_string_long(ssl));
    }
  }
  /* We only care about alerts */
  if (!(where & SSL_CB_ALERT)) {
    return;
  }
  dtlsctx->verify_failed = 1;
}

#if 0
static int cb_verify_peer(int preverify_ok, X509_STORE_CTX *ctx) {
  SSL *ssl = NULL;
  ortp_dtls_srtp_t *dtlsctx;
  X509 *certificate;

  ssl = X509_STORE_CTX_get_app_data(ctx);
  dtlsctx = (ortp_dtls_srtp_t *) SSL_get_app_data(ssl);

  if (!(ssl && dtlsctx)) {
    return 0;
  }

  if (!(certificate = SSL_get_peer_certificate(dtlsctx->ssl))) {
    ortp_error("module_dtls: %p SSL_get_peer_certificate: no certificate", NULL);
    return 1;
  }

  if (dtlsctx->remote_fingerprint_sha256[0]) {
    unsigned char fingerprint[EVP_MAX_MD_SIZE];
    unsigned int size;
    int ret = X509_digest(certificate, EVP_sha256(), fingerprint, &size);
    if (ret==1 && size>0) {
      char remote_fingerprint_sha256[182];
      char *tmp = remote_fingerprint_sha256;
      unsigned int i;
      for (i = 0; i < size; i++) {
        sprintf(tmp, "%.2X:", fingerprint[i]);
        tmp += 3;
      }

      if (memcmp(remote_fingerprint_sha256, dtlsctx->remote_fingerprint_sha256, size)) {
        X509_free(certificate);
        ortp_error("module_dtls: %p Fingerprint provided by remote party does not match that of peer certificate", NULL);
        return -1;
      }
    }
  } else if (dtlsctx->remote_fingerprint_sha1[0]) {
    unsigned char fingerprint[EVP_MAX_MD_SIZE];
    unsigned int size;
    int ret = X509_digest(certificate, EVP_sha1(), fingerprint, &size);
    if (ret==1 && size>0) {
      char remote_fingerprint_sha1[182];
      char *tmp = remote_fingerprint_sha1;
      unsigned int i;
      for (i = 0; i < size; i++) {
        sprintf(tmp, "%.2X:", fingerprint[i]);
        tmp += 3;
      }

      if (memcmp(remote_fingerprint_sha1, dtlsctx->remote_fingerprint_sha1, size)) {
        X509_free(certificate);
        ortp_error("module_dtls: %p Fingerprint provided by remote party does not match that of peer certificate", NULL);
        return -1;
      }
    }
  }

  return 1;
}
#endif

static int _module_dtls_srtp_setup(struct _RtpSession *session, struct _RtpModule *module) {
  ortp_dtls_srtp_t *dtlsctx=(ortp_dtls_srtp_t*)module->data;
  int i;
  const SSL_METHOD *meth;
  EC_KEY* ecdh;

  SSL_library_init ();
  SSL_load_error_strings ();

#if !(OPENSSL_VERSION_NUMBER < 0x10002000L)
  meth = DTLS_method();
#else
  meth = DTLSv1_method();
#endif
  dtlsctx->ssl_ctx = SSL_CTX_new(meth);

  SSL_CTX_set_options(dtlsctx->ssl_ctx, SSL_OP_NO_QUERY_MTU);

#if 0
  SSL_CTX_set_verify(dtlsctx->ssl_ctx, dtlsctx->verify_enable ? SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT : SSL_VERIFY_NONE, cb_verify_peer);
#endif
  SSL_CTX_set_verify_depth(dtlsctx->ssl_ctx, 4);
  SSL_CTX_set_cipher_list(dtlsctx->ssl_ctx, "ALL:!ADH:!LOW:!EXP:!MD5:@STRENGTH");
  SSL_CTX_set_tlsext_use_srtp(dtlsctx->ssl_ctx, "SRTP_AES128_CM_SHA1_80:SRTP_AES128_CM_SHA1_32");

  i = SSL_CTX_use_certificate(dtlsctx->ssl_ctx, ortp_dtls_certificate);
  if (i != 1) {
    return -1;
  }

  i = SSL_CTX_use_PrivateKey(dtlsctx->ssl_ctx, ortp_dtls_key);
  if (i != 1) {
    return -1;
  }

  dtlsctx->ssl = SSL_new(dtlsctx->ssl_ctx);
  SSL_set_ex_data(dtlsctx->ssl, 0, dtlsctx);
  SSL_set_info_callback(dtlsctx->ssl, dtls_info_callback);
  dtlsctx->read_bio = BIO_new(BIO_s_mem());
  BIO_set_mem_eof_return(dtlsctx->read_bio, -1);
  dtlsctx->write_bio = BIO_new(BIO_s_mem());
  BIO_set_mem_eof_return(dtlsctx->write_bio, -1);

  SSL_set_mtu(dtlsctx->ssl, 1200);
  BIO_ctrl(dtlsctx->write_bio, BIO_CTRL_DGRAM_SET_MTU, 1200, NULL);
  SSL_set_bio(dtlsctx->ssl, dtlsctx->read_bio, dtlsctx->write_bio);

  ecdh = EC_KEY_new_by_curve_name(NID_X9_62_prime256v1);
  if (ecdh == NULL)
    return -1;
  SSL_set_options(dtlsctx->ssl, SSL_OP_SINGLE_ECDH_USE);
  SSL_set_tmp_ecdh(dtlsctx->ssl, ecdh);
  EC_KEY_free(ecdh);

  if (dtlsctx->rfc4145_setup == RTP_SESSION_SETUP_PASSIVE) {
    SSL_set_accept_state(dtlsctx->ssl);
  } else {
    SSL_set_connect_state(dtlsctx->ssl);
  }
  return 0;
}

static void dtls_srtp_bio_ctrl_pending(struct _RtpSession *session, struct _RtpModule *module)
{
  ortp_dtls_srtp_t *dtlsctx=(ortp_dtls_srtp_t*)module->data;
  int pending = (int)BIO_ctrl_pending(dtlsctx->write_bio);
  if (pending > 0) {
    mblk_t *ob = allocb(pending, 0);
    int out;
    out = BIO_read(dtlsctx->write_bio, ob->b_rptr, pending);
    ob->b_wptr+=out;
    if (ob!=NULL) {
      const uint8_t* u = ob->b_rptr;
      if (out > 17 && u[0] == 22 && u[13] == 1) {
        ortp_message("module_dtls: sending DtlsClientHelloPacket packet (out=%i)", out);
      } else {
        ortp_message("module_dtls: NOT sending DtlsClientHelloPacket packet (out=%i)", out);
      }
      if (session->bundle!=NULL) rtp_bundle_lock(session->bundle);
      rtp_session_data_send(session, ob, 2);
      if (session->bundle!=NULL) rtp_bundle_unlock(session->bundle);
    }
    return;
  }
  return;
}

static int module_dtls_srtp_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  ortp_dtls_srtp_t *dtlsctx=(ortp_dtls_srtp_t*)module->data;

  if (dtlsctx->rfc4145_setup == RTP_SESSION_SETUP_ACTPASS) /* no meaning here */
    dtlsctx->rfc4145_setup = RTP_SESSION_SETUP_ACTIVE;

  if (dtlsctx->rfc4145_setup == RTP_SESSION_SETUP_HOLDCONN)
    return -1;

  if (dtlsctx->ssl==NULL)
  {
    _module_dtls_srtp_setup(session, module);
  }

  if (dtlsctx->rfc4145_setup == RTP_SESSION_SETUP_ACTIVE) {
    if (session->rtp_icechecklist!=NULL && session->rtp_icechecklist->cand_pairs!=NULL) {
      if (session->flags & RTP_SESSION_DTLS_START) {
        dtlsctx->connect_timeout.tv_sec=0;
        dtlsctx->connect_timeout.tv_usec=0;
      } else {
        uint64_t interval;
        struct timeval now;
        gettimeofday(&now, NULL);
        interval = (now.tv_sec-dtlsctx->connect_timeout.tv_sec)*1000;
        interval+=(now.tv_usec-dtlsctx->connect_timeout.tv_usec)/1000;
        if (interval>5000) { /* 5 seconds? */
          ortp_warning("module_dtls: %p starting DTLS anyway", session);
          rtp_session_set_flag (session, RTP_SESSION_DTLS_START);
          dtlsctx->connect_timeout.tv_sec=0;
          dtlsctx->connect_timeout.tv_usec=0;
        }
      }
    } else {
      rtp_session_set_flag (session, RTP_SESSION_DTLS_START);
      dtlsctx->connect_timeout.tv_sec=0;
      dtlsctx->connect_timeout.tv_usec=0;
    }
    if (dtlsctx->connect_timeout.tv_sec==0 && dtlsctx->ssl_status==ORTP_SRTP_HANDSHAKE) {
      int err;
      if (dtlsctx->ssl==NULL)
      {
        _module_dtls_srtp_setup(session, module);
      }
      if (dtlsctx->ssl==NULL)
        return -1;

      err = SSL_do_handshake(dtlsctx->ssl);
      if (err != 1){
        int ssl_error = SSL_get_error(dtlsctx->ssl, err);
        if (ssl_error!=SSL_ERROR_WANT_READ && ssl_error!=SSL_ERROR_WANT_WRITE && ssl_error!=SSL_ERROR_NONE) {
          dtlsctx->ssl_status = ORTP_SRTP_FAILED;
        }
      }
    }
  }

  dtls_srtp_bio_ctrl_pending(session, module);

  if (dtlsctx->srtp_out!=NULL)
    return _module_srtp_sendto(session, module, m, dtlsctx->srtp_out);
  return -1;
}

#define MASTER_KEY_LEN   30

static int module_dtls_srtp_extract_keying_material(struct _RtpSession *session, struct _RtpModule *module){
  ortp_dtls_srtp_t *dtlsctx=(ortp_dtls_srtp_t*)module->data;

  unsigned char material[MASTER_KEY_LEN * 2];
  SRTP_PROTECTION_PROFILE *srtp_profile;

  char local_key[64];
  char remote_key[64];

  int i;
  srtp_t srtp;
  srtp_policy_t policy;

  memset(&policy, 0, sizeof(srtp_policy_t));

#if 0
  /* TODO: currently, for outgoing calls, SSL_get_peer_certificate returns "no certificate" */
  if (dtlsctx->verify_enable) {
    X509 *certificate;
    if (!(certificate = SSL_get_peer_certificate(dtlsctx->ssl))) {
      ortp_error("module_dtls: %p SSL_get_peer_certificate: no certificate", session);
      return -1;
    }

    if (dtlsctx->remote_fingerprint_sha256[0]) {
      unsigned char fingerprint[EVP_MAX_MD_SIZE];
      unsigned int size;
      int ret = X509_digest(certificate, EVP_sha256(), fingerprint, &size);
      if (ret==1 && size>0) {
        char remote_fingerprint_sha256[182];
        char *tmp = remote_fingerprint_sha256;
        unsigned int i;
        for (i = 0; i < size; i++) {
          sprintf(tmp, "%.2X:", fingerprint[i]);
          tmp += 3;
        }

        if (memcmp(remote_fingerprint_sha256, dtlsctx->remote_fingerprint_sha256, size)) {
          X509_free(certificate);
          ortp_error("module_dtls: %p Fingerprint provided by remote party does not match that of peer certificate", NULL);
          return -1;
        }
      }
    } else if (dtlsctx->remote_fingerprint_sha1[0]) {
      unsigned char fingerprint[EVP_MAX_MD_SIZE];
      unsigned int size;
      int ret = X509_digest(certificate, EVP_sha1(), fingerprint, &size);
      if (ret==1 && size>0) {
        char remote_fingerprint_sha1[182];
        char *tmp = remote_fingerprint_sha1;
        unsigned int i;
        for (i = 0; i < size; i++) {
          sprintf(tmp, "%.2X:", fingerprint[i]);
          tmp += 3;
        }

        if (memcmp(remote_fingerprint_sha1, dtlsctx->remote_fingerprint_sha1, size)) {
          X509_free(certificate);
          ortp_error("module_dtls: %p Fingerprint provided by remote party does not match that of peer certificate", NULL);
          return -1;
        }
      }
    }
    X509_free(certificate);
  }
#endif

#if 0
  if (SSL_get_verify_result(dtlsctx->ssl) != X509_V_OK) {
    ortp_error("module_dtls: %p cannot verify remote fingerprint", session);
    return -1;
  }
#endif

  if (!SSL_export_keying_material(dtlsctx->ssl, material, MASTER_KEY_LEN * 2, "EXTRACTOR-dtls_srtp", 19, NULL, 0, 0)) {
    ortp_error("module_dtls: %p unable to expirt keying material", session);
    return -1;
  }

  if (dtlsctx->rfc4145_setup == RTP_SESSION_SETUP_ACTIVE) {
    memcpy(local_key, material, 16);
    memcpy(local_key+16, material+16+16, 14);
    memcpy(remote_key, material+16, 16);
    memcpy(remote_key+16, material+16+16+14, 14);
  } else {
    memcpy(remote_key, material, 16);
    memcpy(remote_key+16, material+16+16, 14);
    memcpy(local_key, material+16, 16);
    memcpy(local_key+16, material+16+16+14, 14);
  }
  local_key[30]='\0';
  remote_key[30]='\0';

  srtp_profile = SSL_get_selected_srtp_profile(dtlsctx->ssl);
  if (srtp_profile==NULL || srtp_profile->name==NULL) {
    ortp_error ("module_dtls: %p SSL_get_selected_srtp_profile failed\n", session);
    return -1;
  }

  ortp_srtp_create(&srtp, NULL);

  if (strcmp(srtp_profile->name, "SRTP_AES128_CM_SHA1_80")==0) {
    srtp_crypto_policy_set_aes_cm_128_hmac_sha1_80 (&policy.rtp);
    srtp_crypto_policy_set_aes_cm_128_hmac_sha1_80 (&policy.rtcp);
  } else {
    srtp_crypto_policy_set_aes_cm_128_hmac_sha1_32 (&policy.rtp);
    srtp_crypto_policy_set_aes_cm_128_hmac_sha1_80 (&policy.rtcp);
  }

  policy.next = NULL;
  policy.key = (uint8_t *) remote_key;
  policy.ssrc.type = ssrc_any_inbound;
  policy.ssrc.value = 0;
  policy.window_size = 0;
#if 0
  if (remote_wsh != NULL)
    policy.window_size = atoi (remote_wsh + 4);
#endif
  policy.allow_repeat_tx = 0;
  policy.ekt = NULL;

  i = ortp_srtp_add_stream (srtp, &policy);
  if (i != 0) {
    ortp_error ("module_dtls: %p srtp_add_stream failed\n", session);
    return -1;
  }
  dtlsctx->srtp_in = srtp;

  ortp_srtp_create(&srtp, NULL);

  policy.key = (uint8_t *) local_key;
  policy.ssrc.type = ssrc_any_outbound;
  policy.ssrc.value = session->snd.ssrc;
  policy.window_size = 0;
#if 0
  if (local_wsh != NULL)
    policy.window_size = atoi (local_wsh + 4);
#endif
  policy.allow_repeat_tx = 0;
  policy.ekt = NULL;

  i = ortp_srtp_add_stream (srtp, &policy);
  if (i != 0) {
    ortp_error ("module_dtls: %p srtp_add_stream failed\n", session);
    return -1;
  }

  dtlsctx->srtp_out = srtp;
  ortp_message("module_dtls: dtls connection established");
  return 0;
}

static int module_dtls_srtp_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  ortp_dtls_srtp_t *dtlsctx=(ortp_dtls_srtp_t*)module->data;

  if (dtlsctx->rfc4145_setup == RTP_SESSION_SETUP_ACTPASS) /* no meaning here */
    dtlsctx->rfc4145_setup = RTP_SESSION_SETUP_ACTIVE;

  if (dtlsctx->rfc4145_setup == RTP_SESSION_SETUP_HOLDCONN)
    return -1;

  if (dtlsctx->ssl==NULL)
  {
    _module_dtls_srtp_setup(session, module);
  }

  if (dtlsctx->rfc4145_setup == RTP_SESSION_SETUP_ACTIVE) {
    if (dtlsctx->connect_timeout.tv_sec>0) {
      if (session->rtp_icechecklist!=NULL && session->rtp_icechecklist->cand_pairs!=NULL) {
        if (session->flags & RTP_SESSION_DTLS_START) {
          dtlsctx->connect_timeout.tv_sec=0;
          dtlsctx->connect_timeout.tv_usec=0;
        } else {
          uint64_t interval;
          struct timeval now;
          gettimeofday(&now, NULL);
          interval = (now.tv_sec-dtlsctx->connect_timeout.tv_sec)*1000;
          interval+=(now.tv_usec-dtlsctx->connect_timeout.tv_usec)/1000;
          if (interval>5000) { /* 5 seconds? */
            ortp_warning("module_dtls: %p starting DTLS anyway", session);
            rtp_session_set_flag (session, RTP_SESSION_DTLS_START);
            dtlsctx->connect_timeout.tv_sec=0;
            dtlsctx->connect_timeout.tv_usec=0;
          }
        }
      } else {
        rtp_session_set_flag (session, RTP_SESSION_DTLS_START);
        dtlsctx->connect_timeout.tv_sec=0;
        dtlsctx->connect_timeout.tv_usec=0;
      }
    }
    if (dtlsctx->connect_timeout.tv_sec==0 && dtlsctx->ssl_status==ORTP_SRTP_HANDSHAKE) {
      int err;
      if (dtlsctx->ssl==NULL)
      {
        _module_dtls_srtp_setup(session, module);
      }
      if (dtlsctx->ssl==NULL)
        return -1;

      err = SSL_do_handshake(dtlsctx->ssl);
      if (err != 1){
        int ssl_error = SSL_get_error(dtlsctx->ssl, err);
        if (ssl_error!=SSL_ERROR_WANT_READ && ssl_error!=SSL_ERROR_WANT_WRITE && ssl_error!=SSL_ERROR_NONE) {
          dtlsctx->ssl_status = ORTP_SRTP_FAILED;
        }
      }
    }
  }
  if (dtlsctx->ssl==NULL)
    return -1;

  dtls_srtp_bio_ctrl_pending(session, module);

  if ((m->b_rptr[0] >= 20) && (m->b_rptr[0] <= 63)) {
    /* DTLS packet detected */
    /* we receive an incoming ssl connection */
    if (dtlsctx->rfc4145_setup == RTP_SESSION_SETUP_ACTPASS) {
      dtlsctx->rfc4145_setup = RTP_SESSION_SETUP_PASSIVE;
      SSL_set_accept_state(dtlsctx->ssl);
    }

    dtls_srtp_bio_ctrl_pending(session, module);
    BIO_write(dtlsctx->read_bio, m->b_rptr, msgdsize(m));
    SSL_read(dtlsctx->ssl, m->b_rptr, msgdsize(m));
    dtls_srtp_bio_ctrl_pending(session, module);

    if (dtlsctx->verify_failed) {
      ortp_error("module_dtls: %p ssl verification failed", session);
    }

    if (SSL_is_init_finished(dtlsctx->ssl)) {
      /* dtls connection established */
      module_dtls_srtp_extract_keying_material(session, module);
      dtlsctx->ssl_status = ORTP_SRTP_FINISHED;
    }

    /* fake error (to discard packet) */
    return msgdsize(m);
  }

  if (dtlsctx->srtp_in!=NULL)
    return _module_srtp_recvfrom(session, module, m, dtlsctx->srtp_in);
  return -1;
}

static int module_dtls_srtcp_sendto(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  ortp_dtls_srtp_t *dtlsctx=(ortp_dtls_srtp_t*)module->data;

  if (dtlsctx->srtp_out!=NULL)
    return _module_srtcp_sendto(session, module, m, dtlsctx->srtp_out);
  return -1;
}

static int module_dtls_srtcp_recvfrom(struct _RtpSession *session, struct _RtpModule *module, mblk_t *m){
  ortp_dtls_srtp_t *dtlsctx=(ortp_dtls_srtp_t*)module->data;
  if ((m->b_rptr[0] >= 20) && (m->b_rptr[0] <= 64)) {
    /* ssl packet detected */
    return -1;
  }
  if (dtlsctx->srtp_in!=NULL)
    return _module_srtcp_recvfrom(session, module, m, dtlsctx->srtp_in);
  return -1;
}

static void 
  module_dtls_srtp_destroy(struct _RtpModule *module)
{
  module->data=NULL;
}

static int cert_generate(X509 *certificate, EVP_PKEY *privkey)
{
  X509_NAME *name;
  int ret;
  char *cn = "certificate@antisip.com";

  name = X509_NAME_new();
  if (!name) {
    return -1;
  }
  X509_set_version(certificate, 2);

  ASN1_INTEGER_set(X509_get_serialNumber(certificate), ortp_random());

  ret = X509_NAME_add_entry_by_txt(name, "CN", MBSTRING_ASC, (unsigned char *)cn, (int)strlen(cn), -1, 0);
  if (!ret) {
    X509_NAME_free(name);
    return -1;
  }

  if (!X509_set_issuer_name(certificate, name) || !X509_set_subject_name(certificate, name)) {
    X509_NAME_free(name);
    return -1;
  }
  X509_NAME_free(name);

  X509_gmtime_adj(X509_get_notBefore(certificate), 0);
  X509_gmtime_adj(X509_get_notAfter(certificate), 60*60*24*365); /* one year */

  if (!X509_set_pubkey(certificate, privkey)) {
    return -1;
  }

  {
    X509_EXTENSION *ex;
    ex = X509V3_EXT_conf_nid(NULL, NULL, NID_basic_constraints, "CA:FALSE");
    if (1 != X509_add_ext(certificate, ex, -1)) {
      return -1;
    }

    X509_EXTENSION_free(ex);
  }

  return 0;
}

static RSA *
__RSA_generate_key (int bits, unsigned long e_value, void (*callback) (int, int, void *), void *cb_arg)
{
  int i;
  RSA *rsa = RSA_new ();
  BIGNUM *e = BN_new ();

  if (!rsa || !e)
    goto err;

  i = BN_set_word(e, e_value);
  if (i!=1)
    goto err;

  if (RSA_generate_key_ex (rsa, bits, e, NULL)) {
    BN_free (e);
    return rsa;
  }
 err:
  if (e)
    BN_free (e);
  if (rsa)
    RSA_free (rsa);
  return 0;
}

static int ortp_dtls_srtp_generate_selfsigned_cert()
{
  RSA *rsa;

  if (ortp_dtls_certificate!=NULL && ortp_dtls_key!=NULL)
    return 0;

  rsa = __RSA_generate_key(1024, RSA_F4, NULL, NULL);
  if (!rsa)
    return -1;

  ortp_dtls_key = EVP_PKEY_new();
  if (!ortp_dtls_key) {
    RSA_free(rsa);
    return -1;
  }
  if (!EVP_PKEY_set1_RSA(ortp_dtls_key, rsa)) {
    RSA_free(rsa);
    EVP_PKEY_free(ortp_dtls_key);
    ortp_dtls_key=NULL;
    return -1;
  }

  ortp_dtls_certificate = X509_new();
  if (!ortp_dtls_certificate) {
    RSA_free(rsa);
    EVP_PKEY_free(ortp_dtls_key);
    ortp_dtls_key=NULL;
    return -1;
  }

  if (cert_generate(ortp_dtls_certificate, ortp_dtls_key)) {
    RSA_free(rsa);
    EVP_PKEY_free(ortp_dtls_key);
    X509_free(ortp_dtls_certificate);
    ortp_dtls_certificate=NULL;
    ortp_dtls_key=NULL;
    return -1;
  }

  /* Sign the certificate */
  if (!X509_sign(ortp_dtls_certificate, ortp_dtls_key, EVP_sha1())) {
    RSA_free(rsa);
    EVP_PKEY_free(ortp_dtls_key);
    X509_free(ortp_dtls_certificate);
    ortp_dtls_certificate=NULL;
    ortp_dtls_key=NULL;
    return -1;
  }

  RSA_free(rsa);
  return 0;
}

int ortp_dtls_srtp_create(struct ortp_dtls_srtp **dtlsctx) {
  *dtlsctx = (struct ortp_dtls_srtp *)ortp_malloc0(sizeof(struct ortp_dtls_srtp));
  if (*dtlsctx)
  {
    gettimeofday(&(*dtlsctx)->connect_timeout, NULL);
    (*dtlsctx)->verify_enable=1;
    return 0;
  }
  return -1;
}

void ortp_dtls_srtp_dealloc(struct ortp_dtls_srtp *dtlsctx) {
  if (dtlsctx!=NULL) {
    if (dtlsctx->ssl) {
      SSL_free(dtlsctx->ssl);
      dtlsctx->ssl = NULL;
    }
    if (dtlsctx->ssl_ctx) {
      SSL_CTX_free(dtlsctx->ssl_ctx);
      dtlsctx->ssl_ctx = NULL;
    }
    if (dtlsctx->srtp_in) {
      srtp_dealloc(dtlsctx->srtp_in);
    }
    if (dtlsctx->srtp_out) {
      srtp_dealloc(dtlsctx->srtp_out);
    }
  }
  ortp_free(dtlsctx);
}

int ortp_dtls_srtp_get_sha_fingerprint(struct ortp_dtls_srtp *dtlsctx, char *fingerprint, unsigned int *size) {
  uint8_t fg[EVP_MAX_MD_SIZE];
  unsigned int len, i;
  len=sizeof(fg);
  if (1 != X509_digest(ortp_dtls_certificate, EVP_sha1(), fg, &len))
    return -1;

  sprintf(fingerprint, "sha-1 ");
  fingerprint+=6;
  for (i = 0; i < len; i++) {
    sprintf(fingerprint, "%.2X:", fg[i]);
    fingerprint += 3;
  }
  *(fingerprint-1) = 0;
  *size = i*3-1;
  return 0;
}

int ortp_dtls_srtp_get_sha256_fingerprint(ortp_dtls_srtp_t *dtlsctx, char *fingerprint, unsigned int *size) {
  uint8_t fg[EVP_MAX_MD_SIZE];
  unsigned int len, i;
  len=sizeof(fg);
  if (1 != X509_digest(ortp_dtls_certificate, EVP_sha256(), fg, &len))
    return -1;

  sprintf(fingerprint, "sha-256 ");
  fingerprint+=8;
  for (i = 0; i < len; i++) {
    sprintf(fingerprint, "%.2X:", fg[i]);
    fingerprint += 3;
  }
  *(fingerprint-1) = 0;
  *size = i*3-1;
  return 0;
}

int ortp_dtls_srtp_set_sha_remote_fingerprint(struct ortp_dtls_srtp *dtlsctx, char *rfingerprint) {
  if ((rfingerprint[0]=='S' || rfingerprint[0]=='s')
    &&(rfingerprint[1]=='H' || rfingerprint[1]=='h')
    &&(rfingerprint[2]=='A' || rfingerprint[2]=='a')
    &&rfingerprint[3]=='-'
    &&rfingerprint[4]=='1'
    &&rfingerprint[5]==' ')
    snprintf(dtlsctx->remote_fingerprint_sha1, sizeof(dtlsctx->remote_fingerprint_sha1), "%s", rfingerprint+6);
  else if ((rfingerprint[0]=='S' || rfingerprint[0]=='s')
    &&(rfingerprint[1]=='H' || rfingerprint[1]=='h')
    &&(rfingerprint[2]=='A' || rfingerprint[2]=='a')
    &&rfingerprint[3]=='-'
    &&rfingerprint[4]=='2'
    &&rfingerprint[5]=='5'
    &&rfingerprint[6]=='6'
    &&rfingerprint[7]==' ')
    snprintf(dtlsctx->remote_fingerprint_sha256, sizeof(dtlsctx->remote_fingerprint_sha256), "%s", rfingerprint+8);
  else
    return -1;
  return 0;
}

int ortp_dtls_srtp_set_setup_param(struct ortp_dtls_srtp *dtlsctx, char *setup_param) {
  if (0 == strcmp (setup_param, "active")) {
    dtlsctx->rfc4145_setup = RTP_SESSION_SETUP_ACTIVE;
  } else if (0 == strcmp (setup_param, "passive")) {
    dtlsctx->rfc4145_setup = RTP_SESSION_SETUP_PASSIVE;
  } else if (0 == strcmp (setup_param, "holdconn")) {
    dtlsctx->rfc4145_setup = RTP_SESSION_SETUP_HOLDCONN;
  }
  return 0;
}

#else

int ortp_dtls_srtp_create(struct ortp_dtls_srtp **dtlsctx) {
  *dtlsctx=NULL;
  return -1;
}
void ortp_dtls_srtp_dealloc(struct ortp_dtls_srtp *dtlsctx) {
}
int ortp_dtls_srtp_get_sha_fingerprint(struct ortp_dtls_srtp *dtlsctx, char *fingerprint, unsigned int *size) {
  return -1;
}
int ortp_dtls_srtp_get_sha256_fingerprint(struct ortp_dtls_srtp *dtlsctx, char *fingerprint, unsigned int *size) {
  return -1;
}

int ortp_dtls_srtp_set_sha_remote_fingerprint(struct ortp_dtls_srtp *dtlsctx, char *rfingerprint) {
  return -1;
}

int ortp_dtls_srtp_set_setup_param(struct ortp_dtls_srtp *dtlsctx, char *setup_param) {
  return -1;
}

#endif

int dtls_srtp_module_setup(void);

#if defined(HAVE_OPENSSL_SSL_H) && defined(HAVE_SRTP)

static int dtls_srtp_global_init() {
  return ortp_dtls_srtp_generate_selfsigned_cert();
}

static int dtls_srtp_global_shutdown() {
  if (ortp_dtls_certificate)
    X509_free(ortp_dtls_certificate);
  if (ortp_dtls_key)
    EVP_PKEY_free(ortp_dtls_key);
  ortp_dtls_certificate=NULL;
  ortp_dtls_key=NULL;
  return 0;
}
#endif

int dtls_srtp_module_setup(void) {

#if defined(HAVE_OPENSSL_SSL_H) && defined(HAVE_SRTP)
  RtpModule *tr = ortp_new0(RtpModule,1);
  snprintf(tr->name, sizeof(tr->name), "DTLS-SRTP");

  tr->t_global_init = dtls_srtp_global_init;
  tr->t_global_shutdown = dtls_srtp_global_shutdown;
  tr->t_set_option = NULL;

  tr->t_rtp_sendto=module_dtls_srtp_sendto;
  tr->t_rtp_recvfrom=module_dtls_srtp_recvfrom;
  tr->t_rtcp_sendto=module_dtls_srtcp_sendto;
  tr->t_rtcp_recvfrom=module_dtls_srtcp_recvfrom;
  tr->t_destroy=module_dtls_srtp_destroy;

  ortp_module_add(tr);
#endif
  return 0;
}
